#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <time.h>
#include <sys/time.h>
#include <memory.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <poll.h>
#include <signal.h>
#include <setjmp.h>

#include <math.h>
#include <stropts.h>	/* ioctl	*/
#include <semaphore.h>

#include "cdi.h"	// interface to user program
#include "cdiCanLib.h"

__attribute__((constructor)) void cdiSedacMain(void)
{
  int cc = 0;

  if ((cc=cdiRegisterBusFilter("CAN",filterCanbus)) != 0 ) goto err;
  if ((cc=cdiRegisterBusConfiguration("CAN",configCanCard)) != 0) goto err;
  if ((cc=cdiRegisterBusInitialization("CAN",initCanLineForCDI)) != 0 )
    goto err;
  if ((cc=cdiRegisterBusHandler("CAN",execCanbus)) != 0 ) goto err;
  if ((cc=cdiRegisterBusCleanup("CAN",terminateCanLine)) != 0 ) goto err;

err:
  if (cc)
  {
    cdilog("%s : failure in bus registration CAN",erlst[cc]);
  }
  return;
}

int canMsgPrint = 1;
#define MAX_BLEN 256

CAN_LINE canLine[MAX_NUMBER_CAN_LINES] = {
	{-1}, {-1}, {-1}, {-1}
};

int filterCanbus(int *parameter, int accessFlag)
{
  if (parameter == NULL) return argument_list_error;
  if (accessFlag != cdiWriteFlag &&
      (parameter[canOpenId] == PDO_TRANSMIT || parameter[canOpenId] == PDO2_TRANSMIT || parameter[canOpenId] == NMT_COMMANDS) ) return argument_list_error;
  return 0;
}
/****************************************************************************/
/* send request! input active flag = transRequest or transRequest2nd */
/* after, set active flag to finishRequest, recvRequest, or recvRequest2nd */
/****************************************************************************/
void execCanbus(CdiRequestInfoBlk *pReq)
{ 
  int canRet, wFlag;

  if ( pReq->active == transRequest ) /* start first transfere */
  {
    pReq->active = recvRequest; /* afterwards goto receive */
    if ( (pReq->accessFlag == cdiReadFlag || pReq->accessFlag == cdiReadWriteFlag) &&
    pReq->devParameters[canAutoReceive] ) return; /* auto receive, without request! */
    wFlag = (pReq->accessFlag == cdiWriteFlag || pReq->accessFlag == cdiWriteReadFlag) ? 1 : 0;
  }
  else
  {
    pReq->active = recvRequest2nd; /* goto 2ndReceive */
    if ( pReq->accessFlag == cdiWriteReadFlag && pReq->devParameters[canAutoReceive] ) return;
    wFlag = pReq->accessFlag == cdiReadWriteFlag ? 1 : 0;
  }

  canRet = canSend((int)((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine,
              (short)(pReq->devParameters[canBroadcast] ? 0 : pReq->pDev->cdiAddr),
              (short)pReq->devParameters[canOpenId], 
              (short)pReq->outLength, 
              (char *)pReq->pOutData,
              (short)pReq->devParameters[canOpenObj], 
              (short)pReq->devParameters[canOpenSubIndex], 
              (short)wFlag);

  if ( canRet <= 0 )
  {
    pReq->errorCode = -canRet; /* write failed */
    if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
    pReq->active = finishRequest;
    return;
  }
  /* without message back */
  if ( (pReq->devParameters[canOpenId] == PDO_TRANSMIT || pReq->devParameters[canOpenId] == PDO2_TRANSMIT || pReq->devParameters[canOpenId] == NMT_COMMANDS) &&
  pReq->accessFlag == cdiWriteFlag ) pReq->active = finishRequest;

  return;
}

/* Display the contents of a message prepending a text. */
/****************************************/
/*	show messages						*/
/****************************************/
void show_message(const char *text, Message *m)
{
  int spec = m->cmd;
  char    buf[MAX_BLEN];
  char    *mb;

  mb = &buf[0];

  if (spec < 0x20)
                format_modulmgr(mb, MAX_BLEN, m);
  else if (spec < 0x80)
                format_icanos(mb, MAX_BLEN, m);
  else if (spec == M_CMS_DATA)
                format_cmsdata(mb, MAX_BLEN, m);
  else if (spec == M_CMS_DEF)
                format_cmsdef(mb, MAX_BLEN, m);
  else if (spec == M_NMTS)
                format_nmts(mb, MAX_BLEN, m);
  else if (spec == M_NMTM)
                format_nmtm(mb, MAX_BLEN, m);
  else if (spec == M_DBTS)
                format_dbts(mb, MAX_BLEN, m);
  else if (spec == M_DBTM)
                format_dbtm(mb, MAX_BLEN, m);
  else if (spec == M_LMTS)
               format_lmts(mb, MAX_BLEN, m);

  printf("%s%s\n", text, mb);
  fflush(stdout);
}

/****************************************/
/*	can handle messages					*/
/****************************************/
void handle_message(int fd, Message *msg, int line)
{ int i;
  int spec = msg->cmd;
  BYTE_t *d = &msg->data[0];
  WORD_t subspec;
//  char data;

  subspec = word_from_bytes(d[0],d[1]);

  switch (spec)
  {
	case M_DEV_DEBUG:
		printf("Debug Msg arrived: %d %d-%d %d  : %d\n",
				d[0],d[1],d[2],d[3],d[4]);
		printf("                   %d %d\n",d[5],d[6]);
		return;
	case M_MEMORY:
		show_message("master> ", msg);
		return;
	case M_BCAN_RX_ind:
		printf("M_BCAN_RX_ind queue messages arrived: spec: %d, control: %d, len: %d\n",spec,msg->control,msg->len);
		for ( i=0; i<msg->len; i++ ) printf("byte index: %d, data: %x\n",i,msg->data[i] & 0xff);
		return;
	case M_BCAN_EVENT_ind:
		printf("M_BCAN_EVENT_ind(error, overrun) queue messages arrived: spec: %d, control: %d, len: %d\n",spec,msg->control,msg->len);
		for ( i=0; i<msg->len; i++ ) printf("byte index: %d, data: %x\n",i,msg->data[i] & 0xff);
//		data = NMT_START_REMOTE;
//		i = canSendAPI(line, -1, EXEC_REMOTE, 1, &data, 0, 0, 0);
//		printf("init remote control for can line: %d; can_send return: %d\n",line,i);
//		if ( initCanFlag != 0 ) resetLine(line);
		return;
	default:
		printf("Plain queue messages arrived: %d, %d, %d\n",spec,msg->control,msg->len);
		return;
	}
}

/****************************************/
/*	handle fast msg						*/
/****************************************/
void handleMsg(int line)
{ int i;
//  int j, change,k;
  CAN_LINE *pLine = &canLine[line];
  short obj;

  pLine->retId = (pLine->fmsg.data[0] << 3) + ( (pLine->fmsg.data[1] >> 5) & 7 );
  pLine->rtr = (pLine->fmsg.data[1] >> 4) & 0x1;
  pLine->length = pLine->fmsg.data[1] & 0xf;

  for ( i=0; i<pLine->length; i++ ) pLine->retData[i] = pLine->fmsg.data[i+2];

  if ( canMsgPrint )
  {
    printf("Master ID %x, rtr: %d; length: %d : ", pLine->retId, pLine->rtr, pLine->length);
    for(i=0; i<pLine->length; i++) printf("%2x ", ((int)pLine->retData[i] & 0xff));
    printf("\n");
  }
  pLine->remoteNode = pLine->retId & 0x7f;
  pLine->remoteCommand = pLine->retId & 0xff80;

  if ( pLine->receiveCanMessage ) pLine->receiveCanMessage(pLine->cdiLine, pLine->retData, pLine->length, pLine->retId, pLine->rtr, pLine->remoteNode, pLine->remoteCommand);

//  i = getModuleNumber(pLine, pLine->remoteNode);
//  if ( i < 0 ) return;	// not in the device list, wrong!
//  pD = &pLine->pDevice[i];

   obj = *(short *)&pLine->retData[1]; 
}

/****************************************/
/*	can receive thread					*/
/****************************************/
void *GetCanDataThread(void *param)
{ CAN_LINE *pLine = (CAN_LINE *)param;
  unsigned char	index;
  int n, i;

  do 
  {
    n = read(pLine->fd, &index, 1);		// wait on can message
    if(n == 0)
    {
      fprintf(stderr,"(2) panic read ..., can read crash, line: %d\n", pLine->lineNumber);
      exit(1);
    }
    switch(index)
    {
	    case FAST_QUEUE:
        if( can_recv_fast( pLine->fd, &(pLine->fmsg) ) < 0 )
        {
          fprintf(stderr,"(2) can_recv_fast() crash, line: %d\n", pLine->lineNumber);
          exit(1);
        }
        if ( (pLine->fmsg.cmd & 0xf0) != 0 && pLine->canCardType == PC104_CARD )
        {
	        for( i=0; i<MAX_NUMBER_CAN_LINES; i++ )	// find right line
	        {
            if ( canLine[i].fd < 0 ) continue;
            if ( canLine[i].dpm_x != pLine->dpm_x ) continue;
            if ( canLine[i].dpm_y == pLine->dpm_y ) continue;
            break;
	        }
	        memcpy (&(canLine[i].fmsg), &(pLine->fmsg), sizeof(FastMessage));
	        handleMsg(i);
        }
        else
          handleMsg(pLine->lineNumber);
        break;
	    case PLAIN_QUEUE:
        if( can_recv(pLine->fd, &(pLine->msg)) < 0 )
        {
          fprintf(stderr,"(2) can_recv() crash, line: %d\n", pLine->lineNumber);
          exit(1);
        }
        handle_message(pLine->fd, &(pLine->msg), pLine->lineNumber);
        break;
    }
  } while(1);
}

/****************************************/
/*	CAN send API						*/
/*	return -1: line number wrong		*/
/*	return -2: not init yet				*/
/*	return =0: send error				*/
/****************************************/
int canSend(int line, short nodeId, short canOpenId, short length, char *data, short ObjNum, short subIndex, short wFlag)
{ CAN_LINE *pCan;
  FastMessage	FastMsg;
  int c, rtr;

  if ( line < 1 || line > MAX_NUMBER_CAN_LINES ) return -illegal_line;
  line--;

  pCan = &(canLine[line]);
  if ( pCan->fd < 0 ) return -file_not_open;

  FastMsg.cmd = 0;
  if ( pCan->canCardType == PC104_CARD && (pCan->dpm_y % 2) == 1 ) FastMsg.cmd = (1 << 4);
  if ( nodeId < 0 ) nodeId = 0;
//  if ( canLine[line]->canCardType == PC104_CARD && (canLine[line]->dpm_y % 2) == 1 ) FastMsg.cmd = (line << 4);
//  nodeId = devNum < 0 ? 0 : canLine[line]->pDevice[devNum].NodeID;

//  printf("***canSend ==> line: %d; nodeID: %d; CanOPENID: %d\n",line, nodeId, canOpenId);

  switch (canOpenId)
  {
	case NMT_COMMANDS:
	  canOpenId = NMT_COMMANDS;
	  rtr = 0;
	  length = 2;
	  FastMsg.data[2] = data[0];	// only one byte data for EXEC_REMOTE
	  FastMsg.data[3] = nodeId;
	  break;
	case SDO_TRANSMIT:
	  FastMsg.data[2] = wFlag ? SDO_WRITE : SDO_READ;
	  *(short *)&FastMsg.data[3] = ObjNum;	// object 6300; sub index 1
	  FastMsg.data[5] = subIndex;
	  if ( wFlag ) memcpy(&FastMsg.data[6],data,length);

	  if ( wFlag && length == 4 ) printf("can save data: %x\n",*(int *)&FastMsg.data[6]);

	  length = 8;
	  rtr = 0;
	  break;
	case PDO_TRANSMIT:
	case PDO2_TRANSMIT:
	  memcpy(&FastMsg.data[2],data,length);
	  rtr=0;
	  break;
	case PDO_DI_READ:
	case PDO_AI_READ:
	  length = 0;
	  rtr = 1;
	  break;
	default:
	  printf("canSendAPI: ****Wrong CanOpenID: 0x%x\n",canOpenId);
	  return -3;
  }
  c = canOpenId + nodeId;
  FastMsg.data[0] = (c >> 3) & 0xff;
  FastMsg.data[1] = ((c << 5 ) + (rtr ? 0x10 : 0) + length) & 0xff;

  // use can_send for channel 0 in case of pc104
  if ( pCan->canCardType == PC104_CARD && (pCan->dpm_y % 2) == 0 )
  {
//	IcSendReqBCAN(&msg, 0, c=0, (FastMsg.data[0] << 8) | FastMsg.data[1], &FastMsg.data[2]);
//	return can_send(canLine[line]->fd, &msg);
    c = can_fast_send_prio(pCan->fd, HI_PRIORITY, &FastMsg); // swh
  }
  c = can_fast_send(pCan->fd, &FastMsg);

  if ( c <= 0 )
  {
	pCan->numSendErrors++;
//	canLine[line]->pDevice[devNum].numModuleSendErrors++;
//	canLine[line]->pDevice[devNum].errCode |= SEND_ERROR;
  }
//  else canLine[line]->pDevice[devNum].errCode &= ~SEND_ERROR;

//  printf("Can send out! return code: %d (<=0 error)\n",c); 
  return c <= 0 ? -hardware_error : c;
}

/****************************************/
/*	load Can card information			*/
/****************************************/
int configCanCard(void)
{ int i;
  CAN_LINE *pCan;

  for ( i=0; i<MAX_NUMBER_CAN_LINES; i++ )
  {
    pCan = &(canLine[i]);
    pCan->canRate = (i%2) == 0 ? BTR_500KB : BTR_20KB;
    pCan->canCardType = JANZ_PC_CARD;
    pCan->dpm_x = i/2;		// card number
    pCan->dpm_y = i%2;		// channel number
    pCan->lineNumber = i;
  }
  return 0;
}

/****************************************/
/*	translateBaudRate					*/
/****************************************/
int transBaudRate(int baudrate)
{ int rate;
  switch ( baudrate )		// set baud rate
  {
	case 800:
		rate = BTR_800KB;
		break;
	case 500:
		rate = BTR_500KB;
		break;
	case 250:
		rate = BTR_250KB;
		break;
	case 125:
		rate = BTR_125KB;
		break;
	case 50:
		rate = BTR_50KB;
		break;
	case 20:
		rate = BTR_20KB;
		break;
	case 10:
		rate = BTR_10KB;
		break;
	case 1000:
		rate = BTR_1MB;
		break;
	default:
		printf(" baud rate not supported: %d",baudrate);
		return -1;
  }
  return rate;
}

/****************************************/
/*	init can line						*/
/****************************************/
/**************************************/
void canMessage(int cdiLine, char *data, int leng, int retId, int rtr, int remoteNode, int remoteCommand)
{ 
  /* CdiLineInfoBlk *pLine = busHandle[cdiLine]; */
  CdiRequestInfoBlk *pReq;
  short readFlag, returnObj;
  char *str;
  int datalen;
  /*
  if (NGdebug) 
    dbglog("CanMessage call Back: cdi line: %d, can line number: %d, data: %s, length: %d, retId: %d, rtr: %d",
            cdiLine, busHandle[cdiLine]->busLine, data, leng, retId, rtr);
  */
  /* find out device */
  for ( busHandle[cdiLine]->readCounter=0; busHandle[cdiLine]->readCounter < busHandle[cdiLine]->numberActiveRequest; busHandle[cdiLine]->readCounter++ )
  {
    if ( (busHandle[cdiLine]->pReqList[ busHandle[cdiLine]->readCounter ].active != recvRequest &&
      busHandle[cdiLine]->pReqList[ busHandle[cdiLine]->readCounter ].active != recvRequest2nd) ||
      busHandle[cdiLine]->pReqList[ busHandle[cdiLine]->readCounter ].pDev->cdiAddr != remoteNode ) continue;

    pReq = &busHandle[cdiLine]->pReqList[ busHandle[cdiLine]->readCounter ];
    if ( pReq->active == recvRequest ) /* first part */
      readFlag = (pReq->accessFlag == cdiReadFlag || pReq->accessFlag == cdiReadWriteFlag) ? 1 : 0;
    else /* second part! */
      readFlag = (pReq->accessFlag == cdiWriteReadFlag) ? 1 : 0;
    returnObj = *(short *)&data[1];

    datalen = leng;
    str = data;
    if ( pReq->devParameters[canOpenId] == SDO_TRANSMIT ) /* check messagesType */
    {
      if ( remoteCommand != SDO_RECEIVE ) continue;
      if ( readFlag ) { if ( (data[0] & 0xf0) != 0x40 ) continue; } /* SDO read */
      else if ( (data[0] & 0xf0) != 0x60 ) continue; /* SDO write */
      if ( pReq->devParameters[canOpenObj] != returnObj || (char)pReq->devParameters[canOpenSubIndex] != data[3] ) continue;
      /* SDO data back.... */
      datalen = leng - 4;
      str = &data[4];
    }
    else if ( pReq->devParameters[canOpenId] != remoteCommand ) continue; /* same OpenId? */
    else if ( readFlag == 0 ) /* for write command device should not answer! */
    {
      /*
      if (NGdebug) 
      {
        dbglog("canMessage Wrong: unexpected return message for write conmmand!");
        dbglog("CanOpenId: %d; canLine: %d; module CanId: %d; module Name: %s",
                remoteCommand,busHandle[cdiLine]->busLine,remoteNode,pReq->pDev->devName);
      }
      */
      continue;
    }

    if ( readFlag ) /* check return condition */
    {
      checkReturnCondition(pReq, pReq->pInData, str, datalen, 0);
      if ( pReq->dataReturnCondition == 0 ) continue; 
    }
    /*
    if (NGdebug) dbglog("can data return condition OK canOpenID: %d",remoteCommand);
    */
    memcpy( pReq->pInData, str, datalen > pReq->inLength ? pReq->inLength : datalen);

    cdiExecRequest(&(busHandle[cdiLine]->pReqList[busHandle[cdiLine]->readCounter]), readTaskIndex);
  }
}
#define CDI_PARAMETER_DELIMETER ":"
int cdiParseParamString(char *paramstr,int *parameters)
{
  char *s,*c=NULL;
  int i;
  if (paramstr == NULL || parameters == NULL) return argument_list_error;
  for (s=strtok_r(paramstr,CDI_PARAMETER_DELIMETER,&c), i=0; 
       i<NUMBER_EXEC_PARAM && s != NULL; 
       s=strtok_r(NULL,CDI_PARAMETER_DELIMETER,&c), i++)
  {
    parameters[i] = atoi(s);
  }
  return 0;
}
int initCanLineForCDI(int busLine, int cdiLine, int numberDevices, char *parameterList)
{
  int parameter[NUMBER_EXEC_PARAM];
  memset(parameter,0,NUMBER_EXEC_PARAM*sizeof(int));
  cdiParseParamString(parameterList,parameter);
  return initCanLine(busLine,parameter,canMessage,cdiLine,numberDevices);
}
int initCanLine(int line, int *parameters, void (*callBack)(int, char *, int, int, int, int, int), int cdiLine, int numberDevices)
{ Message	msg;
  int channel, i, ana;
  CAN_LINE *pCan;

  if ( line < 1 || line > MAX_NUMBER_CAN_LINES ) return -illegal_line;
  line--;

  pCan = &(canLine[line]);
  if ( pCan->fd >= 0 )
  {
    printf("Can line: %d already opened! \n",line);
    return pCan->fd;
  }

  pCan->canRate = transBaudRate( parameters[paramCanRate] );
  if ( pCan->canRate == -1 )
  {
    printf("Can line: %d; with Baudrate: %d not supported! \n",line,parameters[paramCanRate]);
    return -argument_list_error;
  }
  pCan->NumberDevices = numberDevices;
  pCan->cdiLine = cdiLine;
  pCan->receiveCanMessage = callBack;
  
  sprintf(pCan->devPath, "/dev/dpm_%d%d", pCan->dpm_x, pCan->dpm_y);

//  printf("can open: rate: %d; ndevice: %d; canPath: %s;cardType: %d\n",pCan->canRate,pCan->NumberDevices,pCan->devPath, pCan->canCardType);
  
  // Turn the master module (1st on the carrier board) on
  if ( (pCan->dpm_y % 2) == 0 || pCan->canCardType == JANZ_PC_CARD || pCan->canCardType == JANZ_PC_CARD_RAW)
  {
    pCan->fd = can_open(pCan->devPath);
    if ( pCan->fd < 0 )
    {
      printf("Error: can_open(%s) failed!\n", pCan->devPath);
      return -file_not_open;
    }
    printf("can_open(%s) done! fd: %d\n", pCan->devPath, pCan->fd);

	// Select fast queue
    ana = pCan->NumberDevices / 16 ;	// multipul of 16 buffers
    ana = (ana + 1) * 32;
    ican2_select_hostif(pCan->fd, ana, ana);
//	if( ican2_init_fast_can(canLine[line]->fd, ana, ana) < 0 )
    if( ican2_init_fast_can_prio(pCan->fd, ana, ana,3) != 0 )  //swh
    {
      fprintf(stderr,"ican2_init_fast_can failed!");
      can_close( pCan->fd );
      pCan->fd = -1;
      return -hardware_error;
    }

	// pc 104 style here
	if ( pCan->canCardType == PC104_CARD && (pCan->dpm_y % 2) == 0 )
	{
    // Enable second CAN controller facilities
    printf ("\tEnable 2nd CAN controller usage for PC104 ... "); 
    SwitchCanControllerInterpretation(&msg, 1);
    // send 'baud rate' message 
    if (can_send(pCan->fd, &msg) <= 0) 
    {
      printf ("ERROR : Unable to enable 2nd CAN controller(line number: %d)\n",line); 
      can_close( pCan->fd );
      pCan->fd = -1;
      return -hardware_error;
    }

    // Now we have two CAN channels available and we also need to configure two CAN channels !!!
    for (channel=0; channel<2; channel++)
    {
      sleep (2);

      if ( channel == 0 )
      {
        LsBusOnReq(&msg, pCan->canRate);
        patchCcNo (&msg, channel);
        while ( can_send(pCan->fd, &msg) <= 0 );
      }
      else
      {
		  // create message for baud rate settings //?????????
        IcWriteBtrBCAN(&msg, canLine[line+1].canRate);
        // set CAN controller number in 'Message' data structure
        patchCcNo (&msg, channel);
        // send 'baud rate' message 
        while ( can_send(pCan->fd, &msg) <= 0 );
        sleep(1);

        // create message for 'BusOn' message 
        IcBusOnBCAN(&msg);
        patchCcNo (&msg, channel);
        while ( can_send(pCan->fd, &msg) <= 0 );
      }		  
      sleep (1);

      IcRangeSetAfil(&msg, FILTER_LOW, FILTER_HIGH, 2);
      patchCcNo (&msg, channel);
      while ( can_send(pCan->fd, &msg) <= 0 );
 	  }
 	}
  else
  {
	  if ( pCan->canCardType == JANZ_PC_CARD_RAW )
	  {
		  IcWriteBtrBCAN(&msg, pCan->canRate);
		  while ( can_send(pCan->fd, &msg) <= 0 );
		  sleep(1);
		  IcBusOnBCAN(&msg);
		  while ( can_send(pCan->fd, &msg) <= 0 );
	  }
	  else
	  {
      LsBusOnReq(&msg, pCan->canRate);
      while ( can_send(pCan->fd, &msg) <= 0 );
	  }
	  sleep(1);

	  // Set Massage filter
	  IcRangeSetAfil(&msg, FILTER_LOW, FILTER_HIGH, 2);
	  while ( can_send(pCan->fd, &msg) <= 0 );
	}
	// Waiting for messages and work them out on master module
	if ( pthread_create( &(pCan->tIdMaster), NULL, GetCanDataThread, (void *)pCan ) )
	{
	  printf ("pthread_create line: %d error",line);
	  can_close( pCan->fd );
	  pCan->fd = -1;
	  return -un_allocated;
	}
  }
  else
  {
    for( i=0; i<MAX_NUMBER_CAN_LINES; i++ )	// find right line
    {
      if ( canLine[i].fd < 0 ) continue;
      if ( canLine[i].dpm_x != pCan->dpm_x ) continue;
      if ( canLine[i].dpm_y == pCan->dpm_y ) continue;
      break;
    }
    if ( i >= MAX_NUMBER_CAN_LINES )
    {
      printf("ERROR! can_open for pc104 style has NOT been down!!!! channel number: (%d) \n", line);
      return -not_ready;
    }
    printf("can_open for pc104 style has been down; channel number: (%d) \n", line);
    pCan->fd = canLine[i].fd;
    strcpy(pCan->devPath, canLine[i].devPath);
    pCan->tIdMaster = canLine[i].tIdMaster;
  }

  sleep(1);	// wait for all device become remote contrllable

  pCan->numSendErrors = 0;
  pCan->statusOK = 0;

  return pCan->fd;
}

/****************************************/
/*	terminate can line					*/
/****************************************/
int terminateCanLine(int line)
{ CAN_LINE *pCan;

  if ( line < 1 || line > MAX_NUMBER_CAN_LINES ) return -illegal_line;
  line--;

  pCan = &(canLine[line]);
  if ( pCan->fd < 0 ) return -file_not_open;

  pthread_cancel( pCan->tIdMaster );
  can_close( pCan->fd );
  pCan->fd = -1;

  return 0;
}
