#ifndef CDIRS232_H_
#define CDIRS232_H_

#include "termios.h"

int RegisterRS232Bus(void);
void execRS232(CdiRequestInfoBlk *pReq);
int sendSerialPort(int channelNumber, char *psOutput);
int closeSerialPort(int channelNumber);
int openSerialPort(int channelNumber, int *parameters, void (*callBack)(int, char*, int), int cdiLine);
int initSerialPortForCDI(int busLine, int cdiLine, int numberDevices, char *parameterList);

#define N_SERIAL_LINES	4	// com1 to com4

#define STRING_LENGTH	255

/* open serial parameter index */
enum openSerialParamIndex
{
  paramBaudrate, /* e.g 9600 */
  param8bits, /* flag 1 or 0... */
  param2StopBits,
  paramParity,
  paramOddParity,
  paramFlowControl,
  paramCanonicalInput
};

typedef struct {
  int inputBaud, defBaud;
} BAUD_STRUCT;

#define REALLY_HAVE_RS232
#ifdef REALLY_HAVE_RS232
typedef struct {
  int serFd,		// serial File descripter
	  channel,		// serial port channel, start from 0,1,2,3...
	  cdiLine,		// cdi Line index for getString then callback
	  baudRate,		// baudRate
	  canonicalMode,
	  usOneChar;	// time in us for one Char transmission
  struct termios oldtio;	// stored tio
  char	buf[STRING_LENGTH + 1];	// input string
  
  pthread_t	tIdMaster;		// receive task variables
  fd_set fdset;
  struct timeval tout;
  int ret, length;

  void (*receiveString)(int cdiLine, char *string, int length);
} SER_STRUCT;

int RegisterRS232Bus(void);
void rs232String(int cdiLine, char *str, int leng);
void execRS232(CdiRequestInfoBlk *pReq);
int initSerialPortForCDI(int busLine, int cdiLine, int numberDevices, char *parameterList);
int closeSerialPort(int channelNumber);
int sendSerialPort(int channelNumber, char *psOutput);

extern SER_STRUCT serInfo[N_SERIAL_LINES];

extern BAUD_STRUCT baudRate[];
extern int numBaudRate;

#endif

#endif /* CDIRS232_H_ */
