#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#ifdef UNIX
#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/kd.h>
#include <sys/time.h>
#include <dlfnc.h>
#endif
#include <sys/types.h>
#include <linux/unistd.h>

#include "prolog.h"
#include "cdi.h"	// interface to user program
#include "SedPc.h"
#include "cdiSedpcLib.h"

#define SD_WRITE 0x40
#define SD_READ  0x80

#define N_SEDAC_LINES 8
#define MAX_SEDAC_PUT 350   /* number of telegrams to put (4k FIFO 680 telegrams) */
#define WAIT_TIME_NS 1000000  // 1ms
int nSedacLines = N_SEDAC_LINES;
#define N_THREAD_IDS  4
typedef struct
{
  int numberTelegrams, cdiLine;
	void *sedBuffer;
  int numberOfThreadIds, threadIDs[N_THREAD_IDS];
} SEDPC_DATA;
SEDPC_DATA sedacInfo[N_SEDAC_LINES] = { /* init */
  {0, -1, NULL, 0}, {0, -1, NULL, 0}, {0, -1, NULL, 0}, {0, -1, NULL, 0},  {0, -1, NULL, 0}, {0, -1, NULL, 0}, {0, -1, NULL, 0}, {0, -1, NULL, 0}
};

char *scanCommand[] = {"getcommands","getlines","getcrates","getmodules","readmodules","writemodules"};
enum scommand {getcommands, getlines, getcrates, getmodules, readmodules, writemodules, nCommands};

extern int setIOpermission(int lineNumber, int on);
 
__attribute__((constructor)) void cdiSedacMain(void)
{
    int cc = 0;

    if ((cc=cdiRegisterBusFilter("SEDPC",filterSedacBus)) != 0 ) goto err;
    if ((cc=cdiRegisterBusInitialization("SEDPC",initSedacForCDI)) != 0 ) goto err;
    if ((cc=cdiRegisterBusHandler("SEDPC",execSedacBus)) != 0 ) goto err;
    if ((cc=cdiRegisterBusCleanup("SEDPC",exitSedac)) != 0 ) goto err;
err:
    if (cc)
    {
      cdilog("%s : failure in bus registration SEDAC",erlst[cc]);
    }
    return;
}


int sedacScanFcn(char *target,char *result,int resultBufferSize)
{
  char *c, lclTarget[256];
  int i, j, cnt, ret, slen = 0;
  int nData, data[10];
  short rData, wData;
  short line = -1, crate = -1, subaddr = -1;

  if (result == NULL) return argument_list_error;
  if (resultBufferSize < 32) return buffer_too_small;
  // parse target according to "commandString:<line>:<crate>:<subaddr>"; e.g. NULL or "command:1" or "command:1:48" or "command:1:48:16" 
  if (target == NULL || strlen(target) == 0 ) target = "getcommands";
  strncpy(lclTarget,target,255);
  strlwr(lclTarget);
  // get command string
  if ((c=strtok(lclTarget,":")) == NULL) return 0;  // nothing to do
  for (j=0; j<nCommands; j++) if ( strcmp(scanCommand[j], c) == 0 ) break;
  if ( j == nCommands ) j = getcommands; // return the available commands as 'help'
  for ( nData=0; (c=strtok(NULL,":")) != NULL; nData++ ) data[nData] = strtol(c,NULL,0); //atoi(c);

  
  if (j == getcommands)
  {
    snprintf(result,resultBufferSize,"availabled command strings:\n");
    for (i=0; i<nCommands; i++)
    {
      if ((slen = strlen(result)) > resultBufferSize - 12) break;
      strncat(result,scanCommand[i],resultBufferSize-slen);
      if (i <nCommands-1) strcat(result,",");
    }
    return 0;
  }
  
  
  if ( j == getlines )
  {
    snprintf(result,resultBufferSize,"Sedac lines configured:\n");
    for (i=0, cnt=0; i<N_SEDAC_LINES; i++)
    {
      if ( sedacInfo[i].cdiLine < 0 ) continue;
      if ((slen = strlen(result)) > resultBufferSize - 1) break;
      if ( cnt > 0 )
      {
        snprintf(&result[slen],resultBufferSize-slen,",");
        if ((slen = strlen(result)) > resultBufferSize - 1) break;
      }
       snprintf(&result[slen],resultBufferSize-slen,"%d",i+1);
      cnt++;
    }
    return 0;
  }
  
  line = data[0];
  if ( nData < 1 || line < 1 || line > nSedacLines || sedacInfo[line-1].cdiLine < 0) return no_such_line;
  if ( j == getcrates )
  {
    snprintf(result,resultBufferSize,"crates configured on line %d:\n",line);
    for (crate=0, cnt=0; crate<32; crate++)
    {
      ret = sedac(line, crate, 253, CA_READ, &rData, 1);
      if ( ret ) continue;
      if ((slen = strlen(result)) > resultBufferSize - 1) break;
      if ( cnt > 0 )
      {
        snprintf(&result[slen],resultBufferSize-slen,",");
        if ((slen = strlen(result)) > resultBufferSize - 1) break;
      }
      snprintf(&result[slen],resultBufferSize-slen,"%d",crate);
      cnt++;
    }
    return 0;
  }
  else if ( j == getmodules )
  {
    crate = data[1];
    if ( nData < 2 || crate < 0 || crate >= 32 ) return device_not_connected;
    snprintf(result,resultBufferSize,"Subaddress on line %d crate %d was found:\n",line,crate);
    for (subaddr=0, cnt=0; subaddr<256; subaddr++)
    {
      ret = sedac(line, crate, subaddr, CA_READ, &rData, 1);
      if ( ret ) continue;
      if ((slen = strlen(result)) > resultBufferSize - 1) break;
      if ( cnt > 0 )
      {
        snprintf(&result[slen],resultBufferSize-slen,",");
        if ((slen = strlen(result)) > resultBufferSize - 1) break;
      }
      snprintf(&result[slen],resultBufferSize-slen,"%d",subaddr);
      cnt++;
    }
    return 0;
  }
  crate = data[1];
  subaddr = data[2];
  if ( nData < 3 || crate < 0 || crate >= 32 || subaddr < 0 || subaddr > 255) return device_not_connected;
  if ( j == readmodules )
  {
    snprintf(result,resultBufferSize,"Read Data on line %d crate %d subaddr %d:\n",line,crate,subaddr);
    ret = sedac(line, crate, subaddr, CA_READ, &rData, 1);
    if ((slen = strlen(result)) < resultBufferSize - 18) snprintf(&result[slen],resultBufferSize-slen,"%d <return code %d>",rData,ret);
  }
  else if ( j == writemodules )
  {
    wData = data[3];
    if ( nData < 4 ) return argument_list_error;
    snprintf(result,resultBufferSize,"Write Data on line %d crate %d subaddr %d:\n",line,crate,subaddr);
    ret = sedac(line, crate, subaddr, CA_WRITE, &wData, 1);
    if ((slen = strlen(result)) < resultBufferSize - 18) snprintf(&result[slen],resultBufferSize-slen,"%d <return code %d>",wData,ret);
  }  
  return 0;
}

int waitSedPCfinished()
{ int ln;
//  CdiLineInfoBlk *busHandle;

  for (ln=0; ln<N_SEDAC_LINES; ln++)
  {
    if ( sedacInfo[ln].numberTelegrams <= 0 ) continue;
//    if ((busHandle=cdiGetBusHandle(sedacInfo[ln].cdiLine)) == NULL) return -1;  /* something wrong!!  */
//    cdiMutexLock(&(busHandle->lineMutex));
//    cdiMutexUnlock(&(busHandle->lineMutex));
    exitSedac(ln+1);
  }
  return 0;
}

int filterSedacBus(int *parameter, int accessFlag)
{
  return 0;
}
int initSedacForCDI(int busLine, int cdiLine, int numberDevices, char *parameterList)
{
//  cdi_mutex_t tmpMutex;
  int lastError, error = 0;

  /* check environemnt variables to see what's connected: Stadmueller FIFO  */
  if ( parameterList[0] )
  if ( error=setenv("SEDPC",parameterList,1) )
  {
    cdilog("initSedac setenv error: line: %d; error: %d\n",busLine, error);
    return -non_existent;
  }
  else cdilog("initSedac setenv ok! line: %d\n",busLine);

  /* try to get Mutex, is this a first application  */
//  lastError = cdiMutexInit(&tmpMutex, cdiLine);
//  if ( tmpMutex == NULL ) return -ask_error;
//  if ( lastError == ERROR_ALREADY_EXISTS ) cdiMutexLock(&tmpMutex);
  error = initSedac(busLine);
//  if ( lastError == ERROR_ALREADY_EXISTS ) cdiMutexUnlock(&tmpMutex);
  if ( error == 0 ) sedacInfo[busLine-1].cdiLine = cdiLine;
  
  return error;
}

// if sync call, data not pass filter then repeat Transmit
int checkData(CdiRequestInfoBlk *pReq)
{
  struct timespec tv;

  /* check return condition */
  checkReturnCondition(pReq, (char *)pReq->bufLastData, pReq->pInData, pReq->inLength, 0);
  if ( pReq->dataReturnCondition == DATA_PASS_FILTER)
    memcpy( pReq->bufLastData, pReq->pInData, pReq->inLength < 8 ? pReq->inLength : 8);

 	if ( pReq->conditionFlag > anyData && pReq->dataReturnCondition != DATA_PASS_FILTER )
	{
    if ( pReq->callback == NULL )
    {
      if ( getTimeMilliseconds() - pReq->requestTime >= pReq->timeout )
      {
         pReq->errorCode = invalid_data;	// data back for sync call
         if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
      }
      else
      {
        tv.tv_sec = 0; tv.tv_nsec = WAIT_TIME_NS; // delay 1ms
        nanosleep( &tv, NULL );    
        return 1;
      }
    }
    pReq->active = finishRequest;
  }
  return 0;
}

void execSedacBus(CdiRequestInfoBlk *pReq)
{ 
  /* do first part; */
  do
  {
    doSedacRequest(pReq, 1);
    if ( pReq->active == finishRequest ) return;

    /* check return condition */
    if (pReq->accessFlag != cdiReadFlag && pReq->accessFlag != cdiReadWriteFlag) break;
    if ( checkData(pReq) == 0 ) break;
  } while (TRUE);

  if ( pReq->active == finishRequest ) return;

  if ( pReq->accessFlag == cdiReadFlag || pReq->accessFlag == cdiWriteFlag)
  {
    pReq->active = finishRequest;
    return;
  }

  /* do second part */
  do
  {
    doSedacRequest(pReq, 2);
    if ( pReq->active == finishRequest ) return;

    /* check return condition */
    if (pReq->accessFlag != cdiWriteReadFlag && pReq->accessFlag != cdiWriteReadWriteFlag) break;
    if ( checkData(pReq) == 0 ) break;
  } while (TRUE);

  if (pReq->accessFlag != cdiWriteReadWriteFlag && pReq->accessFlag != cdiWriteReadReadFlag)
  {
    pReq->active = finishRequest; /* finish request */
    return;
  }

  doSedacRequest(pReq, 3);
  pReq->active = finishRequest; /* finish request */
}

int clearSedac(int ln)
{ long err=0,dly=0;

//////////////  SedClr(ln,&dly,&err);
  if ( err )  /* twice clear needed if there was an error befor */
  {
    millisleep(20);
/////////////    SedClr(ln,&dly,&err);
  }
  if ( err ) cdilog("SedClr error: %d with line: %d\n",err,ln);
  return err;
}
int initSedac(int ln)
{
  char *env;
  char str[80];
  long err=0;
  long	xglLogLin[XMLN];		// logical LineNumbers
  long	xglDevAdr[XMLN];		// DeviceAddr of Lines
  long	xglModAdr[XMLN];		// ModuleAddr of Lines (from Sed1Pc)

  if (ln<1 || ln>nSedacLines) return -illegal_line;
  /* check environemnt variables to see what's connected: Stadmueller FIFO  */
//  if ( ( err=setenv("SEDPC","1-378",1) ) )
//  {
//    cdilog("initSedac setenv error: line: %d; error: %d\n",ln, err);
//    return -non_existent;
//  }
  if ( ( env=getenv("SEDPC") ) == NULL )
  {
    cdilog("initSedac getenv error: line: %d\n",ln);
    return -non_existent;
  }
  cdilog("initSedac getenv: %s; line: %d\n",env, ln);

//////////////  SedIni(xglLogLin, xglDevAdr, xglModAdr, &err);

  cdilog("open sedpc line: %d; total line: %d; %d; error: %d\n",ln, xglLogLin[0], xglLogLin[1], err);

  if ( err ) return -file_not_open;

  err = clearSedac((short)ln);
//  printf(" get tid: %d\n",(long)getpid());

  sedacInfo[ln - 1].threadIDs[0] = getpid();  // store the first PID
  sedacInfo[ln - 1].numberOfThreadIds = 1;

  /*  sedacInfo[ln - 1].numberTelegrams = 0;  /* set buffer size  */
  return err ? -not_ready : 0;
}
int exitSedac(int ln)
{ long err;

  if (ln<1 || ln>nSedacLines) return -illegal_line;
  cdilog("close sedpc line! line number: %d\n", ln);
  if ( sedacInfo[ln - 1].numberTelegrams )
  {
    free ( sedacInfo[ln - 1].sedBuffer );
    sedacInfo[ln - 1].numberTelegrams = 0;
    sedacInfo[ln - 1].sedBuffer = NULL;
  }
////////  SedExi ( &err );
  return err;
}

typedef struct
{
  long TypeAndCrate;
  long BaseSubAddress;
  long data;
} SedpcIn;
typedef struct
{
  long status;
  long data;
} SedpcOut;

int sedpcBlock(WORD lin, WORD cra, WORD bsa, int access, WORD *data, int begin, int ende, int putFlag)
{ int i, j, status=0;
  long ntelegrams, tmp;
  SedpcIn *sedin;
  SedpcOut *sedout;

  ntelegrams = ende - begin;
  if ( putFlag )
  {
    sedin = (SedpcIn *)sedacInfo[lin - 1].sedBuffer;
    tmp = ( access == CA_READ ? SD_READ : SD_WRITE ) | cra;
    for (j=begin; j<ende; j++)
    {
      i = j - begin;
      sedin[i].TypeAndCrate = tmp;
      sedin[i].BaseSubAddress = bsa;
      sedin[i].data = access == CA_READ ? 0 : ((WORD *)data)[j];
    }
//    printf("sedput: begin: %d; end: %d\n",begin, ende);
//    p = (long *)sedin;
//    printf("data: 0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x,\n",p[0],p[1],p[2],p[3],p[4],p[5]);
    SedPut(lin,&ntelegrams,(long *)sedin,(long *)&status);
//    printf("sedput: begin: %d; end: %d\n",begin, ende);
  }
  else
  {
    sedout = (SedpcOut *)sedacInfo[lin - 1].sedBuffer;
    SedGet(lin,&ntelegrams, (long *)sedout, (long *)&status);
    if (access == CA_READ) for (i=0; i<ntelegrams; i++) ((WORD *)data)[i+begin] = sedout[i].data;
  }
  if ( status )
  {
    printf("!!!SedPC put or get(putflag: %d) error: 0x%x with line: %d; begin number: %d\n",putFlag,status,lin, begin);
    printf("!!!SedPC put or get(putflag: %d) error: 0x%x with line: %d; begin number: %d\n",putFlag,status,lin, begin);
    printf("!!!SedPC put or get(putflag: %d) error: 0x%x with line: %d; begin number: %d\n",putFlag,status,lin, begin);
  }
  return status;
}

// ================  read/write nwords telegram to same addesses  ===================================
int sedac(WORD lin, WORD cra, WORD bsa, int access, WORD *data, int ntelegrams)
{ 
  int i, steps, ret=0;
  static WORD storedWord = 0;

  if (ntelegrams == 1)
  {
    if (access == CA_READ) *data = storedWord; //SedRd(lin,cra,bsa,(WORD *)data,(WORD *)&ret);
    else storedWord = *data;  //SedWr(lin,cra,bsa,*(WORD *)data,(WORD *)&ret);
    Sleep(1);
    return ret;
  }

  if ( ntelegrams > sedacInfo[lin - 1].numberTelegrams && sedacInfo[lin - 1].numberTelegrams < MAX_SEDAC_PUT )  /* get buffer  */
  {
    if ( sedacInfo[lin - 1].numberTelegrams ) /* freee previous buffer */
    {
      free ( sedacInfo[lin - 1].sedBuffer );
      sedacInfo[lin - 1].numberTelegrams = 0;
      sedacInfo[lin - 1].sedBuffer = NULL;
    }
    steps = ntelegrams > MAX_SEDAC_PUT ? MAX_SEDAC_PUT : ntelegrams;
    if ( (sedacInfo[lin - 1].sedBuffer = (void *)malloc(steps*sizeof(SedpcIn)) ) == NULL ) return -out_of_local_memory;
    sedacInfo[lin - 1].numberTelegrams = steps;
  }
  steps = (ntelegrams - 1) / MAX_SEDAC_PUT + 1;
  for (i=0; i<ntelegrams; i++)
  {
    if (access == CA_READ) data[i] = storedWord; else storedWord = data[i];
  }
  Sleep(ntelegrams/4);
  return ret;
}

// ================  read/write nwords telegram to same addesses  ===================================
int sedacHardware(WORD lin, WORD cra, WORD bsa, int access, WORD *data, int ntelegrams)
{ int i, steps, begin, ende, ret=0;

  if (ntelegrams == 1)
  {
    steps = *data;
    if (access == CA_READ)
    {
      SedRd(lin,cra,bsa,(long *)&steps,(long *)&ret);
      *data = (WORD)steps;
    }
    else SedWr(lin,cra,bsa,*(long *)data,(long *)&ret);
    return ret;
  }

  if ( ntelegrams > sedacInfo[lin - 1].numberTelegrams && sedacInfo[lin - 1].numberTelegrams < MAX_SEDAC_PUT )  /* get buffer  */
  {
    if ( sedacInfo[lin - 1].numberTelegrams ) /* freee previous buffer */
    {
      free ( sedacInfo[lin - 1].sedBuffer );
      sedacInfo[lin - 1].numberTelegrams = 0;
      sedacInfo[lin - 1].sedBuffer = NULL;
    }
    steps = ntelegrams > MAX_SEDAC_PUT ? MAX_SEDAC_PUT : ntelegrams;
    if ( (sedacInfo[lin - 1].sedBuffer = (void *)malloc(steps*sizeof(SedpcIn)) ) == NULL ) return -out_of_local_memory;
    sedacInfo[lin - 1].numberTelegrams = steps;
    cdilog("sedPC get memory, Number: %d; size: %d; line: %d\n", steps, sizeof(SedpcIn), lin);
  }
  steps = (ntelegrams - 1) / MAX_SEDAC_PUT + 1;
  for (i=0; i<steps; i++)
  {
    begin = i * MAX_SEDAC_PUT;
    ende = begin + MAX_SEDAC_PUT;
    if ( ende > ntelegrams ) ende = ntelegrams;
    ret = sedpcBlock(lin, cra, bsa, access, data, begin, ende, 1);
    if ( ret ) break;
    if ( i ) ret = sedpcBlock(lin, cra, bsa, access, data, begin- MAX_SEDAC_PUT, begin, 0);
    if ( ret ) break;
  }
  if ( ret == 0 ) ret = sedpcBlock(lin, cra, bsa, access, data, begin, ende, 0);  /* get the very last sent  */
  if ( ret )
  {
    cdilog("sedpc error! clear sedapc with line: %d\n",lin);
    clearSedac(lin);
  }

  return ret;
}

int checkThreadIDs(CdiRequestInfoBlk *pReq)
{ int lin = ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine;

  if ( sedacInfo[lin - 1].numberOfThreadIds > 1 ) return 0; // background task already registered
  sedacInfo[lin - 1].threadIDs[1] = getpid();
  if ( sedacInfo[lin - 1].threadIDs[1] == sedacInfo[lin - 1].threadIDs[0] ) return 0;
  if ( setIOpermission(lin, 1) )
  {
    cdilog("Sedpc setIOpermission(lineNumber=%d) error!\n", lin);
    return hardware_error;
  }
  sedacInfo[lin - 1].numberOfThreadIds = 2;
  return 0;
}

int doSedacRequest(CdiRequestInfoBlk *pReq, int step)
{ int access, ret, length, writeOffset, done;
  void *pOut;

  pReq->errorCode = checkThreadIDs(pReq);
  if ( pReq->errorCode )
  {
    if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
    return ret;
  }
  //printf("try to get Mutex: process pointer: 0x%x; line: %d\n",pReq, pReq->pDev->cdiLine);
  cdiMutexLock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
  //printf("get Mutex: process pointer: 0x%x; line: %d\n",pReq, pReq->pDev->cdiLine);

  if ( step == 1 ) /* first request */
    access = (pReq->accessFlag != cdiReadFlag && pReq->accessFlag != cdiReadWriteFlag && pReq->accessFlag != cdiReadReadFlag) ? CA_WRITE : CA_READ;
  else if ( step == 2 )/* second */
    access = (pReq->accessFlag == cdiReadWriteFlag || pReq->accessFlag == cdiWriteWriteFlag ) ? CA_WRITE : CA_READ;
  else
    access = pReq->accessFlag == cdiWriteReadReadFlag ? CA_READ : CA_WRITE;

  if ( access == CA_WRITE )
  {
    if ( pReq->accessFlag == cdiWriteWriteFlag || pReq->accessFlag == cdiWriteReadWriteFlag )
    {
      if ( step == 1 )
      {
        length = sizeof(short);
        pOut = (void *)pReq->pOutData;
        writeOffset = sedacSubOffsetSecond;
      }
      else
      {
        length = (pReq->accessFlag == cdiWriteReadWriteFlag && pReq->conditionFlag == anyData) ? sizeof(short) : (pReq->outLength - sizeof(short));
        pOut = (void *)((char *)pReq->pOutData + sizeof(short));
        writeOffset = sedacSubOffsetThird;
      }
    }
    else
    {
      length = (pReq->accessFlag == cdiReadWriteFlag && pReq->conditionFlag == anyData) ? sizeof(short) : pReq->outLength;
      pOut = (void *)pReq->pOutData;
      writeOffset = sedacSubOffsetSecond;
    }
  }
  else
  {
    done = 0;
    if ( pReq->accessFlag == cdiWriteReadReadFlag || pReq->accessFlag == cdiReadReadFlag )
    {
      if ( (pReq->accessFlag == cdiWriteReadReadFlag && step == 2) || (pReq->accessFlag == cdiReadReadFlag && step == 1 ) )
      {
        length = pReq->inLength < pReq->pDev->dataOuputSize ? pReq->inLength : pReq->pDev->dataOuputSize;
        pOut = (void *)pReq->pInData;
        writeOffset = sedacSubOffset;
        done = 1;
      }
      else if ( step == 3 )
      {
        if ( pReq->inLength <= pReq->pDev->dataOuputSize ) return pReq->errorCode;;
        length = pReq->inLength < (pReq->pDev->dataOuputSize + pReq->pDev->dataOuput2ndSize) ? (pReq->pDev->dataOuputSize + pReq->pDev->dataOuput2ndSize - pReq->inLength) : pReq->pDev->dataOuput2ndSize;
        pOut = (void *)((char *)pReq->pInData + pReq->pDev->dataOuputSize);
        writeOffset = sedacSubOffsetThird;
        done = 1;
      }
    }
    if ( done == 0 )
    {
        length = pReq->inLength;
        pOut = (void *)pReq->pInData;
        writeOffset = sedacSubOffset;
    }
  }

  ret = sedac((WORD)((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine, 
          (WORD)pReq->pDev->cdiAddr,
//          (WORD)(pReq->pDev->cdiSubAddr + pReq->devParameters[step == 1 ? sedacSubOffset : sedacSubOffsetSecond]), 
          (WORD)(pReq->pDev->cdiSubAddr + pReq->devParameters[writeOffset]), 
          (int)access,
          (WORD *)pOut,
          (int)length / sizeof(short) );
  if ( ret != 0 )
  {
    pReq->errorCode = ret < 0 ? -ret : sedac_error;
    if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
    if ( ret > 0 )
    {
      pReq->deviceError = ret;    // return from hardware
      pReq->pDev->numberDeviceError++;
    }
    pReq->active = finishRequest; /* finish request */
  }

  cdiMutexUnlock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);

  return pReq->errorCode;
}

