/************ sedaDrv.h for linux kernel sedac driver	*********/
#define MAX_LINES	10				// maxim number of sedac lines
#define NUMBER_INPUT_ENTRY	7		// for each line
#define NUMBER_INPUT_INT	70	// MAX_LINES*NUMBER_INPUT_ENTRY

#define MAX_NUM_TELEGRAM	15000	// maximum # for one block telegram (15000*8 = 128kBytes)
#define	DEFAULT_TEL_NUMBER	10		// pre allocate memory

// PCI 40A carrier card identification
#define	NUMBER_PCI40	4		// maximum number of pci40 cards, will de searched

// sedac ioctl command: Number of Telegram if ioctlcmd > 0 
#define	SEDAC_IOC_NUM_OPEN_PROC			0	// number of processes, which opened this sedac driver
#define	SEDAC_IOC_NUM_OPEN_LINE_PROC	-1	// number of processes, which opened this sedac line
#define	SEDAC_IOC_NUM_QUEUE_PROC		-2	// number of processes queued in this sedac line
#define	SEDAC_IOC_NUM_QUEUE_TELEGRAMS	-3	// number of Telegrams queued in this sedac line
#define	SEDAC_IOC_BUSY					-4	// sedac line busy flag
#define	SEDAC_IOC_CURRENT_PID			-5	// process id if sedac line busy
#define	SEDAC_IOC_TIMEOPUT_FLAG			-6	// timeout flag
#define	SEDAC_IOC_RUNNING_TIME			-7	// since very first open file
#define	SEDAC_IOC_TIME_COUNTER			-8	// since open file
#define	SEDAC_IOC_MMAP					-9	// mmap flag
#define	SEDAC_IOC_IOPORTMEM				-10	// get ioport address for mmap
#define	SEDAC_IOC_DEVTYPE				-11	// get device type for mmap
#define	SEDAC_IOC_PID					-12	// process own id

#define SE_READ		0x80  /* char input mode	*/
#define	SE_WRITE	0x40
#define SE_TEST	(SE_READ + SE_WRITE)
#define	SE_CP		0x20

typedef struct  _SEDAC_HEADER {
	unsigned short	NumberOfError;	// return total number of error
	unsigned short  SedacLine;		// line number start from 0....
	unsigned short	numberRequested;
	unsigned short	numberProcessed;
} SEDAC_HEADER, *PSEDAC_HEADER;

typedef struct  _SEDAC_DATA {
	unsigned short  Data;		// input or output data
	unsigned short	CommandReturn;	// (read, write or test) + cerateAdr and SubAdr (retuen with error)
	unsigned char	ExcessMode,	// input: (read, write or test.. commom pulse... ...)
					Line;		// line number
	unsigned char	SubAdr,		// SubAdr
					CaAdr;		// cerateAdr
} SEDAC_DATA, *PSEDAC_DATA;
#define SEDAC_DATA_SIZE			sizeof(SEDAC_DATA)
#define SEDAC_DATA_NUMBER_LONG	(sizeof(SEDAC_DATA)/sizeof(ULONG))

typedef struct Sedac_Desc {		// sedac descriptor
  int	DeviceType,
		IoPortAddress,
		IoPortCount,
		IrqLevel,
		IntVector,
		IpSlot,
		PciSlot;
} Sedac_Desc, *PSedac_Desc;

// support 3 kinds of sedac card
#define	IPCARD		1			// F58 ip sedac card on isa atc40 carrier board
#define	PCICARD		2			// F58 ip sedac card on pci40A carrier cards
#define	PKTRCARD	3			// pktr sedac card on an isa slot

#ifdef SEDACDRV

static int	sedaclines[MAX_LINES*NUMBER_INPUT_ENTRY] = {0,};

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0))
MODULE_AUTHOR("DESY");
MODULE_DESCRIPTION("Sedac Driver -- PktrSedac, ISA- and PCI-IPSEDAC");
MODULE_PARM(sedaclines,"1-" __MODULE_STRING(NUMBER_INPUT_INT) "i");
#endif

Sedac_Desc *sedDesc = (Sedac_Desc *)&sedaclines[0];

static int numSedac;	// number of configured sedac lines
// device config block
// for PKTRCARD and IPCARD, set hardware jumper, then type all parameters
// for PCICARD, type IntVector, IpSlot, PciSlot parameters
// DeviceType, IoPortAddress, IoPortCount, IrqLevel, IntVector, IpSlot, PciSlot;
//Sedac_Desc sedDesc[] =
//{
//	{3, 0x310,	0x10,	0x5, 0x10, 0, 0},	// device 0 ==> PKTRCARD
//	{0, 0x300,	0x10,	0xc, 0x20, 0, 0},	// device 1	==> not configured
//	{1, 0xD0000,0x4000, 0x9, 0x80, 0, 0},	// device 2 ==> IPCARD
//	{1, 0xD0000,0x4000, 0x9, 0x81, 1, 0},	// device 3 ==> IPCARD
//};
//#define SEDAC_NR_DEVICES	sizeof(sedDesc)/sizeof(Sedac_Desc)

#endif

/****************** example: how to use sedac driver ******************
***********
example 1:
***********

#include "sedacDrv.h"

#define DA_LENGTH	40000

SEDAC_DATA da[DA_LENGTH + 1];		// static header and data buffer

sedacTest()
{
  char *fname = "/dev/sedac0";		// first sedac line
  int sedacfd;						// file id
  int ret;
  unsigned int nData;				// number of sedac telegrams
  unsigned int useSelect;			// use select call flag
  PSEDAC_HEADER phe = (PSEDAC_HEADER)&da[0];
  PSEDAC_DATA	pda = &da[1];

  fd_set rfds;						// variables for using select
  struct timeval tv;
  int retval;

  sedacfd=open(fname,O_RDWR);	// open sedac line
  if ( sedacfd < 0 ) {
	fprintf(stderr, "%s: %s; error Number: %d\n", fname, strerror(errno), errno);
    exit(1);
  }

  for (i=0; i< DA_LENGTH; i++)	// file sedac data
  {
	  pda[i].Data =i*81;		// input or output data
	  pda[i].CommandReturn = 0;	// (read, write or test) + cerateAdr and SubAdr (retuen with error)
	  pda[i].ExcessMode = i%2 ? SE_READ : SE_WRITE;	// input: (read, write or test.. commom pulse... ...)
	  pda[i].Line = 0;			// line number(not used)
	  pda[i].SubAdr = 145;		// SubAdr
	  pda[i].CaAdr = 15;		// cerateAdr
  }

  if ( useSelect )
  {
	FD_ZERO(&rfds);
	FD_SET(sedacfd, &rfds);
	tv.tv_sec = 10;				// wait for 10 seconds
	tv.tv_usec = 0;
	retval = select(sedacfd+1, &rfds, NULL, NULL, &tv);
	if (retval)
	  printf("sedac line is available: %d\n",FD_ISSET(sedacfd,&rfds));
	else
	  printf("sedac line busy! \n");
  }

  nData = 1000;				// send 1000 sedac Telegrams
  ret = ioctl(sedacfd, nData, phe);
  if ( ret != nData + 1 )	// timeout during transmission
	printf("timeout! return code: %d; #sedac error: %d; #Requested: %d; #Procecced: %d\n",ret, phe->NumberOfError, phe->numberRequested, phe->numberProcessed);
  else id ( ret < 0 )
	fprintf(stderr, "ioctl error (request #Telegrams= %d): %s; return code: %d \n",nData,strerror(errno), ret);
  else
	printf("OK! return code: %d; #sedac error: %d; #Requested: %d; #Procecced: %d\n",ret, phe->NumberOfError, phe->numberRequested, phe->numberProcessed);

  close(sedacfd);
  exit(0);
}

***********
example 2:
***********
look at HERA P bpm linux sedac code, use mmap() map atc40IP or pci40AIPcard into a memory region,
using polling mode...

u16 *ATC40Address = NULL;
void initSedac(void)
{
char *fname = "/dev/sedac2";		// first sedac line
mem_fd = open(fileName, O_RDWR) );
ioctl(mem_fd,-10,BASE_ATC40_ADDR);
ioctl(mem_fd,-11,deviceType);
SIZE_ATC40_MEM = deviceType == PCI40AIPCARD ? 0x4000 : 0x300;
atc40_mem = malloc( SIZE_ATC40_MEM + (PAGE_SIZE-1);
if(( unsigned long)atc40_mem % PAGE_SIZE )
        atc40_mem += PAGE_SIZE - ((unsigned long)atc40_mem % PAGE_SIZE );
atc40_mem = (caddr_t)mmap(
        (void *)atc40_mem,
        (size_t)SIZE_ATC40_MEM,
        PROT_READ|PROT_WRITE,
        MAP_SHARED|MAP_FIXED,
        (int)mem_fd,
        (off_t)BASE_ATC40_ADDR );
ATC40Address = (u16 *)atc40_mem;

if (devType == ATC40IP )
   line[i] = (SEDAC_LINE *) (ATC40Address + (i/2)*0x100 + (i%2)*0x04); // for atc40IP card
else
   line[i] = (SEDAC_LINE *) (ATC40Address + 0x800 + (i/2)*0x1000 + (i%2)*0x04); // for PCI40IP card
    
line[i]->csr = 0;		// init sedac ip moudule
line[i]->igr = 0;
}

// ============  exit routine  ===================================
int exitSedac(void)
{
  if ( munmap( (void *)ATC40Address,	(size_t)SIZE_ATC40_MEM ) != 0 )
	  printf("munmap failed, errno: %d: %s\n", errno, strerror(errno));
  close(mem_fd);
  return 0;
}


****************** example: how to use sedac driver ******************/
