// SedCnt.c 240703  Stadtmueller    mst3    230703
//      SedPc & Sed1Pc & Sed2Pc from Version 98
//
//      SedCac, SedCnt, SedErc
// Read and clear ErrorCounter Routines
//

#include <stdlib.h>
#include "SedPc.h"

//-------+---------+---------+---------+---------+---------+---------+---------+

// SedCac   Clear All Error and FunctionCounter
//				!!! This Function takes around 50 ms !!!
//			Sed0Pc don't have this Counters, you get an Error back.
//
//  call: SedCac (lLn, lplErr);
//  para:   lLn     = logical LineNumber
//          lplErr  = 0  = no errors
//            bit 13 = illegal SEDPC type
//            bit 12 = no such LineNumber
//            bit  7 = ParameterError: one parameter incorrect
//            bit  6 = Response/Busy not set; data not taken by SEDPC
//            bit  5 = Bytecount mismatch
//  subp: sedsel, sedwr0, sedwr4, sedrd1, sedrd2

void SedCac ( LONG lLn, LONG *lplErr )

{
	LONG static	dat;

	if ( *lplErr = sedsel(lLn) )    // set device-addresses to externals
		goto exit;

	*lplErr  = sedwr4 (0x2700, 0);  // Clear All Counter
	*lplErr |= sedwr0 (LASTBY);
	if ( *lplErr )              // error set ?
		goto exit;

	*lplErr = sedrd1(&dat);     // get status, error ?
	if ( dat != 0x20 )          // status-error ?
		*lplErr |= 0x20;
	*lplErr |= sedrd2 (&dat);   // get DummyData
	*lplErr |= sedrd1 (&dat);   // get LastByteMarker
	if ( dat != LASTBY )        // LastByteMarker not set ?
		*lplErr |= 0x8;

exit:
	return;

}		// end of SedCac
//-------+---------+---------+---------+---------+---------+---------+---------+

// SedCnt   Get LongWord Counter
//			  Bit 30..0 = Count, Bit 31 = Overflow
//			Sed0Pc don't have this Counters, you get an Error back.
//
//  call: SedCnt (lLn, lplArr, lplErr);
//  para:   lLn		= logical LineNumber
//          lplArr	= pointer to array with LongCounter
//			  lArr[00]	= User ReadTel
//			  lArr[01]	= User WriteTel
//			  lArr[02]	= TestTel
//			  lArr[03]	= SpecialFunctionTel (Look for Cr or Sa)
//			  lArr[04]	= Cp's send, programmed or Hardware
//			  lArr[05]	= programmed changes of Delay
//			  lArr[06]	= programmed changes of TestTel
//			  lArr[07]	= n.a.
//			  lArr[08]	= ResponseErrors	SedacStatus = 0x7 | 0xF
//			  lArr[09]	= SedacErrors		SedacStatus = > 0 but < 7
//			  lArr[10]	= MysteriousErrors	SedacStatus = 8
//			  lArr[11]	= InternalErrors, MicErr will light
//			  lArr[12]	= RspErr BitField, one Bit for each Crate
//			  lArr[13]	= SedErr BitField
//			  lArr[14]	= MysErr BitField
//			  lArr[15]	= n.a.
//          lplErr	= ErrorReturnCode
//          *lplErr	= 0 = no error
//            bit 13 = illegal SEDPC type
//            bit 12 = no such LineNumber
//            bit  7 = ParameterError: one parameter incorrect
//            bit  6 = Response/Busy not set; data not taken by SEDPC
//            bit  5 = Bytecount mismatch
//            bit  4 = Timeout: no response from SEDPC
//            bit  3 = LastByte marker not set or illegal function
//  subp: sedsel, sedwr0, sedwr4, sedrd1, sedrd2

void SedCnt ( LONG lLn, LONG *lplArr, LONG *lplErr )

{
	static LONG	dat, bc, i;
	static WORD	*lpwArr;

	bc = 0;
	if ( *lplErr = sedsel(lLn) )	// set device-addresses to externals
		goto exit;

	*lplErr  = sedwr4(0x2800, 0);	// special function to SedPc, error ?
	*lplErr |= sedwr0(LASTBY);
	if ( *lplErr )
		goto exit;

	lpwArr = (WORD *) lplArr;
	*lplErr = sedrd1(&dat);     // get status, error ?
	if ( dat != 0x30 )          // status-error ?
		*lplErr |= 0x20;
	*lplErr |= sedrd2 (&bc);	// get ByteCount
	if ( bc != 0x40 ) {			// wrong ByteCount ?
		*lplErr = 0x20;
		goto exit;
	}
	for ( i = 0; i < bc / 2; i += 2 ) {	// get half LongCounter
		*lplErr |= sedrd2 (&dat);
		*(lpwArr+i+1) = dat;
		*lplErr |= sedrd2 (&dat);
		*(lpwArr+i) = dat;
	}
	*lplErr |= sedrd1 (&dat);   // get LastByteMarker
	if ( dat != LASTBY )        // LastByteMarker not set ?
		*lplErr |= 0x8;


exit:
	return;

}		// end SedCnt
//-------+---------+---------+---------+---------+---------+---------+---------+

// SedErc   Get ErrorCounter
//			  Bit 14..0 = Count, Bit 15 = Overflow
//			Sed0Pc don't have this Counters, you get an Error back.
//
//  call: SedErc (lLn, lCn, lClass, lplArr, lplErr);
//  para:   lLn		= logical LineNumber
//			lCn		= CrateNumber
//			lClass	= Class of ErrorCounter
//				0	  = ResponseErrors
//				1	  = SedacErrors
//				2	  = MysteriousErrors
//          lplArr	= pointer to array, one Counter per SedacAddress
//          lplErr	= ErrorReturnCode
//          *lplErr	= 0 = no error
//            bit 13 = illegal SEDPC type
//            bit 12 = no such LineNumber
//            bit  7 = ParameterError: one parameter incorrect
//            bit  6 = Response/Busy not set; data not taken by SEDPC
//            bit  5 = Bytecount mismatch
//            bit  4 = Timeout: no response from SEDPC
//            bit  3 = LastByte marker not set or illegal function
//  subp: sedsel, sedwr0, sedwr4, sedrd1, sedrd2

void SedErc ( LONG lLn, LONG lCn, LONG lClass, LONG *lplArr, LONG *lplErr )
{

	static LONG	dat, bc, i;

	bc = 0;
	if ( *lplErr = sedsel(lLn) )	// set device-addresses to externals
		goto exit;

	switch ( lClass ) {
	case 0:							// ResponseErrors
		*lplErr |= sedwr1(0x2A);	// special function to SedPc
		*lplErr |= sedwr1(lCn);		// CrateNumber to SedPc
		*lplErr |= sedwr2(0);		// dummy Data
		break;
	case 1:							// SedacErrors
		*lplErr |= sedwr1(0x2C);	// special function to SedPc
		*lplErr |= sedwr1(lCn);		// CrateNumber to SedPc
		*lplErr |= sedwr2(0);		// dummy Data
		break;
	case 2:							// MysteriousErrors
		*lplErr |= sedwr1(0x2E);	// special function to SedPc
		*lplErr |= sedwr1(lCn);		// CrateNumber to SedPc
		*lplErr |= sedwr2(0);		// dummy Data
		break;
	default:						// illegal Function
		*lplErr	= 0x80;
		goto exit;
	}
	*lplErr |= sedwr0(LASTBY);
	if ( *lplErr )
		goto exit;

	*lplErr = sedrd1(&dat);     // get status, error ?
	if ( dat != 0x30 )          // status-error ?
		*lplErr |= 0x20;
	*lplErr |= sedrd2 (&bc);	// get ByteCount
	if ( bc != 0x200 ) {		// wrong ByteCount ?
		*lplErr = 0x20;
		goto exit;
	}
	for ( i = 0; i < bc / 2; i += 1 ) {	// get Counter
		*lplErr |= sedrd2 (&dat);
		*(lplArr+i) = dat;
	}
	*lplErr |= sedrd1 (&dat);   // get LastByteMarker
	if ( dat != LASTBY )        // LastByteMarker not set ?
		*lplErr |= 0x8;


exit:
	return;

}		// end SedErc
//-------+---------+---------+---------+---------+---------+---------+---------+

//end of SedCnt.c
