// SedCrw.c 230703  Stadtmueller    mst3    160703
//      SedPc & Sed1Pc & Sed2Pc from Version 98
//
//      SedClr, SedRd, SedWr
// clear interface, read and write a telegram
//

#include <stdlib.h>
#include "SedPc.h"

//-------+---------+---------+---------+---------+---------+---------+---------+

// SedClr   perform Init function, clear fifo and buffer. wait for end of clear,
//            read telegram delay-time and set it to external variables.
//
//  call: SedClr (lLn, lplDly, lplErr);
//  para:   lLn     = logical line-number
//          lplDly  = pointer to delay-time
//          *lplErr = 0 = no error
//            bit 13 = illegal SEDPC type
//  subp: sedsel, SedDly
//  used: externals

void SedClr ( LONG lLn, LONG *lplDly, LONG *lplErr )

{
	static LONG    dat, i, n;

	if ( *lplErr = sedsel(lLn) )		// set device-addresses to externals
		goto exit;

	switch ( glModAdr[glSelLin] & 0xFF00 ) {

	case SEDPC :
		_outp (glCtrReg, CLR);			// clear fifo
		_outp (glCtrReg, OUTMOD);		// reset clear

		usleep( 50 );					// wait 50 us

		_outp (glDatReg, glModAdr[glSelLin]);	// test status for Sed1Pc answer
		_outp (glCtrReg, SELECT);
		dat = _inp(glStaReg) & 0xF8;
		_outp (glCtrReg, OUTMOD);		// set output
		if ( dat == SSEL )				// status-lines wrong ?
			goto err13;

		break;

	case SED1PC :
		_outp (glDatReg, 0);
		_outp (glCtrReg, SELECT);		// deselect all SEDPC
		_outp (glCtrReg, IDLE);
		_outp (glDatReg, glModAdr[glSelLin]);
		_outp (glCtrReg, INIT);			// INIT to one SEDPC
		_outp (glCtrReg, IDLE);

		usleep( 50 );					// wait 50 us

		_outp (glDatReg, glModAdr[glSelLin]);	// test status for Sed1Pc answer
		_outp (glCtrReg, SELECT);
		dat = _inp(glStaReg) & 0xF8;
		_outp (glCtrReg, IDLE);
		if ( dat != SSEL )				// status-lines wrong ?
			goto err13;

		break;

	default :
		goto err13;						// error, illegal SEDPC type
	}

	i = 0;

sedclr1:

	*lplDly = -1;						// get delay time from SEDPC
	SedDly (lLn, lplDly, lplErr);

	if ( i++ < 3 ) {					// if error try it two times more
		if ( *lplErr )
			goto sedclr1;
	}

exit:
	return;

err13:
	*lplErr = 0x2000;					// error wrong SEDPC type
	goto exit;

}		// end of SedClr
//-------+---------+---------+---------+---------+---------+---------+---------+

// SedRd    perform a SEDAC read telegram
//
//  call: SedRd (lLn, lCn, lSa, lplDat, lplErr);
//  para:   lLn    = logical line-number
//          lCn    = crate number, 0..31 valid, +32 if CP
//          lSa    = subaddree ss, 0-255 valid
//          lplDat = pointer to data, 16 bit to sedac
//          *lplErr = 0 = no error
//            bit 13 = illegal SEDPC type
//            bit 12 = no such line-number
//            bit  7 = ParameterError: one parameter incorrect
//            bit  6 = Response/Busy not set; data not taken by SEDPC
//            bit  5 = Bytecount mismatch
//            bit  4 = Timeout: no response from SEDPC
//            bit  3 = LastByte marker not set or illegal function
//            bit  2..0 = Code of Sedac Error
//  subp: sedsel, sedwr0, sedwr4, sedrd1, sedrd2

void SedRd  ( LONG lLn, LONG lCn, LONG lSa, LONG *lplDat, LONG *lplErr )

{
	static LONG addr, dat;

	if ( *lplErr = sedsel(lLn) )    // set device-addresses to externals
		goto exit;

	if ( lCn > 63 ) {               // check parameter cn and sa
		if ( lSa > 255 ) {
			*lplErr = 0x80;         // set error-code
			goto exit;
	}  }
	addr = (lCn | 0x80) << 8 | lSa; // addr and ReadBit

	if ( *lplErr = sedwr4 (addr, 0) ) // cn and sa to SedPc, error ?
		goto exit;
	if ( *lplErr = sedwr0 (LASTBY) )  // LastByteMarker to SedPc, error ?
		goto exit;

	*lplErr  = sedrd1 (&dat);       // get status-byte
	*lplErr |= dat & 0xF;           // set sedac-errors to err
	if ( (dat & 0xF0) != 0xA0 )     // read and bc = 2 not set ?
		*lplErr |= 0x20;
	*lplErr |= sedrd2 (lplDat);     // get data
	*lplErr |= sedrd1 (&dat);       // get LastByteMarker
	if ( dat != LASTBY )            // LastByteMarker not set ?
		*lplErr |= 0x8;

exit:
	return;

}		// end SedRd
//-------+---------+---------+---------+---------+---------+---------+---------+

// SedWr    perform a SEDAC write telegram
//
//  call: SedWr (lLn, lCn, lSa, wDat, lplErr);
//  para:   lLn    = logical line-number
//          lCn    = crate number, 0..31 valid, +32 if CP
//          lSa    = subaddress, 0-255 valid
//          wDat   = data, 16 bit to sedac
//          *lplErr = 0 = no error
//            bit 13 = illegal SEDPC type
//            bit 12 = no such line-number
//            bit  7 = ParameterError: one parameter incorrect
//            bit  6 = Response/Busy not set; data not taken by SEDPC
//            bit  5 = Bytecount mismatch
//            bit  4 = Timeout: no response from SEDPC
//            bit  3 = LastByte marker not set or illegal function
//            bit  2..0 = Code of Sedac Error
//  subp: sedsel, sedwr0, sedwr4, sedrd1, sedrd2

void SedWr  ( LONG lLn, LONG lCn, LONG lSa, LONG wDat, LONG *lplErr )

{
	LONG static addr, dat;

	if ( *lplErr = sedsel(lLn) )    // set device-addresses to externals
		goto exit;

	if ( lCn > 63 ) {               // check parameter cn and sa
		if ( lSa > 255 ) {
			*lplErr = 0x80;         // set error-code
			goto exit;
	}  }
	addr = (lCn | 0x40) << 8 | lSa; // addr and WriteBit

	if ( *lplErr = sedwr4(addr, wDat) ) // cn, sa and data to SedPc, error ?
		goto exit;
	if ( *lplErr = sedwr0 (LASTBY) )    // LastByteMarker to SedPc, error ?
		goto exit;

	*lplErr  = sedrd1 (&dat);       // get status-byte
	*lplErr |= dat & 0xF;           // set sedac-errors to err
	if ( (dat & 0xF0) != 0x60 )     // write and bc = 2 not set ?
		*lplErr |= 0x20;
	*lplErr |= sedrd2 (&dat);       // get DummyData
	*lplErr |= sedrd1 (&dat);       // get LastByteMarker
	if ( dat != LASTBY )            // LastByteMarker not set ?
		*lplErr |= 0x8;

exit:
	return;

}		// end SedWr
//-------+---------+---------+---------+---------+---------+---------+---------+

// end of file SedCrw.c

