// SedLow.c 240703  Stadtmueller    mst3    160703
//      SedPc & Sed1Pc & Sed2Pc from Version 98
//
//  sedlpt, sedsel + sed0wt + sed1wt +
//  sedrd1 + sedrd2 + sedwr0 + sedwr1 + sedwr2 + sedwr4
//  dstrtu, hstrtu
// LowLevel Routines to access the ParallelPort and StringFunctions
//

#include "SedPc.h"
#include <time.h>

//-------+---------+---------+---------+---------+---------+---------+---------+

// sedlpt   for: SedPc & Sed1Pc Version 95A & 97A & 98B
//          calculate a number of for-loops per tick (per msec)
//            the for-loop has 2 _outp(..) statements inside
//
//  call: sedlpt;
//  used: external glTickLoop

void sedlpt ( void )

{
	static clock_t	t1, t2;
	static LONG		i, j, n;

	_outp (glCtrReg, IDLE);
	t1 = clock ();

	do {                        // syncronise to start of tick-counter
		t2 = clock ();
	} while ( t2 == t1 );
	t1 = t2 + 10 * TICKINC;     // time = 10 tick-counts = 550 ms
	j = 0;
	n = 300;
	do {                        // you need n*j for-llops for 10 ticks
		for ( i = 0; i < n; i++ ) {
			_outp ( glDatReg, 2 );  // data-bit 1 will be changed
			_outp ( glDatReg, 0 );
		}
		j++;
		t2 = clock ();
	} while ( t2 < t1 );
	glTickLoop = n * j / (10 * TICKINC);

	return;

}		// end of sedlpt
//-------+---------+---------+---------+---------+---------+---------+---------+

// sedsel   for: SedPc & Sed1Pc Version 95A & 97A & 98B
//          setup the device-addresses corresponding with the given
//            logical line-number to the externals, give access to that port
//            and select that module if it should be a Sed1Pc.
//
//  call: err = sedsel (lLn);
//  para: lLn = logical line-number and module-number
//        err = 0 = no error
//            bit 12 = no such line-number
//            bit  4 = Timeout: no response from SEDPC
//  used: externals

LONG sedsel ( LONG lLn )

{
	static LONG i, ret;

	ret = 0;
	for ( i = 0; i < glNoLin; i++ ) {   // search glLogLin-array for that line-number
		if ( lLn == glLogLin[i] ) {     // equal ?
			glSelLin = i;               // set index of last selected line
			glDatReg = glDevAdr[i];     // setup the device-addresses
			glStaReg = glDatReg + 1;
			glCtrReg = glDatReg + 2;
			if ( (glModAdr[i] & 0xFF00) == SED1PC ) {
				_outp(glDatReg, glModAdr[i]);   // set module-number to data-byte
				_outp(glCtrReg, SELECT);        // select module
				if ( (_inp(glStaReg) & 0xF8) != SSEL )
				{
					ret = 0x10;         // error, no response or not OnLine
				}
				_outp(glCtrReg, IDLE);  // leave it in IDLE state
			}
			goto exit;
		}
	}
	ret = (0x1000);             // line-number not found in array
exit:
	return ( ret );

}		// end of sedsel
//-------+---------+---------+---------+---------+---------+---------+---------+

// sed0wt   for: SedPc Version 95A & 97A & 98B
//          Wait a Timeout-time until Busy is set
//             in case of error: LPTn control-register will be set to OUTMOD
//
//  call: err   = sed0wt (&stat);
//  para: &stat = status-byte of LPTn
//        err   = 0 = no error
//            bit  4 = Timeout: no response from SEDPC
//  used: extern LONG   glDatReg, ...

LONG sed0wt ( LONG *stat )

{
	static clock_t	t1, t2;

	*stat = 0;
	t1 = clock ();

	do {
		t2 = clock ();
		*stat = _inp (glStaReg);		// get status-register
		if ( (*stat & BUSY) )			// BUSY set ? >> data present
			goto exit;
	} while ( t2-t1 < TICKINC+5 );		// smallest unit

	_outp (glCtrReg, OUTMOD);			// reset input-strobe
	return (0x10);						// timeout-error

exit:
	return (0);							// ok, BUSY set before timeout

}		// end of sed0wt
//-------+---------+---------+---------+---------+---------+---------+---------+

// sed1wt   for: Sed1Pc Version 95A & 97A & 98B
//          Wait a timeout time until OutFifo is ready (not empty)
//             LPTn control-register will be set to IDLE at end of routine
//
//  call: err = sed1wt;
//  para: err = 0 = no error
//            bit  4 = Timeout: no response from SEDPC
//  used: extern LONG   glDatReg, ...

LONG sed1wt ( void )

{
	static clock_t	t1, t2;
	static LONG		sta;

	t1 = clock ();
	_outp (glCtrReg, RDSTA);

	do {
		t2 = clock ();
		sta = _inp (glStaReg);			// get status-register
		if ( !(sta & OFEF) )			// OutFifo not empty ?
			goto exit;
	} while ( t2-t1 < TICKINC+5 );		// smallest unit

	_outp (glCtrReg, IDLE);				// control to IDLE
	return (0x10);						// timeout-error

exit:
	_outp (glCtrReg, IDLE);				// control to IDLE
	return (0);							// ok, /Response set before timeout

}		// end of sed1wt
//-------+---------+---------+---------+---------+---------+---------+---------+

// sedrdn   for: SedPc & Sed1Pc Version 95A & 97A & 98B
//          Read n Bytes of data from SEDPC
//
//  call: err = sedrd1 (&dat);
//  para:   &dat  = one byte of data, upper byte = 0
//
//  call: err = sedrd2 (&dat);
//  para:   &dat  = two bytes of data, 1. byte = upper byte
//          err = 0 = no error
//            bit  4 = Timeout: no response from SEDPC
//            bit 13 = illegal SEDPC type
//
//  used: externals
//  subp: sed0wt, sed1wt

LONG sedrd1 ( LONG *dat )

{
	LONG static sta, err;

	err = 0x10;

	switch ( glModAdr[glSelLin] & 0xFF00) {

	case SEDPC :
		_outp (glCtrReg, INSTB);    // set input and strobe
		sta = _inp (glStaReg);      // get status-register
		if ( !(sta & BUSY) )        // BUSY not set ?
			if ( err = sed0wt(&sta) ) // wait for BUSY, timeout reached ?
				goto exit;          // yes, error
		*dat = (sta >> 3) & 0xF;    // get lower nibbel
		_outp (glCtrReg, INMOD);    // reset strobe

		_outp (glCtrReg, INSTB);    // set input and strobe
		sta = _inp (glStaReg);      // get status-register
		if ( !(sta & BUSY) )        // BUSY not set ?
			if ( err = sed0wt(&sta) ) // wait for BUSY, timeout reached ?
				goto exit;          // yes, error
		*dat |= (sta << 1) & 0xF0;  // get upper nibbel, that's the byte
		_outp (glCtrReg, OUTMOD);   // reset input-strobe
		err = 0;
		break;

	case SED1PC :
		_outp (glCtrReg, RDSTA);
		sta = _inp (glStaReg);       // get FifoFlags
		if ( (sta & OFEF) )          // OutFifo empty ?
			if ( err = sed1wt() )    // wait until OutFifo not empty
				goto ep1;            // error, TimeOut reached
		_outp (glCtrReg, IDLE);
		_outp (glCtrReg, RDUPP);
		sta = _inp (glStaReg);       // get data
		if ( !(sta & SRSP) )         // /Response not set ?
			goto ep1;                // yes, error
		*dat = (sta << 1) & 0xF0;    // get upper nibbel

		_outp (glCtrReg, RDLOW);
		sta = _inp (glStaReg);       // get data
		if ( !(sta & SRSP) )         // /Response not set ?
			goto ep1;                // yes, error
		*dat |= (sta >> 3) & 0xF;    // get lower nibbel
		err = 0;

ep1:
		_outp (glCtrReg, IDLE);
		break;

	default :
		err = 0x2000;           // error, illegal SEDPC type
	}
exit:
return (err);

}		// end of sedrd1
//-------+---------+---------+---------+---------+---------+---------+---------+

LONG sedrd2 ( LONG *dat )

{
	LONG static sta, err;

	err = 0x10;

	switch ( glModAdr[glSelLin] & 0xFF00) {

	case SEDPC :
		_outp (glCtrReg, INSTB);    // set input and strobe
		sta = _inp (glStaReg);      // get status-register
		if ( !(sta & BUSY) )        // BUSY not set ?
			if ( err = sed0wt(&sta) ) // wait for BUSY, timeout reached ?
				goto exit;          // yes, error
		*dat = (sta >> 3) & 0xF;    // get lower nibbel
		_outp (glCtrReg, INMOD);    // reset strobe

		_outp (glCtrReg, INSTB);    // set input and strobe
		sta = _inp (glStaReg);      // get status-register
		if ( !(sta & BUSY) )        // BUSY not set ?
			if ( err = sed0wt(&sta) ) // wait for BUSY, timeout reached ?
				goto exit;          // yes, error
		*dat |= (sta << 1) & 0xF0;  // get upper nibbel, that's the byte
		_outp (glCtrReg, INMOD);    // reset strobe
		*dat <<= 8;

		_outp (glCtrReg, INSTB);    // set input and strobe
		sta = _inp (glStaReg);      // get status-register
		if ( !(sta & BUSY) )        // BUSY not set ?
			if ( err = sed0wt(&sta) ) // wait for BUSY, timeout reached ?
				goto exit;          // yes, error
		*dat |= (sta >> 3) & 0xF;   // get lower nibbel
		_outp (glCtrReg, INMOD);    // reset strobe

		_outp (glCtrReg, INSTB);    // set input and strobe
		sta = _inp (glStaReg);      // get status-register
		if ( !(sta & BUSY) )        // BUSY not set ?
			if ( err = sed0wt(&sta) ) // wait for BUSY, timeout reached ?
				goto exit;          // yes, error
		*dat |= (sta << 1) & 0xF0;  // get upper nibbel, that's the byte
		_outp (glCtrReg, OUTMOD);   // reset input-strobe
		err = 0;
		break;

	case SED1PC :
		_outp (glCtrReg, RDSTA);
		sta = _inp (glStaReg);      // get FifoFlags
		if ( (sta & OFEF) )         // OutFifo empty ?
			if ( err = sed1wt() )   // wait until OutFifo not empty
				goto ep1;           // error, TimeOut reached
		_outp (glCtrReg, IDLE);
		_outp (glCtrReg, RDUPP);
		sta = _inp (glStaReg);      // get data
		if ( !(sta & SRSP) )        // /Response not set ?
			goto ep1;               // yes, error
		*dat = (sta << 1) & 0xF0;   // get upper nibbel

		_outp (glCtrReg, RDLOW);
		sta = _inp (glStaReg);      // get data
		if ( !(sta & SRSP) )        // /Response not set ?
			goto ep1;               // yes, error
		*dat |= (sta >> 3) & 0xF;   // get lower nibbel
		_outp (glCtrReg, IDLE);

		*dat <<= 8;

		_outp (glCtrReg, RDSTA);
		sta = _inp (glStaReg);      // get FifoFlags
		if ( (sta & OFEF) )         // OutFifo empty ?
			if ( err = sed1wt() )   // wait until OutFifo not empty
				goto ep1;           // error, TimeOut reached
		_outp (glCtrReg, IDLE);
		_outp (glCtrReg, RDUPP);
		sta = _inp (glStaReg);      // get data
		if ( !(sta & SRSP) )        // /Response not set ?
			goto ep1;               // yes, error
		*dat |= (sta << 1) & 0xF0;  // get upper nibbel

		_outp (glCtrReg, RDLOW);
		sta = _inp (glStaReg);      // get data
		if ( !(sta & SRSP) )        // /Response not set ?
			goto ep1;               // yes, error
		*dat |= (sta >> 3) & 0xF;   // get lower nibbel
		err = 0;

ep1:
		_outp (glCtrReg, IDLE);
		break;

	default :
		err = 0x2000;           // error, illegal SEDPC type
	}
exit:
return (err);

}		// end of sedrd2
//-------+---------+---------+---------+---------+---------+---------+---------+

// sedwrn   for: SedPc & Sed1Pc Version 95A & 97A & 98B
//          Write n Bytes to Sed1Pc
//
//  call: err = sedwr0 (long1);
//  para:   long1 = lower byte is data with PETF
//
//  call: err = sedwr1 (long1);
//  para:   long1 = lower byte is data without PETF
//
//  call: err = sedwr2 (long1);
//  para:   long1 = one LONG of data, upper byte = 1. byte
//
//  call: err = sedwr4 (long1, long2);
//  para:   long1 = one LONG of data, upper byte = 1. byte
//          long2 = one LONG of data, lower byte = 4. byte
//          err = 0     = no error
//            bit  6 = Response/Busy not set; data not taken by SEDPC
//            bit 13 = illegal SEDPC type
//
//  used:   externals

LONG sedwr0 ( LONG long1 )

{
	LONG static err;

	err = 0;

	switch ( glModAdr[glSelLin] & 0xFF00) {

	case SEDPC :
		_outp (glDatReg, long1);        // data to port
		_outp (glCtrReg, OUTEND);       // strobe data out with end-bit
		if ( ! (_inp (glStaReg) & BUSY) ) // busy not set : error
			err = 0x40;
		_outp (glCtrReg, OUTMOD);       // reset strobe-bit
		break;

	case SED1PC :
		_outp (glDatReg, long1);        // data to port
		_outp (glCtrReg, WRILA);        // strobe data out with end-mark
		if ( ! (_inp (glStaReg) & SRSP) ) // response not set : error
		err = 0x40;
		_outp (glCtrReg, IDLE);
		break;

	default :
		err = 0x2000;           // error, illegal SEDPC type
	}

	return (err);

}		// end of sedwr0
//-------+---------+---------+---------+---------+---------+---------+---------+

LONG sedwr1 ( LONG long1 )

{
	LONG static err;

	err = 0;

	switch ( glModAdr[glSelLin] & 0xFF00) {

	case SEDPC :
		_outp (glDatReg, long1);        // data to port
		_outp (glCtrReg, OUTSTB);       // strobe data out
		if ( ! (_inp (glStaReg) & BUSY) ) // busy not set : error
			err = 0x40;
		_outp (glCtrReg, OUTMOD);       // reset strobe-bit
		break;

	case SED1PC :
		_outp (glDatReg, long1);        // data to port
		_outp (glCtrReg, WRIDA);        // strobe data out
		if ( ! (_inp (glStaReg) & SRSP) ) // response not set : error
			err = 0x40;
		_outp (glCtrReg, IDLE);
		break;

	default :
		err = 0x2000;           // error, illegal SEDPC type
	}

	return (err);

}		// end of sedwr1
//-------+---------+---------+---------+---------+---------+---------+---------+

LONG sedwr2 ( LONG long1 )

{
	LONG static err;

	err = 0;

	switch ( glModAdr[glSelLin] & 0xFF00) {

	case SEDPC :
		_outp (glDatReg, long1 >> 8);   // msb-data to port
		_outp (glCtrReg, OUTSTB);       // strobe data out
		if ( ! (_inp (glStaReg) & BUSY) ) // busy not set : error
		err = 0x40;
		_outp (glCtrReg, OUTMOD);       // reset strobe-bit
		_outp (glDatReg, long1);        // lsb-data to port
		_outp (glCtrReg, OUTSTB);       // strobe data out
		if ( ! (_inp (glStaReg) & BUSY) ) // busy not set : error
		err = 0x40;
		_outp (glCtrReg, OUTMOD);       // reset strobe-bit
		break;

	case SED1PC :
		_outp (glDatReg, long1 >> 8);   // msb-data to port
		_outp (glCtrReg, WRIDA);        // strobe data out
		if ( ! (_inp (glStaReg) & SRSP) ) // response not set : error
		err = 0x40;
		_outp (glCtrReg, IDLE);
		_outp (glDatReg, long1);        // lsb-data to port
		_outp (glCtrReg, WRIDA);        // strobe data out
		if ( ! (_inp (glStaReg) & SRSP) ) // response not set : error
		err = 0x40;
		_outp (glCtrReg, IDLE);
		break;

	default :
		err = 0x2000;           // error, illegal SEDPC type
	}

	return (err);

}		// end of sedwr2
//-------+---------+---------+---------+---------+---------+---------+---------+

LONG sedwr4 ( LONG long1, LONG long2 )

{
	LONG static err;

	err = 0;

	switch ( glModAdr[glSelLin] & 0xFF00) {

	case SEDPC :
		_outp (glDatReg, long1 >> 8);   // msb-data to port
		_outp (glCtrReg, OUTSTB);       // strobe data out
		if ( ! (_inp (glStaReg) & BUSY) ) // busy not set : error
		err = 0x40;
		_outp (glCtrReg, OUTMOD);       // reset strobe-bit
		_outp (glDatReg, long1);        // lsb-data to port
		_outp (glCtrReg, OUTSTB);       // strobe data out
		if ( ! (_inp (glStaReg) & BUSY) ) // busy not set : error
		err = 0x40;
		_outp (glCtrReg, OUTMOD);       // reset strobe-bit
		_outp (glDatReg, long2 >> 8);   // msb-data to port
		_outp (glCtrReg, OUTSTB);       // strobe data out
		if ( ! (_inp (glStaReg) & BUSY) ) // busy not set : error
		err = 0x40;
		_outp (glCtrReg, OUTMOD);       // reset strobe-bit
		_outp (glDatReg, long2);        // lsb-data to port
		_outp (glCtrReg, OUTSTB);       // strobe data out
		if ( ! (_inp (glStaReg) & BUSY) ) // busy not set : error
		err = 0x40;
		_outp (glCtrReg, OUTMOD);       // reset strobe-bit
		break;

	case SED1PC :
		_outp (glDatReg, long1 >> 8);   // msb-data to port
		_outp (glCtrReg, WRIDA);        // strobe data out
		if ( ! (_inp (glStaReg) & SRSP) ) // response not set : error
		err = 0x40;
		_outp (glCtrReg, IDLE);
		_outp (glDatReg, long1);        // lsb-data to port
		_outp (glCtrReg, WRIDA);        // strobe data out
		if ( ! (_inp (glStaReg) & SRSP) ) // response not set : error
		err = 0x40;
		_outp (glCtrReg, IDLE);
		_outp (glDatReg, long2 >> 8);   // msb-data to port
		_outp (glCtrReg, WRIDA);        // strobe data out
		if ( ! (_inp (glStaReg) & SRSP) ) // response not set : error
		err = 0x40;
		_outp (glCtrReg, IDLE);
		_outp (glDatReg, long2);        // lsb-data to port
		_outp (glCtrReg, WRIDA);        // strobe data out
		if ( ! (_inp (glStaReg) & SRSP) ) // response not set : error
		err = 0x40;
		_outp (glCtrReg, IDLE);
		break;

	default :
		err = 0x2000;           // error, illegal SEDPC type
	}

	return (err);

}		// end of sedwr4

//-------+---------+---------+---------+---------+---------+---------+---------+

// hstrtu	convert a Hex-STRing To Unsigned long
//
//	call:	err		 =	hstrtu (&hstr[0], &dat)
//	para:	&hstr[0] =	far-pointer to first element of hex-string
//			&dat     =	pointer to unsigned data
//			dat 	 =	unsigned long
//			err = 0  =	no error
//			  bit 7 = parameter error, illegal character in hex-string

LONG hstrtu (CHAR *hstr, LONG *pdat)

{
	CHAR    c;
	LONG    hu, err, i;

	hu = err = i = 0;
	while ( c = *(hstr+i) ) {		// loop until NUL reached
		hu <<= 4;					// shift 4 bit
		if ( c >= '0' && c <= '9' )
			hu |= c & 0xF;			// value 0 .. 9
		else {
			c &= 0xDf;				// make upper-case
			if ( c >= 'A' && c<= 'F' )
				hu |= (c & 7) + 9;	// value A .. F
			else {
				err = 0x80;			// illegal character
				break;
		}  }
		i++;
	}
	*pdat = hu;

	return (err);
}		// end of hstrtu
//-------+---------+---------+---------+---------+---------+---------+---------+

// dstrtu	convert a Decimal-STRing To Unsigned long
//
//	call:	err		 =	dstrtu (&dstr[0], &dat)
//	para:	&dstr[0] =	far-pointer to first element of dec-string
//			&dat     =	pointer to unsigned data
//			dat 	 =	unsigned long
//			err = 0  =	no error
//			  bit 7 = parameter error, illegal character in dec-string

LONG dstrtu (CHAR *dstr, LONG *pdat)

{
	CHAR    c;
	LONG    du, err, i;

	du = err = i = 0;
	while ( c = *(dstr+i) ) {	// loop until NUL reached
		du *= 10;
		if ( c >= '0' && c <= '9' )
			du += c & 0xF;
		else {
			err = 0x80;			// illegal character
			break;
		}
		i++;
	}
	*pdat = du;

	return (err);
}		// end of dstrtu
//-------+---------+---------+---------+---------+---------+---------+---------+

// end of file SedLow.c

