// SedSer.c 230703  Stadtmueller    mst3    210703
//      SedPc & Sed1Pc & Sed2Pc from Version 98
//
//      SedDly, SedLcs, SedVer, SedTyp, SedDef, SedSta
// ServiceRoutines
//

#include <stdlib.h>
#include "SedPc.h"

//-------+---------+---------+---------+---------+---------+---------+---------+

// SedDly   set / get response-delay to / from SEDPC
//          A TEST-telegram get also the delay-time, so an old TEST-tel
//          takes the full delay-time.
//
//  call: SedDly (lLn, lplDly, lplErr);
//  para:   lLn     = logical line-number
//          lplDly  = pointer to delay-time in 256 sec, max. 6 bit
//                     the delay-time stored in SEDPC will be read back
//              -1  = only get the delay-time from SEDPC
//          lplErr  = 0  = no errors
//            bit 13 = illegal SEDPC type
//            bit 12 = no such line-number
//            bit  7 = ParameterError: one parameter incorrect
//            bit  6 = Response/Busy not set; data not taken by SEDPC
//            bit  5 = Bytecount mismatch
//  subp: sedsel, sedwr0, sedwr4, sedrd1, sedrd2

void SedDly ( LONG lLn, LONG *lplDly, LONG *lplErr )

{
	#define SPFS    0x2500      // SpecialFunction SetDelay
	#define SPFG    0x2600      // SpecialFunction GetDelay

	static LONG dat;

	if ( *lplErr = sedsel(lLn) )    // set device-addresses to externals
		goto exit;

	if ( *lplDly != -1 ) {			// set delay ?

		if ( *lplErr = sedwr4(SPFS, *lplDly) )  // fun & dly to SedPc, error ?
			goto exit;
		if ( *lplErr = sedwr0 (LASTBY) ) // LastByteMarker to SedPc, error ?
			goto exit;

		*lplErr = sedrd1(&dat);     // get status, error ?
		if ( dat != 0x20 )          // status-error ?
			*lplErr |= 0x20;
		*lplErr |= sedrd2 (&dat);   // get DummyData
		*lplErr |= sedrd1 (&dat);   // get LastByteMarker
		if ( dat != LASTBY )        // LastByteMarker not set ?
			*lplErr |= 0x20;
	}

						// get delay and setup external tdly

	if ( *lplErr = sedwr4(SPFG, 0) )    // function to SedPc, error ?
		goto exit;
	if ( *lplErr = sedwr0 (LASTBY) )    // LastByteMarker to SedPc, error ?
		goto exit;

	*lplErr = sedrd1(&dat);     // get status, error ?
	if ( dat != 0x20 )          // status error ?
		*lplErr |= 0x20;
	*lplErr |= sedrd2(lplDly);  // get delay, error ?
	glTelDly = *lplDly;
	*lplErr |= sedrd1 (&dat);   // get LastByteMarker
	if ( dat != LASTBY )        // LastByteMarker not set ?
		*lplErr |= 0x8;

exit:
	return;

}		// end of SedDly
//-------+---------+---------+---------+---------+---------+---------+---------+

// SedLcs   Look for Crates or Subaddresses
//
//  call: SedLcs (lLn, lCn, lplNum, lplArr, lplErr);
//  para: lLn    = logical line-number
//        lCn    = crate number, 0..31 valid for look for subaddresses
//                               128  valid for look for crates
//        lplNum = number of crates or subaddresses within a crate found
//        lplArr = pointer to array with crate- or subaddresses
//        lplErr = error return code
//        *lplErr = 0 = no error
//            bit 13 = illegal SEDPC type
//            bit 12 = no such line-number
//            bit  7 = ParameterError: one parameter incorrect
//            bit  6 = Response/Busy not set; data not taken by SEDPC
//            bit  5 = Bytecount mismatch
//            bit  4 = Timeout: no response from SEDPC
//            bit  3 = LastByte marker not set or illegal function
//  subp: sedsel, sedwr0, sedwr4, sedrd1, sedrd2

void SedLcs ( LONG lLn, LONG lCn, LONG *lplNum, LONG *lplArr, LONG *lplErr )
{

	static LONG dat, spf, bc, icn, max, i;

	bc = 0;
	if ( *lplErr = sedsel(lLn) ) // set device-addresses to externals
		goto exit;

	if ( lCn == 128 ) {         // look for crates ?
		spf = 0x900;            // yes
		icn = 0;
		max = 32 * 17;
	} else {
		if ( lCn <= 31 ) {      // look for subaddresses ?
			spf = 0xA00;        // yes
			icn = lCn;
			max = 256;
		} else
			goto err80;
	}

	dat = spf | icn;
	*lplErr  = sedwr4(dat, 0);  // special function and cn to SedPc, error ?
	*lplErr |= sedwr0(LASTBY);
	if ( *lplErr )
		goto exit;

	for ( i = 0; i < max; i++ ) { // wait max * timeout of sedrd1 (5sec)
		if ( ! (*lplErr = sedrd1(&dat)) ) // get status & count, no error ?
			break;
	}
	if ( i >= max )             // time-out reached ?
		goto errF;
	if ( dat != 0x30 )          // special function code not set ?
		goto err8;
	*lplErr |= sedrd2(&bc);     // get ByteCount
	if ( bc > 256 )             // overflow ?
		goto err20;

	for ( i = 0; i < bc; i++ ) { // get all addresses from Fifo
		*lplErr |= sedrd1(&dat); // get a byte
		*(lplArr+i) = dat;      // addresses to array
	}

	*lplErr |= sedrd1(&dat);    // get a byte
	if ( dat != LASTBY )        // LastByteMarker missing ?
		*lplErr |= 0x8;

exit:
	*lplNum = bc;
	return;

err8:
	*lplErr = 0x8;              // SpecialFunctionCode not set
	goto exit;

errF:
	*lplErr = 0xF;              // TimeOut reached
	goto exit;

err20:
	*lplErr = 0x20;             // ByteCount overflow
	bc = 0;
	goto exit;

err80:
	*lplErr = 0x80;             // ParameterError
	goto exit;

}		// end SedLcs
//-------+---------+---------+---------+---------+---------+---------+---------+

// SedVer   get a version-string from SEDPC and put it into an array.
//            in the next line you get the used port-address and delay-time.
//
//  call: SedVer (lLn, lpszVer, lplErr);
//  para:   lLn     = logical line-number
//          lpszVer = array to put the token string, >= 128 character
//          lplErr  = 0 = no errors
//            bit 13 = illegal SEDPC type
//            bit 12 = no such line-number
//            bit  7 = ParameterError: one parameter incorrect
//            bit  6 = Response/Busy not set; data not taken by SEDPC
//            bit  5 = Bytecount mismatch
//  subp: sedsel, sedwr0, sedwr4, sedrd1, sedrd2

void SedVer ( LONG lLn, CHAR *lpszVer, LONG *lplErr )

{
	#define SPFV    0x2000      // special-function get version-string
	#define MAXLIN  64          // 2*64 = 128 = max StringLength

	static LONG dat, bc, i;
	static CHAR lin[MAXLIN];
	static CHAR *p;

	if ( *lplErr = sedsel(lLn) )    // set device-addresses to externals
		goto exit;

	p = lpszVer;
	sprintf (lin, "\r\nAddr = %X hex   Delay = %u usec\r\n",
				glDatReg, glTelDly);

	if ( *lplErr = sedwr4(SPFV, 0) )  // function-code to SedPc, error ?
		goto exit;
	if ( *lplErr = sedwr0 (LASTBY) )  // LastByteMarker to SedPc, error ?
		goto exit;

	*lplErr = sedrd1(&dat);     // get status, error ?
	if ( dat != 0x30 )          // special function mark wrong ?
		*lplErr |= 0x20;
	*lplErr |= sedrd2(&bc);     // get byte-count, error ?
	if ( bc > MAXLIN ) {        // ByteCount too big ?
		bc = MAXLIN;
		*lplErr |= 0x20;
	}
	if ( bc < 1 ) {		        // ByteCount too less ?
		bc = 0;
		*lplErr |= 0x20;
		p++;
	}

	for ( i = 0; i < bc; i++) { // copy all bytes to array
		*lplErr |= sedrd1(&dat);
		*(p++) = (BYTE) dat;
	}
	p--;
	*lplErr |= sedrd1 (&dat);   // get LastByteMarker
	if ( dat != LASTBY )        // LastByteMarker not set ?
		*lplErr |= 0x8;

exit:
	i = 0;
	do {                        // copy rest to string
		*(p++) = lin[i];
	} while ( lin[i++] );

	return;

}		// end of SedVer
//-------+---------+---------+---------+---------+---------+---------+---------+

// SedTyp   get Information of SedPc Typ
//          HardwareTyp and Release, FirmwareVersion and Release
//          FifoLength (Number of Bytes)
//
//  call: SedTyp (lLn, lplHdwTyp, lplHdwRel, lplFirVer, lplFirRel, lplFifLen, lplErr);
//  para:   lLn         = logical line-number
//          lplHdwTyp   = pointer to HardwareTyp
//          lplHdwRel   = pointer to HardwareRelease
//          lplFirVer   = pointer to FirmwareVersion
//          lplFirRel   = pointer to FirmwareRelease
//          lplFifLen   = pointer to FifoLength
//          lplErr = 0  = no errors
//            bit 13 = illegal SEDPC type
//            bit 12 = no such line-number
//            bit  7 = ParameterError: one parameter incorrect
//            bit  6 = Response/Busy not set; data not taken by SEDPC
//            bit  5 = Bytecount mismatch
//  subp: sedsel, sedwr0, sedwr4, sedrd1, sedrd2

void SedTyp ( LONG lLn, LONG *lplHdwTyp, LONG *lplHdwRel,
       LONG *lplFirVer, LONG *lplFirRel, LONG *lplFifLen, LONG *lplErr )

{
	static LONG dat;

	if ( *lplErr = sedsel(lLn) )        // set device-addresses to externals
		goto exit;

	*lplErr  = sedwr4 (0x2100, 0);
	*lplErr |= sedwr0 (LASTBY);
	if ( *lplErr )						// error set ?
		goto exit;
	*lplErr = sedrd1(&dat);             // get status, error ?
	if ( dat != 0x20 )                  // status error ?
		*lplErr |= 0x20;
	*lplErr |= sedrd1(lplHdwTyp);       // get info, error ?
	*lplErr |= sedrd1(lplHdwRel);       // get info, error ?
	*lplErr |= sedrd1 (&dat);           // get LastByteMarker
	if ( dat != LASTBY )                // LastByteMarker not set ?
		*lplErr |= 0x20;
	if ( *lplErr )						// error set ?
		goto exit;

	*lplErr  = sedwr4 (0x2200, 0);
	*lplErr |= sedwr0 (LASTBY);
	if ( *lplErr )						// error set ?
		goto exit;
	*lplErr = sedrd1(&dat);             // get status, error ?
	if ( dat != 0x20 )                  // status error ?
		*lplErr |= 0x20;
	*lplErr |= sedrd1(lplFirVer);       // get info, error ?
	*lplErr |= sedrd1(lplFirRel);       // get info, error ?
	*lplErr |= sedrd1 (&dat);           // get LastByteMarker
	if ( dat != LASTBY )                // LastByteMarker not set ?
		*lplErr |= 0x20;
	if ( *lplErr )						// error set ?
		goto exit;

	*lplErr  = sedwr4 (0x2300, 0);
	*lplErr |= sedwr0 (LASTBY);
	if ( *lplErr )						// error set ?
		goto exit;
	*lplErr = sedrd1(&dat);             // get status, error ?
	if ( dat != 0x20 )                  // status error ?
		*lplErr |= 0x20;
	*lplErr |= sedrd2(lplFifLen);       // get info, error ?
	*lplErr |= sedrd1 (&dat);           // get LastByteMarker
	if ( dat != LASTBY )                // LastByteMarker not set ?
		*lplErr |= 0x8;

exit:
	return;
}		// end of SedTyp
//-------+---------+---------+---------+---------+---------+---------+---------+

// SedDef   set SedPc defaults: Delay = 0, TestTel = OldTestTel = disabled
//
//  call: SedDef (lLn, lplErr);
//  para:   lLn         = logical line-number
//          lplErr = 0  = no errors
//            bit 13 = illegal SEDPC type
//            bit 12 = no such line-number
//            bit  7 = ParameterError: one parameter incorrect
//            bit  6 = Response/Busy not set; data not taken by SEDPC
//            bit  5 = Bytecount mismatch
//  subp: sedsel, sedwr0, sedwr4, sedrd1, sedrd2

void SedDef ( LONG lLn, LONG *lplErr )

{
	static LONG dat;

	if ( *lplErr = sedsel(lLn) ) // set device-addresses to externals
		goto exit;

	*lplErr  = sedwr4 (0x3300, 0);
	*lplErr |= sedwr0 (LASTBY);
	if ( *lplErr )              // error set ?
		goto exit;
	*lplErr = sedrd1 (&dat);    // get sedac-status
	if ( (dat & 0xF0) != 0x20 ) // function or byte-count error ?
		*lplErr |= 0x20;
	*lplErr |= sedrd2 (&dat);   // get DummyData
	*lplErr |= sedrd1 (&dat);   // get LastByteMarker
	if ( dat != LASTBY )
		*lplErr |= 0x8;

exit:
	return;
}		// end of SedDef
//-------+---------+---------+---------+---------+---------+---------+---------+

// SedSta   get (and clear) SedPc Status
//
//  call: SedSta (lLn, lClr lplSta, lplErr);
//  para:   lLn     = logical line-number
//			lClr	= clear Bit 7..5 after read if != 0
//          lplSta  = pointer to status
//            lSta Bit: 7 - Reset with SelfTest
//            lSta Bit: 6 - Reset without SelfTest
//            lSta Bit: 5 - SelfTestError
//            lSta Bit: 3 - ErrorCounter != 0
//            lSta Bit: 1 - TestTel enabled
//            lSta Bit: 0 - Delay != 0
//          lplErr  = 0  = no errors
//            bit 13 = illegal SEDPC type
//            bit 12 = no such line-number
//            bit  7 = ParameterError: one parameter incorrect
//            bit  6 = Response/Busy not set; data not taken by SEDPC
//            bit  5 = Bytecount mismatch
//  subp: sedsel, sedwr0, sedwr4, sedrd1, sedrd2

void SedSta ( LONG lLn, LONG lClr, LONG *lplSta, LONG *lplErr )

{
	static LONG dat;

	if ( *lplErr = sedsel(lLn) )    // set device-addresses to externals
		goto exit;

	if ( lClr == 0 )
		*lplErr  = sedwr4 (0x3000, 0);  // read
	else
		*lplErr  = sedwr4 (0x3100, 0);  // read and clear
	*lplErr |= sedwr0 (LASTBY);
	if ( *lplErr )              // error set ?
		goto exit;

	*lplErr = sedrd1(&dat);     // get status, error ?
	if ( dat != 0x20 )          // status-error ?
		*lplErr |= 0x20;
	*lplErr |= sedrd1 (lplSta); // get SedPc Status
	*lplErr |= sedrd1 (&dat);   // get DummyData
	*lplErr |= sedrd1 (&dat);   // get LastByteMarker
	if ( dat != LASTBY )        // LastByteMarker not set ?
		*lplErr |= 0x8;

exit:
	return;

}		// end of SedSta
//-------+---------+---------+---------+---------+---------+---------+---------+

//end of SedSer.c
