#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#ifdef UNIX
#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/kd.h>
#include <sys/time.h>
#include <dlfnc.h>
#endif
#include <sys/types.h>
#include <linux/unistd.h>

#include "prolog.h"
#include "cdi.h"	// interface to user program
#include "SedPc.h"
#include "cdiSedpcLib.h"

#define SD_WRITE 0x40
#define SD_READ  0x80

#define N_SEDAC_LINES 8
#define MAX_SEDAC_PUT 350   /* number of telegrams to put (4k FIFO 680 telegrams) */
#define NS_WAIT_MAX_PUT 48000000     /* ns wait time of sending MAX_SEDAC_PUT / 2 telegrams  */
#define WAIT_TIME_NS 1000000  // 1ms

typedef struct
{
  long TypeAndCrate;
  long BaseSubAddress;
  long data;
} SedpcIn;
typedef struct
{
  long status;
  long data;
} SedpcOut;
typedef struct
{
  SedpcIn *sedMem;
  int *reqList;
  int numTelegrams, numRequests, reqestCnt, redCnt;
} SedBlock;

int nSedacLines = N_SEDAC_LINES;
#define N_THREAD_IDS  4
typedef struct
{
  int numberTelegrams, cdiLine;
	void *sedBuffer;
  int numberOfThreadIds, threadIDs[N_THREAD_IDS];
} SEDPC_DATA;
SEDPC_DATA sedacInfo[N_SEDAC_LINES] = { /* init */
  {0, -1, NULL, 0}, {0, -1, NULL, 0}, {0, -1, NULL, 0}, {0, -1, NULL, 0},  {0, -1, NULL, 0}, {0, -1, NULL, 0}, {0, -1, NULL, 0}, {0, -1, NULL, 0}
};

char *scanCommand[] = {"getcommands","getlines","getcrates","getmodules","readmodules","writemodules"};
enum scommand {getcommands, getlines, getcrates, getmodules, readmodules, writemodules, nCommands};

extern int setIOpermission(int lineNumber, int on);
 
__attribute__((constructor)) void cdiSedacMain(void)
{
    int cc = 0;

    if ((cc=cdiRegisterBusFilter("SEDPC",filterSedacBus)) != 0 ) goto err;
    if ((cc=cdiRegisterBusInitialization("SEDPC",initSedacForCDI)) != 0 ) goto err;
    if ((cc=cdiRegisterBusHandler("SEDPC",execSedacBus)) != 0 ) goto err;
    if ((cc=cdiRegisterBusCleanup("SEDPC",exitSedac)) != 0 ) goto err;
    if ((cc=cdiRegisterBusScanner("SEDPC",sedacScanFcn)) != 0 ) goto err;
    if ((cc=cdiRegisterBusBlockFunction("SEDPC",sedacBlockTrans)) != 0 ) goto err;

err:
    if (cc)
    {
      cdilog("%s : failure in bus registration SEDAC",erlst[cc]);
    }
    return;
}

int sedacScanFcn(char *target,char *result,int resultBufferSize)
{
  char *c, lclTarget[256];
  int i, j, cnt, ret, slen = 0;
  int nData, data[10];
  short rData, wData;
  short line = -1, crate = -1, subaddr = -1;

  if (result == NULL) return argument_list_error;
  if (resultBufferSize < 32) return buffer_too_small;
  // parse target according to "commandString:<line>:<crate>:<subaddr>"; e.g. NULL or "command:1" or "command:1:48" or "command:1:48:16" 
  if (target == NULL || strlen(target) == 0 ) target = "getcommands";
  strncpy(lclTarget,target,255);
  strlwr(lclTarget);
  // get command string
  if ((c=strtok(lclTarget,":")) == NULL) return 0;  // nothing to do
  for (j=0; j<nCommands; j++) if ( strcmp(scanCommand[j], c) == 0 ) break;
  if ( j == nCommands ) j = getcommands; // return the available commands as 'help'
  for ( nData=0; (c=strtok(NULL,":")) != NULL; nData++ ) data[nData] = strtol(c,NULL,0); //atoi(c);

  
  if (j == getcommands)
  {
    snprintf(result,resultBufferSize,"availabled command strings:\n");
    for (i=0; i<nCommands; i++)
    {
      if ((slen = strlen(result)) > resultBufferSize - 12) break;
      strncat(result,scanCommand[i],resultBufferSize-slen);
      if (i <nCommands-1) strcat(result,",");
    }
    return 0;
  }
  
  
  if ( j == getlines )
  {
    snprintf(result,resultBufferSize,"Sedac lines configured:\n");
    for (i=0, cnt=0; i<N_SEDAC_LINES; i++)
    {
      if ( sedacInfo[i].cdiLine < 0 ) continue;
      if ((slen = strlen(result)) > resultBufferSize - 1) break;
      if ( cnt > 0 )
      {
        snprintf(&result[slen],resultBufferSize-slen,",");
        if ((slen = strlen(result)) > resultBufferSize - 1) break;
      }
       snprintf(&result[slen],resultBufferSize-slen,"%d",i+1);
      cnt++;
    }
    return 0;
  }
  
  line = data[0];
  if ( nData < 1 || line < 1 || line > nSedacLines || sedacInfo[line-1].cdiLine < 0) return no_such_line;

  if ( j == getcrates )
  {
    snprintf(result,resultBufferSize,"crates configured on line %d:\n",line);
    cdiMutexLock(&busHandle[sedacInfo[line-1].cdiLine]->lineMutex);
    for (crate=0, cnt=0; crate<32; crate++)
    {
      ret = sedac(line, crate, 253, CA_READ, &rData, 1, NULL);
      if ( ret ) continue;
      if ((slen = strlen(result)) > resultBufferSize - 1) break;
      if ( cnt > 0 )
      {
        snprintf(&result[slen],resultBufferSize-slen,",");
        if ((slen = strlen(result)) > resultBufferSize - 1) break;
      }
      snprintf(&result[slen],resultBufferSize-slen,"%d",crate);
      cnt++;
    }
    cdiMutexUnlock(&busHandle[sedacInfo[line-1].cdiLine]->lineMutex);
    return 0;
  }
  else if ( j == getmodules )
  {
    crate = data[1];
    if ( nData < 2 || crate < 0 || crate >= 32 ) return device_not_connected;
    snprintf(result,resultBufferSize,"Subaddress on line %d crate %d was found:\n",line,crate);
    cdiMutexLock(&busHandle[sedacInfo[line-1].cdiLine]->lineMutex);
    for (subaddr=0, cnt=0; subaddr<256; subaddr++)
    {
      ret = sedac(line, crate, subaddr, CA_READ, &rData, 1, NULL);
      if ( ret ) continue;
      if ((slen = strlen(result)) > resultBufferSize - 1) break;
      if ( cnt > 0 )
      {
        snprintf(&result[slen],resultBufferSize-slen,",");
        if ((slen = strlen(result)) > resultBufferSize - 1) break;
      }
      snprintf(&result[slen],resultBufferSize-slen,"%d",subaddr);
      cnt++;
    }
    cdiMutexUnlock(&busHandle[sedacInfo[line-1].cdiLine]->lineMutex);
    return 0;
  }
  crate = data[1];
  subaddr = data[2];
  if ( nData < 3 || crate < 0 || crate >= 32 || subaddr < 0 || subaddr > 255) return device_not_connected;
  if ( j == readmodules )
  {
    snprintf(result,resultBufferSize,"Read Data on line %d crate %d subaddr %d:\n",line,crate,subaddr);
    cdiMutexLock(&busHandle[sedacInfo[line-1].cdiLine]->lineMutex);
    ret = sedac(line, crate, subaddr, CA_READ, &rData, 1, NULL);
    cdiMutexUnlock(&busHandle[sedacInfo[line-1].cdiLine]->lineMutex);
    if ((slen = strlen(result)) < resultBufferSize - 18) snprintf(&result[slen],resultBufferSize-slen,"%d <return code %d>",rData,ret);
  }
  else if ( j == writemodules )
  {
    wData = data[3];
    if ( nData < 4 ) return argument_list_error;
    snprintf(result,resultBufferSize,"Write Data on line %d crate %d subaddr %d:\n",line,crate,subaddr);
    cdiMutexLock(&busHandle[sedacInfo[line-1].cdiLine]->lineMutex);
    ret = sedac(line, crate, subaddr, CA_WRITE, &wData, 1, NULL);
    cdiMutexUnlock(&busHandle[sedacInfo[line-1].cdiLine]->lineMutex);
    if ((slen = strlen(result)) < resultBufferSize - 18) snprintf(&result[slen],resultBufferSize-slen,"%d <return code %d>",wData,ret);
  }  
  return 0;
}

int sedacBlockTrans(int *requestIndex, int numberRequest, void **blockBuffer, int openFlag)
{ int i, j, cnt, retCode, busLine, cdiLine, access, wTele, rTele, numRead, numTele;
  CdiRequestInfoBlk *pReq;
  SedBlock *psedBlock;
  SedpcIn *sedin;

  if ( openFlag == 0  )  /* free buffer  */
  {
    if ( *blockBuffer == NULL || ((SedBlock *)*blockBuffer)->reqList[0] == 0 )
    {
      cdilog("sedacBlockTrans close Error! buffer pointer wrong !! buffer: 0x%x\n", *blockBuffer);
      return -invalid_parameter;
    }
    pReq = (CdiRequestInfoBlk *)((SedBlock *)*blockBuffer)->reqList[0];
    if ( pReq->pBlockBuffer != *blockBuffer )
    {
      cdilog("sedacBlockTrans close Error! buffer pointer wrong !! input buffer: 0x%x, allocated buffer: 0x%x, reqestIndex: %d\n", *blockBuffer, pReq->pBlockBuffer, pReq->requestNumber + 0x10000);
      return -invalid_parameter;
    }
    free(((SedBlock *)*blockBuffer)->reqList);
    free(((SedBlock *)*blockBuffer)->sedMem);
    free(*blockBuffer);
    *blockBuffer = NULL;
    cdilog("sedacBlockTrans close OK! reqestIndex: %d\n", pReq->requestNumber + 0x10000);
    return 0;
  }
  *blockBuffer = NULL;
  for (i=0; i<numberRequest; i++)   // check line
  {
    if ( (retCode = get_RequestBufferPointer(requestIndex[i], &pReq)) < 0 ) return retCode;
    if ( i == 0 )
    {
      access = pReq->accessFlag;
      if ( access != cdiReadFlag && access != cdiWriteReadFlag && access != cdiReadReadFlag && access != cdiWriteReadReadFlag) return -illegal_read_write;
      busLine = ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine;
      cdiLine = ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->cdiLine;
      wTele = pReq->outLength / sizeof(short); 
      rTele = pReq->inLength / sizeof(short); 
    }
    else if (busLine != ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine || cdiLine != ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->cdiLine ) return -illegal_line;
  }
  numTele = (rTele + wTele) * numberRequest;
  if ( (psedBlock = calloc(1, sizeof(SedBlock))) == NULL )
  {
    cdilog("sedacBlockTrans calloc SedBlock memory error: line: %d; num Telegram: %d\n",busLine, numTele);
    return -out_of_local_memory;
  }
  if ( (psedBlock->sedMem = calloc(numTele, sizeof(SedpcIn))) == NULL )
  {
    free(psedBlock);
    cdilog("sedacBlockTrans calloc sedBuffer memory error: line: %d; num Telegram: %d\n",busLine, numTele);
    return -out_of_local_memory;
  }
  if ( (psedBlock->reqList = calloc(numberRequest, sizeof(int))) == NULL )
  {
    free(psedBlock->sedMem);
    free(psedBlock);
    cdilog("sedacBlockTrans calloc reqList memory error: line: %d; num Request: %d\n",busLine, numberRequest);
    return -out_of_local_memory;
  }
  psedBlock->numTelegrams = numTele;
  psedBlock->numRequests = numberRequest;
  *blockBuffer = psedBlock;
  sedin = psedBlock->sedMem;
  for (i=0, cnt=0; i<numberRequest; i++)   // store data
  {
    get_RequestBufferPointer(requestIndex[i], &pReq);
    psedBlock->reqList[i] = (int)pReq;
    if ( access == cdiWriteReadFlag || access == cdiWriteReadReadFlag ) for (j=0; j<wTele; j++, cnt++)
    {
      sedin[cnt].TypeAndCrate = (WORD)SD_WRITE | pReq->pDev->cdiAddr;
      sedin[cnt].BaseSubAddress = (WORD)(pReq->pDev->cdiSubAddr + pReq->devParameters[sedacSubOffsetSecond]);
      sedin[cnt].data = ((WORD *)pReq->pOutData)[j];
    }
    if ( access == cdiWriteReadReadFlag || access == cdiReadReadFlag )
      numRead = (pReq->inLength < pReq->pDev->dataOuputSize ? pReq->inLength : pReq->pDev->dataOuputSize) / sizeof(short);
    else numRead = rTele;
    for (j=0; j<numRead; j++, cnt++)
    {
      sedin[cnt].TypeAndCrate = (WORD)SD_READ | pReq->pDev->cdiAddr;
      sedin[cnt].BaseSubAddress = (WORD)(pReq->pDev->cdiSubAddr + pReq->devParameters[sedacSubOffset]);
      sedin[cnt].data = 0;
    }
    if ( (access == cdiWriteReadReadFlag || access == cdiReadReadFlag) && j < rTele ) for (; j<rTele; j++, cnt++)
    {
      sedin[cnt].TypeAndCrate = (WORD)SD_READ | pReq->pDev->cdiAddr;
      sedin[cnt].BaseSubAddress = (WORD)(pReq->pDev->cdiSubAddr + pReq->devParameters[sedacSubOffsetThird]);
      sedin[cnt].data = 0;
    }
  }
  cdilog("sedacBlockTrans init sucess! line: %d; num Request: %d, reqestIndex: %d\n",busLine, numberRequest, requestIndex[0]);
  return 0;
}
void blockTrans(CdiRequestInfoBlk *pReq)
{ int i;
  SedBlock *psedBlock;
  CdiRequestInfoBlk *tmpReq;

  if ( pReq->groupReq != pReq ) return;   // skip if not the first Group member
  psedBlock = (SedBlock *)pReq->pBlockBuffer;
  psedBlock->redCnt = 0;
  psedBlock->reqestCnt = 0;
  for (i=0; i<psedBlock->numRequests; i++)   // reset rest of the request flag
  {
    tmpReq = (CdiRequestInfoBlk *)psedBlock->reqList[i];
    if ( tmpReq->groupReq != tmpReq)
    {
      tmpReq->errorCode = 0;
      if (tmpReq->perror != NULL) *tmpReq->perror = 0;
    }
  }
  sedac((WORD)((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine,0,0,0,NULL,psedBlock->numTelegrams,pReq);
  for (i=0; i<psedBlock->numRequests; i++)   // set rest of the request data back flag
  {
    tmpReq = (CdiRequestInfoBlk *)psedBlock->reqList[i];
    if ( tmpReq->groupReq != tmpReq)
    {
      tmpReq->active = finishRequest; /* finish request */
      tmpReq->NgroupRets = 1;
    }
  }
}

int waitSedPCfinished()
{ int ln;
//  CdiLineInfoBlk *busHandle;

  for (ln=0; ln<N_SEDAC_LINES; ln++)
  {
    if ( sedacInfo[ln].numberTelegrams <= 0 ) continue;
//    if ((busHandle=cdiGetBusHandle(sedacInfo[ln].cdiLine)) == NULL) return -1;  /* something wrong!!  */
//    cdiMutexLock(&(busHandle->lineMutex));
//    cdiMutexUnlock(&(busHandle->lineMutex));
    exitSedac(ln+1);
  }
  return 0;
}

int filterSedacBus(int *parameter, int accessFlag)
{
  return 0;
}
int initSedacForCDI(int busLine, int cdiLine, int numberDevices, char *parameterList)
{
//  cdi_mutex_t tmpMutex;
  int lastError, error = 0;

  /* check environemnt variables to see what's connected: Stadmueller FIFO  */
  if ( parameterList[0] )
  if ( error=setenv("SEDPC",parameterList,1) )
  {
    cdilog("initSedac setenv error: line: %d; error: %d\n",busLine, error);
    return -non_existent;
  }
  else cdilog("initSedac setenv ok! line: %d\n",busLine);

  /* try to get Mutex, is this a first application  */
//  lastError = cdiMutexInit(&tmpMutex, cdiLine);
//  if ( tmpMutex == NULL ) return -ask_error;
//  if ( lastError == ERROR_ALREADY_EXISTS ) cdiMutexLock(&tmpMutex);
  error = initSedac(busLine);
//  if ( lastError == ERROR_ALREADY_EXISTS ) cdiMutexUnlock(&tmpMutex);
  if ( error == 0 ) sedacInfo[busLine-1].cdiLine = cdiLine;
  
  return error;
}
// if sync call, data not pass filter then repeat Transmit
int checkData(CdiRequestInfoBlk *pReq)
{
  struct timespec tv;

  /* check return condition */
  checkReturnCondition(pReq, (char *)pReq->bufLastData, pReq->pInData, pReq->inLength, 0);
  if ( pReq->dataReturnCondition == DATA_PASS_FILTER)
    memcpy( pReq->bufLastData, pReq->pInData, pReq->inLength < 8 ? pReq->inLength : 8);

 	if ( pReq->conditionFlag > anyData && pReq->dataReturnCondition != DATA_PASS_FILTER )
	{
    if ( pReq->callback == NULL )
    {
      if ( getTimeMilliseconds() - pReq->requestTime >= pReq->timeout )
      {
         pReq->errorCode = invalid_data;	// data back for sync call
         if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
      }
      else
      {
        tv.tv_sec = 0; tv.tv_nsec = WAIT_TIME_NS; // delay 1ms
        nanosleep( &tv, NULL );    
        return 1;
      }
    }
    pReq->active = finishRequest;
  }
  return 0;
}

void execSedacBus(CdiRequestInfoBlk *pReq)
{ 
  /* do first part; */
  cdiMutexLock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
  if ( pReq->groupReq && pReq->pBlockBuffer ) blockTrans(pReq);
  else
  {

  do
  {
    doSedacRequest(pReq, 1);
    if ( pReq->active == finishRequest )
    {
      cdiMutexUnlock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
      return;
    }
    /* check return condition */
    if (pReq->accessFlag != cdiReadFlag && pReq->accessFlag != cdiReadWriteFlag) break;
    if ( checkData(pReq) == 0 ) break;
  } while (TRUE);

  if ( pReq->active == finishRequest )
  {
    cdiMutexUnlock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
    return;
  }
  if ( pReq->accessFlag == cdiReadFlag || pReq->accessFlag == cdiWriteFlag)
  {
    pReq->active = finishRequest;
    cdiMutexUnlock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
    return;
  }

  /* do second part */
  do
  {
    doSedacRequest(pReq, 2);
    if ( pReq->active == finishRequest )
    {
      cdiMutexUnlock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
      return;
    }

    /* check return condition */
    if (pReq->accessFlag != cdiWriteReadFlag && pReq->accessFlag != cdiWriteReadWriteFlag) break;
    if ( checkData(pReq) == 0 ) break;
  } while (TRUE);

  if (pReq->accessFlag != cdiWriteReadWriteFlag && pReq->accessFlag != cdiWriteReadReadFlag)
  {
    pReq->active = finishRequest; /* finish request */
    cdiMutexUnlock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
    return;
  }

  /* do third part */
  doSedacRequest(pReq, 3);

  }

  pReq->active = finishRequest; /* finish request */
  cdiMutexUnlock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
}

int clearSedac(int ln)
{ long err,dly=0;

  SedClr(ln,&dly,&err);
  if ( err )  /* twice clear needed if there was an error befor */
  {
    millisleep(20);
    SedClr(ln,&dly,&err);
  }
  if ( err ) cdilog("SedClr error: %d with line: %d\n",err,ln);
  return err;
}
int initSedac(int ln)
{
  char *env;
  char str[80];
  long err;
  long	xglLogLin[XMLN];		// logical LineNumbers
  long	xglDevAdr[XMLN];		// DeviceAddr of Lines
  long	xglModAdr[XMLN];		// ModuleAddr of Lines (from Sed1Pc)

  if (ln<1 || ln>nSedacLines) return -illegal_line;
  /* check environemnt variables to see what's connected: Stadmueller FIFO  */
//  if ( ( err=setenv("SEDPC","1-378",1) ) )
//  {
//    cdilog("initSedac setenv error: line: %d; error: %d\n",ln, err);
//    return -non_existent;
//  }
  if ( ( env=getenv("SEDPC") ) == NULL )
  {
    cdilog("initSedac getenv error: line: %d\n",ln);
    return -non_existent;
  }
  cdilog("initSedac getenv: %s; line: %d\n",env, ln);

  SedIni(xglLogLin, xglDevAdr, xglModAdr, &err);

  cdilog("open sedpc line: %d; total line: %d; %d; error: %d\n",ln, xglLogLin[0], xglLogLin[1], err);

  if ( err )
  {
    if ( err == 0x4000 ) cdilog("io permission failed! error number: 0x%x with line: %d\n",err,ln);
    return -file_not_open;
  }
  err = clearSedac((short)ln);
//  printf(" get tid: %d\n",(long)getpid());

  sedacInfo[ln - 1].threadIDs[0] = getpid();  // store the first PID
  sedacInfo[ln - 1].numberOfThreadIds = 1;

  /*  sedacInfo[ln - 1].numberTelegrams = 0;  /* set buffer size  */
  return err ? -not_ready : 0;
}
int exitSedac(int ln)
{ long err;

  if (ln<1 || ln>nSedacLines) return -illegal_line;
  cdilog("close sedpc line! line number: %d\n", ln);
  if ( sedacInfo[ln - 1].numberTelegrams )
  {
    free ( sedacInfo[ln - 1].sedBuffer );
    sedacInfo[ln - 1].numberTelegrams = 0;
    sedacInfo[ln - 1].sedBuffer = NULL;
  }
  SedExi ( &err );
  return err;
}

int sedpcBlock(WORD lin, WORD cra, WORD bsa, int access, WORD *data, int begin, int ende, int putFlag, CdiRequestInfoBlk *pReq)
{ int i, j, status=0;
  long ntelegrams, tmp, errFlg;
  SedpcIn *sedin;
  SedpcOut *sedout;
  SedBlock *psedBlock;
  CdiRequestInfoBlk *tmpReq;

  ntelegrams = ende - begin;
  if ( putFlag )
  {
    if ( pReq && pReq->pBlockBuffer ) SedPut(lin,&ntelegrams,(long *)&psedBlock->sedMem[begin],(long *)&status);
    else
    {
      sedin = (SedpcIn *)sedacInfo[lin - 1].sedBuffer;
      tmp = ( access == CA_READ ? SD_READ : SD_WRITE ) | cra;
      for (j=begin; j<ende; j++)
      {
        i = j - begin;
        sedin[i].TypeAndCrate = tmp;
        sedin[i].BaseSubAddress = bsa;
        sedin[i].data = access == CA_READ ? 0 : ((WORD *)data)[j];
      }
//    printf("sedput: begin: %d; end: %d\n",begin, ende);
//    p = (long *)sedin;
//    printf("data: 0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x,\n",p[0],p[1],p[2],p[3],p[4],p[5]);
      SedPut(lin,&ntelegrams,(long *)sedin,(long *)&status);
//    printf("sedput: begin: %d; end: %d\n",begin, ende);
    }
  }
  else
  {
    sedout = (SedpcOut *)sedacInfo[lin - 1].sedBuffer;
    SedGet(lin,&ntelegrams, (long *)sedout, (long *)&status);
    if ( pReq && pReq->pBlockBuffer ) for (i=begin; i<ende; i++)   // get data
    {
      j = i * psedBlock->numRequests / psedBlock->numTelegrams;
      tmpReq = (CdiRequestInfoBlk *)psedBlock->reqList[j];
      if ( psedBlock->reqestCnt != j )
      {
        psedBlock->reqestCnt = j;
        psedBlock->redCnt = 0;
      }
      errFlg = (tmpReq->perror && (psedBlock->sedMem[i].TypeAndCrate & SD_READ) == SD_READ ) ? 1 : 0;
      if ( status > 0x7 || (status > 0 && (sedout[i-begin].status & 0x7) != 0) )  /* sedpc error or single telegram error */
      {
        tmpReq->errorCode = sedac_error;
        if ( errFlg ) tmpReq->perror[psedBlock->redCnt] = tmpReq->errorCode;
        tmpReq->deviceError = sedout[i - begin].status;
        tmpReq->pDev->numberDeviceError++;
      } else if ( errFlg ) tmpReq->perror[psedBlock->redCnt] = 0;
      if ( (psedBlock->sedMem[i].TypeAndCrate & SD_WRITE) == SD_WRITE ) continue;
      j = i - begin;
//        if ( status == 0 || (sedout[j].status & 0x7) == 0 ) ((WORD *)tmpReq->pInData)[psedBlock->redCnt] = sedout[j].data;
      ((WORD *)tmpReq->pInData)[psedBlock->redCnt] = sedout[j].data;
      psedBlock->redCnt++;
    }
    else
    {
      if (access == CA_READ)
      {
        errFlg = (pReq && pReq->perror) ? 1 : 0;
        for (i=0; i<ntelegrams; i++)
        {
          ((WORD *)data)[i+begin] = sedout[i].data;
          if (errFlg) pReq->perror[i+begin] = (status == 0 || (sedout[i].status & 0x7) == 0) ? 0 : sedac_error;
        }
      }
    }
  }
  if ( status )
  {
    printf("!!!SedPC put or get(putflag: %d) error: 0x%x with line: %d; begin number: %d\n",putFlag,status,lin, begin);
  }
  return status;
}

// ================  read/write nwords telegram to same addesses  ===================================
int sedac(WORD lin, WORD cra, WORD bsa, int access, WORD *data, int ntelegrams, CdiRequestInfoBlk *pReq)
{ int i, steps, begin, ende, ret=0;
  struct timespec wait_time;
  struct timeval tv; 

  if (ntelegrams == 1)
  {
    steps = *data;
    if (access == CA_READ)
    {
      SedRd(lin,cra,bsa,(long *)&steps,(long *)&ret);
      *data = (WORD)steps;
    }
    else SedWr(lin,cra,bsa,*(long *)data,(long *)&ret);
    if ( ret > 7 )
    {
      gettimeofday (&tv, NULL);
      cdilog("SedRd or SdWt error! clear sedapc with line: %d, accessFlag: %d, time: %d\n",lin, access, tv.tv_sec);
      clearSedac(lin);
    }
    return ret;
  }

  if ( ntelegrams > sedacInfo[lin - 1].numberTelegrams && sedacInfo[lin - 1].numberTelegrams < MAX_SEDAC_PUT )  /* get buffer  */
  {
    if ( sedacInfo[lin - 1].numberTelegrams ) /* freee previous buffer */
    {
      free ( sedacInfo[lin - 1].sedBuffer );
      sedacInfo[lin - 1].numberTelegrams = 0;
      sedacInfo[lin - 1].sedBuffer = NULL;
    }
    steps = ntelegrams > MAX_SEDAC_PUT ? MAX_SEDAC_PUT : ntelegrams;
    if ( (sedacInfo[lin - 1].sedBuffer = (void *)malloc(steps*sizeof(SedpcIn)) ) == NULL ) return -out_of_local_memory;
    sedacInfo[lin - 1].numberTelegrams = steps;
    cdilog("sedPC get memory, Number: %d; size: %d; line: %d\n", steps, sizeof(SedpcIn), lin);
  }
  steps = (ntelegrams - 1) / MAX_SEDAC_PUT + 1;
  wait_time.tv_sec = 0;
  wait_time.tv_nsec = NS_WAIT_MAX_PUT;
  for (i=0; i<steps; i++)
  {
    begin = i * MAX_SEDAC_PUT;
    ende = begin + MAX_SEDAC_PUT;
    if ( ende > ntelegrams ) ende = ntelegrams;
    ret = sedpcBlock(lin, cra, bsa, access, data, begin, ende, 1, pReq);
    if ( ret ) break;
    if ( i )  // after two puts, then get telegram back
    {
      nanosleep( &wait_time, NULL );
      ret = sedpcBlock(lin, cra, bsa, access, data, begin- MAX_SEDAC_PUT, begin, 0, pReq);
    }
    if ( ret > 7 ) break;
  }
  if ( ret <= 7 )
  {
    if ( ende - begin > 105 )  // wait # of ms time
    {
      wait_time.tv_sec = 0;
      wait_time.tv_nsec = ((ende - begin) - 100 ) * 49 * 4000;  /* from 1 ms to 49 ms */
      nanosleep( &wait_time, NULL );
    }
    ret = sedpcBlock(lin, cra, bsa, access, data, begin, ende, 0, pReq);  /* get the very last sent  */
  }
  if ( ret )
  {
    cdilog("sedpc put or get error number(0x%x) with line: %d\n",ret,lin);
    if ( ret > 7 )
    {
      cdilog("sedpc error! clear sedapc with line: %d\n",lin);
      clearSedac(lin);
    }
  }

  return ret;
}

int checkThreadIDs(CdiRequestInfoBlk *pReq)
{ int lin = ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine;

  if ( sedacInfo[lin - 1].numberOfThreadIds > 1 ) return 0; // background task already registered
  sedacInfo[lin - 1].threadIDs[1] = getpid();
  if ( sedacInfo[lin - 1].threadIDs[1] == sedacInfo[lin - 1].threadIDs[0] ) return 0;
  if ( setIOpermission(lin, 1) )
  {
    cdilog("Sedpc setIOpermission(lineNumber=%d) error!\n", lin);
    return hardware_error;
  }
  sedacInfo[lin - 1].numberOfThreadIds = 2;
  return 0;
}

int doSedacRequest(CdiRequestInfoBlk *pReq, int step)
{ int access, ret, length, writeOffset, done;
  void *pOut;

  pReq->errorCode = checkThreadIDs(pReq);
  if ( pReq->errorCode )
  {
    if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
    return ret;
  }
  //printf("try to get Mutex: process pointer: 0x%x; line: %d\n",pReq, pReq->pDev->cdiLine);
//  cdiMutexLock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
  //printf("get Mutex: process pointer: 0x%x; line: %d\n",pReq, pReq->pDev->cdiLine);

  if ( step == 1 ) /* first request */
    access = (pReq->accessFlag != cdiReadFlag && pReq->accessFlag != cdiReadWriteFlag && pReq->accessFlag != cdiReadReadFlag) ? CA_WRITE : CA_READ;
  else if ( step == 2 )/* second */
    access = (pReq->accessFlag == cdiReadWriteFlag || pReq->accessFlag == cdiWriteWriteFlag ) ? CA_WRITE : CA_READ;
  else
    access = pReq->accessFlag == cdiWriteReadReadFlag ? CA_READ : CA_WRITE;

  if ( access == CA_WRITE )
  {
    if ( pReq->accessFlag == cdiWriteWriteFlag || pReq->accessFlag == cdiWriteReadWriteFlag )
    {
      if ( step == 1 )
      {
        length = sizeof(short);
        pOut = (void *)pReq->pOutData;
        writeOffset = sedacSubOffsetSecond;
      }
      else
      {
        length = (pReq->accessFlag == cdiWriteReadWriteFlag && pReq->conditionFlag == anyData) ? sizeof(short) : (pReq->outLength - sizeof(short));
        pOut = (void *)((char *)pReq->pOutData + sizeof(short));
        writeOffset = sedacSubOffsetThird;
      }
    }
    else
    {
      length = (pReq->accessFlag == cdiReadWriteFlag && pReq->conditionFlag == anyData) ? sizeof(short) : pReq->outLength;
      pOut = (void *)pReq->pOutData;
      writeOffset = sedacSubOffsetSecond;
    }
  }
  else
  {
    done = 0;
    if ( pReq->accessFlag == cdiWriteReadReadFlag || pReq->accessFlag == cdiReadReadFlag )
    {
      if ( (pReq->accessFlag == cdiWriteReadReadFlag && step == 2) || (pReq->accessFlag == cdiReadReadFlag && step == 1 ) )
      {
        length = pReq->inLength < pReq->pDev->dataOuputSize ? pReq->inLength : pReq->pDev->dataOuputSize;
        pOut = (void *)pReq->pInData;
        writeOffset = sedacSubOffset;
        done = 1;
      }
      else if ( step == 3 )
      {
        if ( pReq->inLength <= pReq->pDev->dataOuputSize ) return pReq->errorCode;;
        length = pReq->inLength < (pReq->pDev->dataOuputSize + pReq->pDev->dataOuput2ndSize) ? (pReq->pDev->dataOuputSize + pReq->pDev->dataOuput2ndSize - pReq->inLength) : pReq->pDev->dataOuput2ndSize;
        pOut = (void *)((char *)pReq->pInData + pReq->pDev->dataOuputSize);
        writeOffset = sedacSubOffsetThird;
        done = 1;
      }
    }
    if ( done == 0 )
    {
        length = pReq->inLength;
        pOut = (void *)pReq->pInData;
        writeOffset = sedacSubOffset;
    }
  }

  ret = sedac((WORD)((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine, 
          (WORD)pReq->pDev->cdiAddr,
//          (WORD)(pReq->pDev->cdiSubAddr + pReq->devParameters[step == 1 ? sedacSubOffset : sedacSubOffsetSecond]), 
          (WORD)(pReq->pDev->cdiSubAddr + pReq->devParameters[writeOffset]), 
          (int)access,
          (WORD *)pOut,
          (int)length / sizeof(short), pReq );
  if ( ret != 0 )
  {
    pReq->errorCode = ret > 7 ? hardware_error : sedac_error;
    if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
    if ( ret > 0 )
    {
      printf("Sedpc error: line: %d; ca: %d sub: %d; rw_Flag: %d; numTele: %d, error: 0x%x\n",
          ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine,
          pReq->pDev->cdiAddr,
          pReq->pDev->cdiSubAddr + pReq->devParameters[access == CA_READ ? sedacSubOffset : writeOffset],
          access,
          (access == CA_READ ? pReq->inLength : pReq->outLength) / sizeof(short),
          ret);
      pReq->deviceError = ret;    // return from hardware
      pReq->pDev->numberDeviceError++;
    }
    pReq->active = finishRequest; /* finish request */
  }

//  cdiMutexUnlock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);

  return pReq->errorCode;
}

