#include "cdi.def"
#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cdi.h"
#include "cdiSimulateLib.h"
#define SD_WRITE 0x40
#define SD_READ  0x80

#define NUM_SIMULATION_LINES 8
int nSimulationLines = NUM_SIMULATION_LINES;

char *scanCommand[] = {"getcommands","getlines","getcrates","getmodules","readmodule","writemodule"};
enum scommand {getcommands, getlines, getcrates, getmodules, readmodule, writemodule, nCommands};


int initHardware(int line)
{
  /* e.g. some routine to initialize the real hardware */
  return 0;
}

int simulateBusAction(WORD lin, WORD cra, WORD bsa, int access, WORD *data, int ntelegrams)
{ 
  int i;
  static WORD storedWord[32][256]; /* just keep a local static buffer */

  if (cra < 0) cra = 0;
  if (cra > 31) cra = 31;
  if (bsa < 0) bsa = 0;
  if (bsa > 255) bsa = 255;
  if (ntelegrams == 1)
  {
    if (access == CA_READ) *data = storedWord[cra][bsa]; 
    else storedWord[cra][bsa] = *data;
    return 0;
  }

  for (i=0; i<ntelegrams; i++)
  {
    if (access == CA_READ) data[i] = storedWord[cra][bsa]; else storedWord[cra][bsa] = data[i];
  }
  Sleep(ntelegrams/4); // a 250 usec delay per telegram ...
  return 0;
}

CDISIMULATE_EXPORT int scanSimulationBus(char *target,char *result,int resultBufferSize)
{
  char *c, lclTarget[256];
  int i, j, cnt, ret, slen = 0;
  int nData, data[10];
  short rData, wData;
  short line = -1, crate = -1, subaddr = -1;

  if (result == NULL) return argument_list_error;
  if (resultBufferSize < 32) return buffer_too_small;
  /* It depends completely on the bus as to what consitutes a scan operation
   * Here, we will fake it as follows:
   * Parse target according to "commandString:<line>:<crate>:<subaddr>"; 
   * e.g. NULL or "getcrates:1" or "getmodules:1:48" or "readmodule:1:48:16" 
   */
  if (target == NULL || strlen(target) == 0) target = "getcommands";
  strncpy(lclTarget,target,255);
  strlwr(lclTarget);
  /* get command string */
  if ((c=strtok(lclTarget,":")) == NULL) return 0;  /* nothing to do */
  for (j=0; j<nCommands; j++) if ( strcmp(scanCommand[j], c) == 0 ) break;
  if ( j == nCommands ) j = getcommands; /* return the available commands as 'help' */
  for ( nData=0; (c=strtok(NULL,":")) != NULL; nData++ ) data[nData] = strtol(c,NULL,0);
  line = data[0];
  crate = data[1];
  subaddr = data[2];

  switch (j)
  {
    default:
    case getcommands:
      _snprintf(result,resultBufferSize,"availabled command strings:\n");
      for (i=0; i<nCommands; i++)
      {
        if ((slen=(int)strlen(result)) > resultBufferSize - 12) break;
        strncat(result,scanCommand[i],resultBufferSize-slen);
        if (i <nCommands-1) strcat(result,",");
      }
      return 0;
    case getlines:
      _snprintf(result,resultBufferSize,"lines configured:\n");
      for (i=0, cnt=0; i<NUM_SIMULATION_LINES; i++)
      {
        if ((slen = (int)strlen(result)) > resultBufferSize - 1) break;
        if ( cnt > 0 )
        {
          _snprintf(&result[slen],resultBufferSize-slen,",");
          if ((slen = (int)strlen(result)) > resultBufferSize - 1) break;
        }
        _snprintf(&result[slen],resultBufferSize-slen,"%d",i+1);
        cnt++;
      }
      return 0;
    case getcrates:
      if ( nData < 1 || line < 1 || line > nSimulationLines ) return no_such_line;
      if ( j == getcrates )
      {
        _snprintf(result,resultBufferSize,"crates configured on line %d:\n",line);
        for (crate=0, cnt=0; crate<10; crate++)
        {
          if ((slen = (int)strlen(result)) > resultBufferSize - 1) break;
          _snprintf(&result[slen],resultBufferSize-slen,"%d",crate);
          cnt++;
        }
        return 0;
      }
    case getmodules:
      if ( nData < 2 || crate < 0 || crate >= 32 ) return device_not_connected;
      if ( line < 1 || line > nSimulationLines ) return no_such_line;
      _snprintf(result,resultBufferSize,"Subaddresses on line %d crate %d:\n",line,crate);
      for (subaddr=0, cnt=0; subaddr<256; subaddr++)
      {
        if ((slen = (int)strlen(result)) > resultBufferSize - 1) break;
        _snprintf(&result[slen],resultBufferSize-slen,"%d",subaddr);
        cnt++;
      }
      return 0;
    case readmodule:
      if ( nData < 3 || crate < 0 || crate >= 32 || subaddr < 0 || subaddr > 255) return device_not_connected;
      if ( line < 1 || line > nSimulationLines ) return no_such_line;
      _snprintf(result,resultBufferSize,"Read Data on line %d crate %d subaddr %d:\n",line,crate,subaddr);
      ret = simulateBusAction(line, crate, subaddr, CA_READ, &rData, 1);
      if ((slen = (int)strlen(result)) < resultBufferSize - 4) _snprintf(&result[slen],resultBufferSize-slen,"%d <return code %d>",rData,ret);
      return 0;
    case writemodule:
      if ( nData < 3 || crate < 0 || crate >= 32 || subaddr < 0 || subaddr > 255) return device_not_connected;
      if ( line < 1 || line > nSimulationLines ) return no_such_line;
      wData = data[3];
      if ( nData < 4 ) return argument_list_error;
      _snprintf(result,resultBufferSize,"Write Data on line %d crate %d subaddr %d:\n",line,crate,subaddr);
      ret = simulateBusAction(line, crate, subaddr, CA_WRITE, &wData, 1);
      if ((slen = (int)strlen(result)) < resultBufferSize - 4) _snprintf(&result[slen],resultBufferSize-slen,"%d <return code %d>",wData,ret);
      return 0;
  }
  return 0;
}

CDISIMULATE_EXPORT int filterSimulation(int *parameter, int accessFlag)
{
  /* data can be filter according to the bus's filtering criteria here */
  return 0;
}

CDISIMULATE_EXPORT int initSimulation(int busLine, int cdiLine, int numberDevices, char *parameterList)
{
  cdi_mutex_t tmpMutex;
  int lastError, error = 0;
  /* try to get Mutex if this is the first application  */
  lastError = cdiMutexInit(&tmpMutex, cdiLine);
# ifdef WIN32
  if ( lastError == ERROR_ALREADY_EXISTS ) cdiMutexLock(&tmpMutex);
# endif
  error = initHardware(busLine);
# ifdef WIN32
  if ( lastError == ERROR_ALREADY_EXISTS ) cdiMutexUnlock(&tmpMutex);
# endif
  return error;
}

CDISIMULATE_EXPORT void simulationHandler(CdiRequestInfoBlk *pReq)
{ 
  int access, length = sizeof(short); /* fix this to read 1 word from the bus */
  void *pData;

  /* Note: 
     pReq->active can contain 'state' information in case the handler needs to be re-entered
     We will not make use of it here in this simple simulation example 
  */
  switch (pReq->accessFlag)
  {
    default: /* there are more complicated cases, but let's keep it simple here */
    case cdiReadFlag:
      access = CA_READ;
      /* a 'READ' is regarded as input from the hardware into the bus plug */
      pData = (void *)pReq->pInData;
      break;
    case cdiWriteFlag:
      access = CA_WRITE;
      /* a 'WRITE' is regarded as output from the bus plug into the hardware */
      pData = (void *)pReq->pOutData;
      break;
  }

  cdiMutexLock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);

  pReq->errorCode = simulateBusAction(
          (WORD)((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine, 
          (WORD)pReq->pDev->cdiAddr[caAddrIndex],
          (WORD)(pReq->pDev->cdiAddr[subAddrIndex] + pReq->devParameters[0]), 
          (int)access,
          (WORD *)pData,
          (int)length / sizeof(short) );

  if (pReq->errorCode != 0)
  { /* bus action not successful */
    if (pReq->perror != NULL) *(pReq->perror) = pReq->errorCode;
    pReq->deviceError = pReq->errorCode;    
    pReq->pDev->numberDeviceError++;
  }

  pReq->active = finishRequest; /* must signal that the request is finished */
  cdiMutexUnlock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
  return;
}

CDISIMULATE_EXPORT int exitSimulation(int ln)
{
  return 0;
}


