#include "prolog.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/timeb.h>
#include "tine.h"
#include "cdiListener.h"
#include "tinecdi.h"

cdiListener **cdiListenerTable = NULL;
int nListeners = 0;
int defaultPollingMode = CM_POLL;

/**
 * \internal used internally: don't bother checking arguments vs. NULL
 */
int getScheduledProperty(char *prp,char *dev)
{ /* this should be a short list (worry about a hash table sometime ?) */
  int i;
  char *devp=dev;
  if (*devp == '/')
  { /* fully specified device name */
    devp = strrchr(dev,'/'); devp++;
  }
  for (i=0; i<sizCdiSchdLst; i++)
  {
    if (!gCdiSchdLst[i].enabled) continue;
    if (strlen(gCdiSchdLst[i].dev) > 0)
    { /* device specific */
      if (strchr(devp,',') != NULL) continue; /* device groups not allowed here */
      if (wc_stricmp(gCdiSchdLst[i].dev,devp)) continue;
    }
    if (!strnicmp(prp,gCdiSchdLst[i].prp,64)) return i;
  }
  return -1;
}
float getFloatAtIndex(BYTE *dat,int idx,int fmt)
{
   switch (fmt)
   {
     case CF_BYTE:
       return (float)((BYTE *)dat)[idx];
     case CF_INT16:
       return (float)((short *)dat)[idx];
     case CF_INT32:
       return (float)((int *)dat)[idx];
     case CF_FLOAT:
       return (float)((float *)dat)[idx];
     case CF_DOUBLE:
       return (float)((double *)dat)[idx];
     default:
       break;
   }
   return 0;
}
#define isNumberFormat(f) (isPrimitiveFormat(f) && (f) != CF_CHAR)
void listenerCb(int id,int cc)
{
  int i, siz, fmt, mlid = id - 1, fireSched = FALSE;
  float t = 0, frb, fv;
  cdiListener *ml;
  if (cdiDebugLevel) dbglog("listenerCb: entered with id %d cc %d",id,cc);
  if (mlid < 0 || mlid >= MAX_NUM_LISTENERS)
  {
    dbglog("listenerCb: listener id is invalid!",mlid);
  }
  if ((ml=cdiListenerTable[mlid]) == NULL) 
  {
    if (cdiDebugLevel) dbglog("listenerCb: could not find listener for this id!");
    return;
  }
  fmt = ml->dout.dFormat;
  ml->status = cc;
  ml->timestamp = makeDataTimeStamp();
  siz = ml->dout.dArrayLength*fmtsizeof(fmt);

  if (ml->isSched)
  { /* check scheduling criteria */
    if (isNumberFormat(fmt) && (ml->atolSched + ml->ptolSched) != 0)
    { /* is a number type and has a tolerance set */
      for (i=0; i<siz; i++)
      { /* run thru the list */
        frb = getFloatAtIndex(ml->rbbuf,i,fmt);
        fv = getFloatAtIndex(ml->outbuf,i,fmt);
        t = (float)fabs(ml->atolSched + frb * ml->ptolSched);
        if (fabs(fv-frb) > t)
        { /* this can creep without ever firing! (successive changes just under tolerance) */
          fireSched = TRUE;
          break;
        }
      }
    }
    else
    {
      if (memcmp(ml->rbbuf,ml->outbuf,siz) != 0) 
      {
        fireSched = TRUE;
      }
    }
  }
  memcpy(ml->rbbuf,ml->outbuf,siz);

  if (fireSched) 
  {
    char schstr[132];
    if (ml->isSched & CDI_SCHED_DEV)
    { /* schedule property and specific device */
      char *devp=ml->devName;
      if (*devp == '/')
      { /* fully specified device name */
        devp = strrchr(ml->devName,'/'); devp++;
      }
      sprintf(schstr,"%.64s<%.64s>",ml->devProperty,devp);
    }
    else
    {
      sprintf(schstr,"%.64s",ml->devProperty);
    }
    SystemScheduleProperty(CDIEQM_TAG,schstr);
  }
  return;
}
int startListenerCompCode = 0;
int isExiting = FALSE;
cdiListener *startListener(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,int pollrate,int flags)
{
  int i,p,lid,siz,mode=defaultPollingMode|CM_WAIT;
  cdiListener *ml=NULL;
  char *xc = strlen(devName) > 64 ? "(+device sequence)" : "";

  if (isExiting) return NULL;
  cdilog("startListener: starting new listener for %.64s %s %.64s at %d msec",devName,xc,devProperty,pollrate);
  if (cdiListenerTable == NULL)
  {
    if ((cdiListenerTable=(cdiListener **)SystemCalloc(MAX_NUM_LISTENERS,sizeof(cdiListener *))) == NULL)
    {
      cdilog("Cannot allocate listener table !");
      startListenerCompCode = out_of_local_memory;
      return NULL;
    }
  }
  for (i=0; i<MAX_NUM_LISTENERS && cdiListenerTable[i] != NULL; i++);
  if (i == MAX_NUM_LISTENERS) 
  {
    cdilog("listener table full !");
    startListenerCompCode = resources_exhausted;
    return NULL;
  }
  if (i == nCdiListeners) nCdiListeners++;
  siz = dout->dArrayLength * fmtsizeof(dout->dFormat);
  if ((ml=(cdiListener *)calloc(1,sizeof(cdiListener))) == NULL) 
  {
    cdilog("Cannot allocate new listener !");
    startListenerCompCode = out_of_local_memory;
    return NULL;
  }
  if ((ml->outbuf=(BYTE *)calloc(siz,1)) == NULL) 
  {
    free(ml);
    cdilog("Cannot allocate data buffer for listener !");
    startListenerCompCode = out_of_local_memory;
    return NULL;
  }
  if ((ml->rbbuf=(BYTE *)calloc(siz,1)) == NULL) 
  {
    free(ml->outbuf);
    free(ml);
    cdilog("Cannot allocate data buffer for listener !");
    startListenerCompCode = out_of_local_memory;
    return NULL;
  }
  ml->outlen = dout->dArrayLength;
  ml->dout.dArrayLength = dout->dArrayLength;
  ml->dout.dFormat = dout->dFormat;
  memcpy(ml->dout.dTag,dout->dTag,TAG_NAME_SIZE);
  ml->rbbuf;  /* pointer to original data reference */
  ml->dout.data.vptr = ml->outbuf; /* pointer to working area buffer */
  if (din != NULL && (siz=din->dArrayLength * fmtsizeof(din->dFormat)) > 0)
  {
    if ((ml->inbuf=(BYTE *)calloc(siz,1)) == NULL) 
    {
      if (ml->outbuf != NULL) free(ml->outbuf);
      if (ml->rbbuf != NULL) free(ml->rbbuf);
      free(ml);
      cdilog("Cannot allocate data buffer for listener !");
      startListenerCompCode = out_of_local_memory;
      return NULL;
    }
    ml->din.data.vptr = ml->inbuf;
    memcpy(ml->din.data.vptr,din->data.vptr,siz);
    ml->din.dArrayLength = din->dArrayLength;
    ml->din.dFormat = din->dFormat;
    memcpy(ml->din.dTag,din->dTag,TAG_NAME_SIZE);
  }
  else
  {
    ml->din.dFormat = CF_NULL;
    ml->din.dArrayLength = 0;
    ml->din.data.vptr = NULL;
    memset(ml->din.dTag,0,TAG_NAME_SIZE);
  }
  siz = (int)strlen(devName)+4; siz += 4-(siz%4);
  if ((ml->devName=(char *)calloc(siz,1)) == NULL) 
  {
    if (ml->outbuf != NULL) free(ml->outbuf);
    if (ml->inbuf != NULL) free(ml->inbuf);
    if (ml->rbbuf != NULL) free(ml->rbbuf);
    free(ml);
    cdilog("Cannot allocate device name space for listener !");
    startListenerCompCode = out_of_local_memory;
    return NULL;
  }
  strcpy(ml->devName,devName);
  strncpy(ml->devProperty,devProperty,PROPERTY_NAME_SIZE);

  ml->mlid = i;
  cdiListenerTable[i] = ml;

  if ((p=getScheduledProperty(devProperty,devName)) >= 0)
  { /* property is scheduled ! */
    ml->isSched = CDI_SCHED_PRP;
    if (strlen(gCdiSchdLst[p].dev) > 0) ml->isSched |= CDI_SCHED_DEV;
    ml->atolSched = gCdiSchdLst[p].atol;
    ml->ptolSched = gCdiSchdLst[p].ptol;

  }
  if (flags > 0) mode |= flags;
  if (pollrate < 0) pollrate = DEFAULT_POLLING_INTERVAL;
  lid = AttachLinkEx(devName,devProperty,&ml->dout,&ml->din,CA_READ,pollrate,listenerCb,mode,ml->mlid+1);

  if (lid < 0)
  {
    if (ml->rbbuf != NULL) free(ml->rbbuf);
    if (ml->outbuf != NULL) free(ml->outbuf);
    if (ml->inbuf != NULL) free(ml->inbuf);
    if (ml->devName != NULL) free(ml->devName);
    free(ml);
    cdiListenerTable[i] = NULL;
    cdilog("startListener for %.64s %s %.64s failed : %s",devName,xc,devProperty,erlst[-lid]);
    startListenerCompCode = -lid;
    return NULL;
  }
  if (ml->dout.dArrayLength != 0)
  {
    memcpy(dout->data.vptr,ml->outbuf,dout->dArrayLength*fmtsizeof(dout->dFormat));
  }
  ml->pollrate = pollrate;
  ml->lid = lid;
  startListenerCompCode = 0;
  cdilog("startListener: assign listener %d (link %d) for %.64s %s %.64s",i,lid,devName,xc,devProperty);
  return ml;
}
void stopAllListeners(void)
{
  int i;
  cdiListener *ml=NULL;
  for (i=0; i<nCdiListeners; i++)
  {
    if ((ml=cdiListenerTable[i]) == NULL) continue;
    stopListener(ml);
  }
  nCdiListeners = 0;
  return;
}
void freeListenerMemory(void)
{
  stopAllListeners();
  if (cdiListenerTable != NULL) free(cdiListenerTable);
  cdiListenerTable = NULL;
  isExiting = TRUE;
  return;
}
int stopListener(cdiListener *ml)
{
  int mlid, cc;
  cdiListener *m = NULL;
  if (ml == NULL) return non_existent_elem;
  cdilog("stopListener: stop listener %d (link %d)",ml->mlid,ml->lid);
  cc = CloseLink(ml->lid);
  if (cc != 0) cdilog("stopListener: could not stop listener link %d (link %d) : %s",ml->mlid,ml->lid,cc2str(cc));
  mlid = ml->mlid;
  if (cdiListenerTable[mlid] == NULL) return not_running;
  if (ml->rbbuf != NULL) free(ml->rbbuf);
  if (ml->outbuf != NULL) free(ml->outbuf);
  if (ml->inbuf != NULL) free(ml->inbuf);
  if (ml->devName != NULL) free(ml->devName);
  free(ml);
  cdiListenerTable[mlid] = NULL;
  return 0;
}
int formatsCanMap(short fmt1,short fmt2)
{
  if (fmt1 == fmt2) return TRUE;
  if (!isNumberFormat(fmt1)) return FALSE;
  if (!isNumberFormat(fmt2)) return FALSE;
  return TRUE;
}
int mapOutputData(DTYPE *d,cdiListener *ml)
{ /* used locally: no NULL ptr checking */
  int siz = MIN(d->dArrayLength,ml->dout.dArrayLength);
  d->dTimeStamp = ml->timestamp;
  if (d->dFormat == ml->dout.dFormat)
  {
    siz *= fmtsizeof(d->dFormat);
    memcpy(d->data.vptr,ml->rbbuf,siz);
    return 0;
  }
  switch (ml->dout.dFormat)
  {
    case CF_BYTE:
      return GetValuesAsByte(d,ml->rbbuf,siz);
    case CF_INT16:
      return GetValuesAsShort(d,(short *)ml->rbbuf,siz);
    case CF_INT32:
      return GetValuesAsLong(d,(int *)ml->rbbuf,siz);
    case CF_FLOAT:
      return GetValuesAsFloat(d,(float *)ml->rbbuf,siz);
    case CF_DOUBLE:
      return GetValuesAsDouble(d,(double *)ml->rbbuf,siz);
    default:
      break;
  }
  return illegal_format;
}
int nCdiListeners = 0;
cdiListener *getListener(char *devName,char *devProperty,DTYPE *dout,DTYPE *din)
{
  cdiListener *ml=NULL;
  UINT32 dsiz;
  int i;
  char *xc = strlen(devName) > 64 ? "(+device sequence)" : "";

  if (cdiListenerTable == NULL) return NULL;
  if (cdiDebugLevel) dbglog("getListener: looking for %.64s %s %.64s among %d registered listeners",devName,xc,devProperty,nCdiListeners);
  for (i=0; i<nCdiListeners; i++)
  {
    if ((ml=cdiListenerTable[i]) == NULL) continue;
    if (strcmp(devName,ml->devName)) continue;
    if (strlen(devProperty) > 0 && strncmp(devProperty,ml->devProperty,PROPERTY_NAME_SIZE)) continue;
    if (dout != NULL)
    {
      if (!formatsCanMap(dout->dFormat,ml->dout.dFormat)) continue;
      if (dout->dArrayLength != ml->outlen) continue;
    }
    if (din != NULL && ml->din.dArrayLength != 0)
    {
      if (din->dFormat != ml->din.dFormat) continue;
      if (din->dArrayLength != ml->din.dArrayLength) continue;
      dsiz = ml->din.dArrayLength*fmtsizeof(ml->din.dFormat);
      if (memcmp(din->data.vptr,ml->inbuf,dsiz)) continue;
    }
    if (cdiDebugLevel) dbglog("getListener: found listener (%d of %d)",i,nCdiListeners);
    break;
  }
  return i < nCdiListeners ? ml : NULL;  
}
int cdiDumpListeners(int i1,int i2,int i3,int i4)
{
  cdiListener *ml=NULL;
  int i,n=0;
  char *xc;
  i1 = i1; i2 = i2; i3 = i3; i4 = i4;
  dbglog("Currently running CDI Listeners :\n");
  if (cdiListenerTable == NULL) return 0;
  for (i=0; i<nCdiListeners; i++)
  {
    if ((ml=cdiListenerTable[i]) == NULL) continue;
    xc = strlen(ml->devName) > 64 ? "(+device sequence)" : "";
    dbglog("%.64s %s %.64s @ %d ms : last access %10.3f <%d>",ml->devName,xc,ml->devProperty,ml->pollrate,ml->timestamp,ml->status);
    n++;
  }
  if (nCdiListeners > 0) dbglog("\nTotal CDI Listeners : %d active from %d\n",n,nCdiListeners);
  return 0;
}
int getDataFromListener(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,double *dTime,int rate,int flags)
{
  cdiListener *ml;
  struct timespec ts;
  char *xc;
  int cc, wait_cnt=0, to_cnt=0;
  struct timeval tv_start, tv_now;

  if (devName == NULL || devProperty == NULL || dout == NULL) 
  { /* used only internally -> this should never happen ! */
    if (cdiDebugLevel) dbglog("getDataFromListener: device %.64s property %.64s data %s",
      devName != NULL ? devName : "null", devProperty != NULL ? devProperty : "null", dout != NULL ? "okay" : "null");
    return argument_list_error;
  }
  xc = strlen(devName) > 64 ? "(+device sequence)" : "";
  if (cdiDebugLevel) dbglog("getDataFromListener: %.64s %s %.64s",devName,xc,devProperty);
  if ( (ml=getListener(devName,devProperty,dout,din)) == NULL && 
       (ml=startListener(devName,devProperty,dout,din,rate,flags)) == NULL )
  {
    if (cdiDebugLevel) dbglog("getDataFromListener: startListener returns %s",erlst[startListenerCompCode]);
    return startListenerCompCode;
  }
  if (dout->dFormat != ml->dout.dFormat) 
  {
    if (cdiDebugLevel) dbglog("getDataFromListener: calling format %d not equal to listener format %d",
      dout->dFormat,ml->dout.dFormat);
   /* return illegal_format; */
  }
  if (dout->data.vptr == NULL) 
  {
    if (cdiDebugLevel) dbglog("getDataFromListener: data pointer is null !");
    return argument_list_error;
  }
  if (flags & CM_WAIT)
  {
    if (cdiDebugLevel) dbglog("getDataFromListener: %s %s wait for timestamp!",devName,devProperty);
    gettimeofday(&tv_start,(struct timezone *)NULL);
    while (ml->timestamp == 0)
    {
      ts.tv_sec = 0;
      ts.tv_nsec = LSNR_STARTUP_WAITTIME;
      cdi_nanosleep(&ts,NULL);
      if (wait_cnt++ > 200)
      {
        gettimeofday(&tv_now,(struct timezone *)NULL);
        if (cdiDebugLevel) 
          dbglog("getDataFromListener: %s %s still waiting for timestamp! (%d msec)",
            devName,devProperty,MSECS(tv_now,tv_start));
        wait_cnt = 0;
        to_cnt++;
      }
      if (to_cnt > 3)
      {
        cdilog("getDataFromListener: %s %s abandon listener startup wait (%d msec)",
          devName,devProperty,MSECS(tv_now,tv_start));
        break;
      }
    }
  }
  if ((cc=mapOutputData(dout,ml)) != 0) return cc;
  if (dTime != NULL) *dTime = dout->dTimeStamp;
  if (cdiDebugLevel) dbglog("getDataFromListener: %s %s returning status %d (timestamp %g)",
    devName,devProperty,ml->status,ml->timestamp);
  cc = ml->status;
  if (cc && dout->dArrayLength > 0 && !isPrimitiveFormat(dout->dFormat)) cc |= CE_SENDDATA;
  return cc;
}
