/*
 * sedactest.c -- sedac driver test program
 *
 */
#include "prolog.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <time.h>

#ifdef UNIX
#include <unistd.h>
#include <ctype.h>
#include <sys/mman.h>
#include <errno.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/kd.h>
#include <sys/time.h>
#endif

#include "tinecdi.h"
#include "rpctypes.h"
#include "..\\BusPlugs\Platform\win32\SedPc\\cdiSedpcLib.h"


//#include "..\\BusPlugs\Platform\win32\CanEsd\\cdiCanLib.h"
#define SDO_TRANSMIT 0x600 /* request on remote node (read or write) */
#define SDO_RECEIVE 0x580 /* device back command on sdo_transmit(read or write) */
#define NMT_COMMANDS 0x0
#define NMT_START_REMOTE 0x1 /* NMT data byte */
#define NMT_STOP_REMOTE 0x2
#define NMT_ENTER_PREOPERATION 0x80
#define NMT_RESET_NODE 0x81
#define NMT_RESET_COMMUNICATION 0x82
#define PDO_TRANSMIT	0x200
#define PDO_DI_READ		0x180	// digital input

/* open CAN parameter index */
enum openCanParamIndex
{
  paramCanRate, /* e.g 500 in KHz */
};
enum execCanParamIndex
{
  canOpenId, /* can open id */
  canOpenObj, /* can open object number */
  canOpenSubIndex,
  canBroadcast, /* canID forced to be 0, one telegram for all devices */
  canAutoReceive, /* device sends data regulary (for Read or ReadthenWrite) */
};

#define	nLine		2
#define	nSedLine	1
#define	nCanLine	4

#define	maskLSOctet	0xff00
#define	maskMSOctet	0x00ff
#define	mask7Bit	0x7f

#define NUM_ANSAGE_TEXT		2	// start and end text
#define MAX_ANSAGE_LETTER	12	// Tunnel Sucher Gebiete

char aliveText[MAX_ANSAGE_LETTER];
char *hdr = "RBTSTI,RETSTI,ABTSTI,AETSTI,";
char AnsageRequestText[NUM_ANSAGE_TEXT][MAX_ANSAGE_LETTER];
char AnsageResponseText[NUM_ANSAGE_TEXT][MAX_ANSAGE_LETTER];

extern int sendSerialPort(int channelNumber, char *psOutput);
//extern int SedacSimulate;		// simulate true
int SedacSimulate;		// simulate true

#define NDATA	4096
char wtData[2 * NDATA];
char rtData[2 * NDATA];

char backstr[40];
int backlength = 40;
char back2str[40];
int back2length = 40;
char reqstr[140];
int reqlength = 140;

/************************************************************/
/*	8 bit 1s-complement sum of a string, 7 bits returned	*/
/************************************************************/
char sum7(char *ps)
{ short sum = 0;

  while ( *ps )
  {
    sum += *ps++;
    if ( sum & maskLSOctet ) sum = maskMSOctet & ++sum;
  }
  return sum & mask7Bit;  
}

/********************************************/
void makeAnsageText(void)
{ int j;
  char *pstart, *pdest, retBBC;

  memset(aliveText, '\0', MAX_ANSAGE_LETTER);
  aliveText[0] = '\2';	// starts with STX char
  strcpy(&aliveText[1],"CBAGOK");
  retBBC = sum7( &(aliveText[1]) );
  aliveText[7]='\3';
  aliveText[8]=retBBC;
  printf("return text: %s\n", aliveText);

  pstart = hdr;
  printf("text %s\n", pstart);
  for ( j=0; j<4; j++ )		// skip nText cells
  {
    if ( j < 2 ) {
      memset(AnsageRequestText[j], '\0', MAX_ANSAGE_LETTER);
      memset(AnsageResponseText[j], '\0', MAX_ANSAGE_LETTER);
      AnsageRequestText[j][0] = '\2';	// starts with STX char
      AnsageResponseText[j][0] = '\2';
 	  }
    pdest = strchr( pstart, ',' );
    memcpy(j < 2 ? &(AnsageRequestText[j%2][1]) : &(AnsageResponseText[j%2][1]), pstart, pdest-pstart);
    retBBC = sum7(j < 2 ? &(AnsageRequestText[j%2][1]) : &(AnsageResponseText[j%2][1]));
    if (j < 2)
    {
      AnsageRequestText[j%2][pdest-pstart+1] = '\3';
      AnsageRequestText[j%2][pdest-pstart+2] = retBBC;
    }
    else
    {
      AnsageResponseText[j%2][pdest-pstart+1] = '\3';
      AnsageResponseText[j%2][pdest-pstart+2] = retBBC;
    }
    printf("text Number: %d; %s\n", j, j < 2 ? AnsageRequestText[j%2] : AnsageResponseText[j%2]);
    pstart = pdest + 1;
  }
  printf("return text: %s\n", aliveText);
}

/********************************************/
void callBack(int index, int error)
{ int i, stat, errorArray[100];
  static cont;
  short *pSh;
  
  pSh = (short *)rtData;
	printf("**************callback: index: %x, error: %d; rdData: %d; str: %s; counter: %d\n",index,error,pSh[1],backstr, cont++);
  for (i=0; i<10; i++) printf("short data: 0x%x \n",pSh[i]);
  stat = cdi_GetRequestErrors(index, errorArray);
  printf("cdi_GetRequestError; return: %d, error: 0x%x, 0x%x... \n",stat, errorArray[0], errorArray[1]);
}

/********************************************/
void PreSystemInit(void)
{
}
/********************************************/
void PostSystemInit(void)
{
}

/********************************************/
int main(int argc, char **argv)
{ int i, comport = 1, j, k, nr = 1;
  int parameters[10], getparameters[10];
  int lineIndex[nLine];
  int canIndex[nCanLine];
  int sedacIndex[nSedLine];
  int tIndex[100], errorbuffer[100];
  char *tryText="This is a test string";
  
  int ret, num1, num2;
  char devName[32];
  char *forms = "%d,%dY";
  void *pmem[2];
  char strdata[8];
  int dataPatternWindow;
  int cc=0;
  DTYPE din, dout;

// test database!!!!
# define READ_DATABASE
# ifdef READ_DATABASE
//  SedacSimulate = 1;		// simulate true
//  readCdiDataBase();
  //initTineCDI();
# else
# define TESTRS232
# endif

  makeAnsageText();

  pmem[0] = &num1;
  pmem[1] = &num2;
  strcpy(devName,"12345,987Y");
  ret = sscanf(devName,(char *)forms,pmem[0],pmem[1]);
  printf("retCode: %d; number : %d,,, %d\n",ret, *(int *)pmem[0], *(int *)pmem[1]);
/*2006-4
  parameters[paramBaudrate] = 9600;
  parameters[param8bits] = 0;
  parameters[param2StopBits] = 1;
  parameters[paramParity] = 1;
  parameters[paramOddParity] = 0;
  parameters[paramFlowControl] = 0;
  parameters[paramCanonicalInput] = 0;
*/
  for (i=0; i< NDATA; i++) ((short *)wtData)[i] = i + 1;

// test redirect!!!!
//#define TESTREDIRECT
#ifdef TESTREDIRECT
  maxCdiRedirectedDevices = 16;
  cdi_RegisterRedirectedDevice("SerialTestDevice", 0,"/TEST/CDISRV.CDI/SerialTestDevice");

#endif

// test parameters!!!!
#ifdef TESTPARAM
  for(j=0;j<2;j++)
  {
    for(i=0;i<4;i++)
    {
      sprintf(devName,"ParamTest_%d",i+1);
/* 2006-4
      parameters[paramBaudrate] = 9600 + i + j*10;
*/
      cdi_RegisterParameters(devName, parameters);
    }
    for(i=0;i<4;i++)
    {
      sprintf(devName,"ParamTest_%d",i+1);
      cdi_GetParameters(devName, getparameters);
/* 2006-4
      printf("param: %d \n",getparameters[paramBaudrate]);
*/
    }
  }
#endif

// test rs232!!!!
//#define TESTRS232
#ifdef TESTRS232
  for (i=0; i<nLine; i++ ) lineIndex[i] = cdi_Open("RS232", comport+i, parameters, 1);

  cdi_RegisterDevice(lineIndex[1], "SerialTestDevice", NULL, 0, nr++, 0, 0);
  millisleep(2000);
/*
  SedacSimulate = 1;		// simulate true
  sedacIndex[0] = cdi_Open("SEDAC", comport, parameters, 20);
  for (i=0; i<20; i++ )
  {
    sprintf(devName,"SedacDevNr_%d",i+1);
    cdi_RegisterDevice(sedacIndex[0], devName, NULL, i, i+1, 0);
  }
*/
#ifdef DO_THE_TEST_STUFF
  din.dArrayLength = strlen(AnsageRequestText[0]);
  din.dFormat = CF_TEXT;
  din.data.vptr = AnsageRequestText[0];
  for (i=0; i<5; i++)
  {
    j = ExecLinkEx("/LOCAL/HARDWARE/SerialTestDevice","WRITE",NULL,&din,CA_WRITE, 3000);
//    j = cdi_Request("SerialTestDevice", NULL, 0, AnsageRequestText[0], strlen(AnsageRequestText[0]), NULL, NULL, &i, cdiWriteFlag, maskChange, 3000, NULL, 0, -1);
  	printf("cdi_exec return error code: %d \n",j);
    millisleep(3000);
  }
/*
  j = cdi_Request("SerialTestDevice", NULL, 0, AnsageRequestText[0], strlen(AnsageRequestText[0]), NULL, NULL, &i, cdiWriteFlag, maskChange, 3000, callBack, 1, -1);
  printf("cdi_POll return: %x \n",j);
*/

  j = cdi_RegisterParameters("ParamRS232alive", parameters);
  printf("return Register parameters: %d\n",j);

  dout.dArrayLength = backlength;
  dout.dFormat = CF_TEXT;
  dout.data.vptr = backstr;
  strcpy(backstr,aliveText);
  sprintf(reqstr,"/LOCAL/HARDWARE/SerialTestDevice/%s","ParamRS232alive");
  printf("%s;;; %s;;; %s\n",reqstr, aliveText, AnsageResponseText[0]);

  j = AttachLinkEx(reqstr,"READ",&dout,&din,CA_READ,5000,callBack,CM_REFRESH|CM_SEEKPATTERN,-1);
//  j = cdi_Request("SerialTestDevice", backstr, backlength, AnsageRequestText[0], strlen(AnsageRequestText[0]), aliveText, NULL, &i, cdiReadFlag, maskCompEqual, 5000, callBack, 1, -1);
  printf("cdi_Request return: %x \n",j);

  dout.dArrayLength = back2length;
  dout.dFormat = CF_TEXT;
  dout.data.vptr = back2str;
  strcpy(back2str,AnsageResponseText[0]);
  j = AttachLinkEx(reqstr,"WRITEREAD",&dout,&din,CA_READ,5000,callBack,CM_REFRESH|CM_SEEKPATTERN,-1);
//  j = cdi_Request("SerialTestDevice", back2str, back2length, AnsageRequestText[0], strlen(AnsageRequestText[0]), AnsageResponseText[0], NULL, &i, cdiWriteReadFlag, maskCompEqual, 5000, callBack, 1, -1);
  printf("cdi_Request return: %x  \n",j);
/*
  millisleep(20000);

	k = cdi_Request("SerialTestDevice", NULL, 0, AnsageRequestText[0], strlen(AnsageRequestText[0]), NULL, NULL, &i, cdiWriteFlag, maskChange, 3000, NULL, 0, -1);
	printf("WWWWWWWWWWWWWWWW cdi_exec return: %d \n",k);

  i = cdi_CancelRequest(j);
  printf("--------------cdi_Cancel return: %d  \n",i);
*/
  millisleep(10000);
  i = cdi_CancelRequest(j);
  printf("--------------cdi_Cancel return: %d  \n",i);
#endif // DO_THE_TEST_STUFF
  
#endif
// test rs232!!!!


#ifdef TEST_ESD_CAN
// test can!!!!
  printf("busname: %s\n",cdi_getBusNameForDevice("ESDCAN_INIT"));
	*(short *)strdata = (short)NMT_START_REMOTE; // start an all
	parameters[canOpenId] = NMT_COMMANDS;			// can open id
	parameters[canOpenObj] = 0;				// can open object number
	parameters[canOpenSubIndex] = 0;
	parameters[canBroadcast] = 1;			// canID forced to be 0, one telegram to everyone
	parameters[canAutoReceive] = 0;			// device sends data regulary (for Read or ReadthenWrite)
//	k = cdi_Request("ESDCAN_1", NULL, 0, strdata, 1, NULL, NULL, parameters, cdiWriteFlag, maskChange, 3000, NULL, 0, -1);
//	printf("WWWWWWWWWWWWWWWW cdi_exec return error: %d\n",k);

  tIndex[0] = cdi_Request("ESDCAN_INIT", rtData, cdiShort, 2, strdata, cdiShort, 2, &dataPatternWindow, NULL, parameters, cdiWriteFlag, anyData, 3000, callBack, 0, &errorbuffer[0], -1, 1);
  printf("WWWWWWWWWWWWWWWW cdi_exec return error: %d\n",tIndex[0]);
	j = cdi_StartRequest(tIndex[0]);
	printf("***start request (read): ret: %d; data: %hd, %hd\n", j, ((short *)rtData)[0], ((short *)rtData)[1]);

  millisleep(5000);

	parameters[canOpenId] = SDO_TRANSMIT;			// can open id
//	parameters[canOpenId] = PDO_DI_READ;			// can open id
	parameters[canOpenObj] = 0x1000;				// can open object number
	parameters[canOpenSubIndex] = 0;
	parameters[canBroadcast] = 0;			// canID forced to be 0, one telegram to everyone
	parameters[canAutoReceive] = 0;
  //tIndex[1] = cdi_Request("ESDCAN_READ1", rtData, cdiShort, 4, strdata, cdiShort, 2, &dataPatternWindow, NULL, parameters, cdiReadFlag, anyData, 5000, callBack, 0, &errorbuffer[0], -1, 1);
  tIndex[1] = cdi_Request("ESDCAN_READ1", rtData, cdiInteger, 4, strdata, cdiShort, 2, &dataPatternWindow, NULL, NULL, cdiReadFlag, anyData, 5000, callBack, 0, &errorbuffer[0], -1, 1);
  printf("WWWWWWWWWWWWWWWW cdi_exec return error: %d\n",tIndex[1]);
	j = cdi_StartRequest(tIndex[1]);
	printf("***start request (read): ret: %d; data: %hd, %hd\n", j, ((short *)rtData)[0], ((short *)rtData)[1]);

  millisleep(1000);

	parameters[canOpenId] = PDO_DI_READ;			// can open id
	parameters[canOpenObj] = 0x1800;				// can open object number
  for (i=0; i<1; i++)
  {
	parameters[canOpenSubIndex] = i;
  //tIndex[2] = cdi_Request("ESDCAN_READ2", rtData, cdiShort, 4, strdata, cdiShort, 2, &dataPatternWindow, NULL, parameters, cdiReadFlag, anyData, 5000, callBack, 0, &errorbuffer[0], -1, 1);
  tIndex[2] = cdi_Request("ESDCAN_READ2", rtData, cdiShort, 4, strdata, cdiShort, 2, &dataPatternWindow, NULL, NULL, cdiReadFlag, anyData, 5000, callBack, 0, &errorbuffer[0], -1, 1);
  printf("WWWWWWWWWWWWWWWW cdi_exec return error: %d\n",tIndex[2]);
	j = cdi_StartRequest(tIndex[2]);
	printf("***start request (read): ret: %d; data: %hd, %hd\n", j, ((short *)rtData)[0], ((short *)rtData)[1]);
  millisleep(1000);

	j = cdi_CancelRequest(tIndex[2]);
	printf("cancel request (read): ret: %d\n", j);
  millisleep(1000);

  }

  millisleep(1000);
	parameters[canOpenId] = PDO_TRANSMIT;			// can open id
  rtData[0]=0xf0;
  tIndex[2] = cdi_Request("ESDCAN_INIT", rtData, cdiChar, 1, strdata, cdiShort, 2, &dataPatternWindow, NULL, parameters, cdiWriteFlag, anyData, 5000, callBack, 0, &errorbuffer[0], -1, 1);
  printf("WWWWWWWWWWWWWWWW cdi_exec return error: %d\n",tIndex[2]);
	j = cdi_StartRequest(tIndex[2]);
	printf("***start request (read): ret: %d; data: %hd, %hd\n", j, ((short *)rtData)[0], ((short *)rtData)[1]);
  millisleep(1000);

	j = cdi_CancelRequest(tIndex[2]);
	printf("cancel request (read): ret: %d\n", j);


  millisleep(15000);

	j = cdi_CancelRequest(tIndex[0]);
	printf("cancel request (read): ret: %d\n", j);

	j = cdi_CancelRequest(tIndex[1]);
	printf("cancel request (read): ret: %d\n", j);

  printf("type a char to continue!\n");

  getchar();
#endif  // TEST_ESD_CAN


//#define TESTCAN
#ifdef TESTCAN

  parameters[paramCanRate] = 500;		// e.g 500 in KHz

  canIndex[0] = cdi_Open("CAN", 1, parameters, 32);
  for (i=0; i<20; i++ )
  {
    sprintf(devName,"CanDevNr_%d",i+1);
    cdi_RegisterDevice(canIndex[0], devName, NULL, i, nr++, i+1, 0);
  }

//	strdata[0] = NMT_ENTER_PREOPERATION;
	strdata[0] = NMT_START_REMOTE;
	parameters[canOpenId] = NMT_COMMANDS;			// can open id
	parameters[canOpenObj] = 0;				// can open object number
	parameters[canOpenSubIndex] = 0;
	parameters[canBroadcast] = 1;			// canID forced to be 0, one telegram to everyone
	parameters[canAutoReceive] = 0;			// device sends data regulary (for Read or ReadthenWrite)
	k = cdi_Request("CanDevNr_4", NULL, 0, strdata, 1, NULL, NULL, parameters, cdiWriteFlag, maskChange, 3000, NULL, 0, -1);
	printf("WWWWWWWWWWWWWWWW cdi_exec return error: %d\n",k);
    
  millisleep(1000);
//	{0x3011, 0, 1, 0x19,	0},	// 25ms for logic block 0 and block
	parameters[canOpenId] = SDO_TRANSMIT;			// can open id
//	parameters[canOpenId] = PDO_DI_READ;			// can open id
	parameters[canOpenObj] = 0x3011;				// can open object number
	parameters[canOpenSubIndex] = 0;
	parameters[canBroadcast] = 0;			// canID forced to be 0, one telegram to everyone
	parameters[canAutoReceive] = 0;
	dataPatternWindow = 32;
	wtData[0] = 32;
//	k = cdi_Request("CanDevNr_4", strdata, 1, NULL, 0, NULL, NULL, &dataWindow, parameters, cdiReadFlag, maskChange, 3000, callBack, 0, -1);

	k = cdi_Request("CanDevNr_4", strdata, 1, wtData, 1, &dataPatternWindow, NULL, parameters, cdiReadWriteFlag, maskCompNotEqual, 3000, callBack, 1, -1);
//	k = cdi_Request("CanDevNr_4", strdata, 1, wtData, 1, &dataPatternWindow, NULL, parameters, cdiReadWriteFlag, maskCompNotEqual, 3000, NULL, 0, -1);

//	k = cdi_Request("CanDevNr_5", strdata, 8, NULL, 0, &dataPatternWindow, NULL, parameters, cdiReadFlag, maskChange, 5000, callBack, 1, -1);
//	k = cdi_Request("CanDevNr_5", strdata, 8, NULL, 0, NULL, NULL, parameters, cdiReadFlag, maskChange, 0, callBack, 1, -1);
	printf("WWWWWWWWWWWWWWWW cdi_exec return: %d,  data: %d \n",k, strdata[0]);

  millisleep(5000);
  getchar();
#endif

#define TESTTWINCAT
#ifdef TESTTWINCAT
  NGdebug = 1;
  printf("busname: %s\n",cdi_getBusNameForDevice("MAIN.iCount"));
  printf("sleep 10 seconds!\n");
  millisleep(10000);

  tIndex[0] = cdi_Request("#1", rtData, cdiShort, 2, wtData, cdiShort, 2, &dataPatternWindow, NULL, parameters, cdiReadFlag, anyData, 3000, callBack, 1, &errorbuffer[0], -1, 1);
  printf("WWWWWWWWWWWWWWWW cdi_exec return error: %d\n",tIndex[0]);
	j = cdi_StartRequest(tIndex[0]);
	printf("***start request (read): ret: %d; data: %hd, %hd\n", j, ((short *)rtData)[0], ((short *)rtData)[1]);


  *(short *)wtData = 14;
  tIndex[0] = cdi_Request("#2", rtData, cdiShort, 2, wtData, cdiShort, 2, &dataPatternWindow, NULL, parameters, cdiWriteFlag, anyData, 3000, NULL, 0, &errorbuffer[0], -1, 1);
  printf("WWWWWWWWWWWWWWWW cdi_exec return error: %d\n",tIndex[0]);
	j = cdi_StartRequest(tIndex[0]);
	printf("***start request (read): ret: %d; data: %hd, %hd\n", j, ((short *)rtData)[0], ((short *)rtData)[1]);
  
  
  getchar();
  exit(0);

#endif

#define TESTSEDAC
#ifdef TESTSEDAC
  SedacSimulate = 0;		// simulate true
  NGdebug = 1;
/*
  sedacIndex[0] = cdi_Open("SEDPC", comport, parameters, 20, 0);
  printf("open sedac: %d\n", sedacIndex[0]);
  for (i=0; i<20; i++ )
  {
    sprintf(devName,"SedacDevNr_%d",i+1);
    if ( i == 0 ) k=145;
    else if ( i == 1 ) k=32;
    else if ( i == 2 ) k=176;
    else if ( i == 3 ) k=192;
    else k=i;
    j = cdi_RegisterDevice(sedacIndex[0], devName, NULL, i, nr++,15, k);
    printf("register sedac device: %s; ret: %d\n",devName, j);
  }
*/
  parameters[sedacSubOffset] = 0;

// test group call
  rtData[0] = 55;
  rtData[1] = 22;
//  printf("busname: %s\n",cdi_getBusNameForDevice("ESDCAN_1"));
  printf("busname: %s\n",cdi_getBusNameForDevice("ESDCAN_INIT"));
  printf("busname: %s\n",cdi_getBusNameForDevice("SerialDevice1"));
  printf("busname: %s\n",cdi_getBusNameForDevice("SedacDevNr_1"));

  printf("sleep 10 seconds!\n");
  millisleep(10000);

  tIndex[0] = cdi_Request("#1-#100", rtData, cdiShort, 2, wtData, cdiShort, 2 * NDATA, &dataPatternWindow, NULL, parameters, cdiWriteFlag, anyData, 3000, callBack, 1, &errorbuffer[0], -1, 1);
  printf("WWWWWWWWWWWWWWWW cdi_exec return error: %d\n",tIndex[0]);
	j = cdi_StartRequest(tIndex[0]);
	printf("***start request (read): ret: %d; data: %hd, %hd\n", j, ((short *)rtData)[0], ((short *)rtData)[1]);

  getchar();

  tIndex[1] = cdi_Request("SedacDevNr_3,SedacDevNr_4", rtData, cdiShort, 2, wtData, cdiShort, 2 * NDATA, &dataPatternWindow, NULL, parameters, cdiWriteFlag, anyData, 3000, callBack, 1, &errorbuffer[0], -1, 1);
  printf("WWWWWWWWWWWWWWWW cdi_exec return error: %d\n",tIndex[1]);
	j = cdi_StartRequest(tIndex[1]);
	printf("***start request (read): ret: %d; data: %hd, %hd\n", j, ((short *)rtData)[0], ((short *)rtData)[1]);


  printf("sleep 40 seconds!\n");
  millisleep(40000);
	j = cdi_CancelRequest(tIndex[0]);
	printf("cancel request (read): ret: %d\n", j);
	j = cdi_CancelRequest(tIndex[1]);
	printf("cancel request (read): ret: %d\n", j);

  printf("type a char to continue!\n");
  getchar();



  //	tIndex[0] = cdi_Request("SerialDevice1", NULL, cdiShort, 0, AnsageRequestText[0], cdiChar, strlen(AnsageRequestText[0]), NULL, NULL, &i, cdiWriteFlag, anyData, 3000, NULL, 0, NULL, -1, 1);
	tIndex[0] = cdi_Request("SerialDevice1", NULL, cdiShort, 0, tryText, cdiChar, strlen(tryText), NULL, NULL, &i, cdiWriteFlag, anyData, 3000, NULL, 0, NULL, -1, 1);
	printf("serial WWWWWWWWWWWWWWWW cdi_exec return: %d \n",tIndex[0]);
	j = cdi_StartRequest(tIndex[0]);
	printf("***start request (read): ret: %d; data: %hd, %hd\n", j, ((short *)rtData)[0], ((short *)rtData)[1]);
/*
  printf("sleep 20 seconds!\n");
  millisleep(20000);
*/
  tIndex[0] = cdi_Request("SedacDevNr_1", rtData, cdiShort, 2, wtData, cdiShort, 2 * NDATA, &dataPatternWindow, NULL, parameters, cdiWriteFlag, anyData, 3000, callBack, 1, &errorbuffer[0], -1, 1);
  printf("WWWWWWWWWWWWWWWW cdi_exec return error: %d\n",tIndex[0]);

  tIndex[1] = cdi_Request("SedacDevNr_1", rtData, cdiShort, 2 * NDATA, wtData, cdiShort, 2, &dataPatternWindow, NULL, parameters, cdiReadFlag, anyData, 3000, callBack, 1, &errorbuffer[1], -1, 1);
  printf("WWWWWWWWWWWWWWWW cdi_exec return error: %d\n",tIndex[1]);

  tIndex[2] = cdi_loadGroupRequests(tIndex[0], tIndex[1]);
  printf("WWWWWWWWWWWWWWWW cdi_exec return error: %d\n",tIndex[2]);

	j = cdi_StartRequest(tIndex[0]);
	printf("***start request (read): ret: %d; data: %hd, %hd\n", j, ((short *)rtData)[0], ((short *)rtData)[1]);

  millisleep(20000);

	j = cdi_CancelRequest(tIndex[0]);
	printf("cancel request (read): ret: %d\n", j);
	
	j = cdi_StopRequest(tIndex[0]);
	printf("stop request (read): ret: %d\n", j);

/*//*
  tIndex[1] = cdi_Request("SedacDevNr_1", rtData, cdiShort, 2 * NDATA, wtData, cdiShort, 2, &dataPatternWindow, NULL, parameters, cdiReadFlag, anyData, 4000, callBack, 1, -1, 1);
  printf("WWWWWWWWWWWWWWWW cdi_exec(read) return error: %d\n",tIndex[1]);
	j = cdi_StartRequest(tIndex[1]);
	printf("star request (read): ret: %d\n", j);
*/
	/*
  millisleep(10000);
	j = cdi_StopRequest(tIndex[1]);
	printf("stop request (read): ret: %d\n", j);
  millisleep(10000);
	j = cdi_StartRequest(tIndex[1]);
	printf("star request (read): ret: %d\n", j);
  */

// test data pattern window
/*
  *(short *)strdata = 0;
  dataPatternWindow = 1;
  tIndex[0] = cdi_Request("SedacDevNr_1", strdata, cdiShort, 2, wtData, cdiShort, 2 * NDATA, &dataPatternWindow, NULL, parameters, cdiReadWriteFlag, maskChangeWindow, 3000, callBack, 1, -1, 0);
  printf("WWWWWWWWWWWWWWWW cdi_exec return error: %d\n",tIndex[0]);
	j = cdi_StartRequest(tIndex[0]);
	printf("star request (read): ret: %d, data: %d\n", j, *(short *)strdata);
  millisleep(12000);
	j = cdi_StopRequest(tIndex[0]);
//	j = cdi_StartRequest(tIndex[0]);
//	printf("star request (read): ret: %d, data: %d\n", j, *(short *)strdata);
*/

/*
  *(short *)wtData = 32;
  tIndex[0] = cdi_Request("SedacDevNr_1", strdata, cdiShort, 2, wtData, cdiShort, 2 * NDATA, &dataPatternWindow, NULL, parameters, cdiWriteFlag, maskCompNotEqual, 3000, NULL, 0, -1, 1);
  printf("WWWWWWWWWWWWWWWW cdi_exec return error: %d\n",tIndex[0]);
  if ( tIndex[0] > 0 )
  {
	   	for ( num1=0; num1<10; num1++)
		{
	j = cdi_StartRequest(tIndex[0]);
	printf("star request number: %d: ret: %d\n", num1, j);
  millisleep(2000);
  *(short *)wtData = 255;
	j = cdi_StartRequest(tIndex[0]);
	printf("2nd star request number: %d: ret: %d\n", num1, j);
  millisleep(2000);
		}
	j = cdi_CancelRequest(tIndex[0]);
  }
*/

/*
  for (i=0; i<10; i++)
  {
    k = cdi_Request("SedacDevNr_1", strdata, cdiShort, 2, wtData, cdiShort, 2, NULL, NULL, parameters, cdiReadFlag, maskChange, 3000, NULL, 0, -1, 0);
    printf("WWWWWWWWWWWWWWWW cdi_exec return error: %d; data : %d\n",k, *(short *)strdata);
  }

  *(short *)wtData = 88;//64;
  k = cdi_Request("SedacDevNr_4", strdata, 2, wtData, 2, &dataPatternWindow, NULL, parameters, cdiWriteFlag, maskCompNotEqual, 3000, NULL, 0, -1, 0);
  printf("WWWWWWWWWWWWWWWW cdi_exec return error: %d\n",k);
  k = cdi_Request("SedacDevNr_4", strdata, 2, wtData, 2, NULL, NULL, parameters, cdiReadFlag, maskChange, 3000, NULL, 0, -1);
  printf("WWWWWWWWWWWWWWWW cdi_exec return error: %d; data : %d\n",k, *(short *)strdata);

  *(short *)wtData = 16;
  dataPatternWindow = 16;
  k = cdi_Request("SedacDevNr_4", strdata, 2, wtData, 2, &dataPatternWindow, NULL, parameters, cdiReadWriteFlag, maskCompNotEqual, 5000, callBack, 1, -1, 0);
*/  
  millisleep(29000);
	j = cdi_CancelRequest(tIndex[1]);
	printf("cancel request (read): ret: %d\n", j);
  getchar();

#endif

  // start server ...
  PreSystemInit();

  /* initialize RPC server: */
  printf("initializing server\n");
  NGdebug=3;
  if ((cc=SystemInit(TRUE)) != 0)
  {
     printf(erlst[cc]);
     exit(1);
  }

  /* Make all other registrations in PostSystemInit() */

  PostSystemInit();

  printf("starting server cycle\n");
  for(;;)
  {
    SystemCycle(TRUE);
  }
  printf("quitting server\n");

# ifdef RS232
# ifndef TESTREDIRECT
  for (i=0; i<5; i++)
  {
    sendSerialPort(2, AnsageRequestText[0]);
    millisleep(3000);
  }
  sendSerialPort(2, AnsageRequestText[1]);

  getchar();

  for (i=0; i<nLine; i++ ) cdi_Close(lineIndex[i]);
  cdi_Close(canIndex[0]);
  cdi_Close(sedacIndex[0]);
  
# endif
# endif
  return 0;

}


//int main(int argc, char **argv)
//{
//    char *fname;
//	int data, ret, i, forever;
//	int sedacfd;
//  unsigned int nData;
//    unsigned int nRepeat;
//    unsigned int sel, map;
//	time_t tstart, tstop, t1, t2;
//	int tm;
//	fd_set rfds;
//	struct timeval tv;
//	int retval;

//    if (argc != 7 || sscanf(argv[2],"%i", &nData) == 0 || sscanf(argv[3],"%i", &nRepeat) == 0 || sscanf(argv[4],"%i", &sel) == 0 || sscanf(argv[5],"%i", &map) == 0 || sscanf(argv[6],"%i", &forever) == 0)
//	{
//	  fprintf(stderr, "%s: Usage \"%s <file> <numberTelegrams> <repeatTimes> <select> <map> <forever>\"\n", argv[0],
//                argv[0]);
//      exit(1);
//    }
//    fname = argv[1];
//	printf("file name = %s, number of data: %d; N of repeat: %d; Mmap: %d\n", fname, nData, nRepeat, map);


//    printf("wait for a char!!\n");
//	ret = getchar();

//	printf("bye bye!\n");
//	exit(0);
//}
        
