/*
 * 02.01.2012 Modified version of devsrv.c
 * Includes code for parallel port blinkers
 * and PDO Init/Monitor activation
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "tine.h"
#include "cdi.h"
#include "pdoclient.h"
#include "tinecdi.h"      /* for def of cdiSetExportName */
#include "srversion.h"
#include <dlfcn.h>
#include <sys/utsname.h>

#ifdef MULTITHREADED_XX
TINE_EXPORT void SystemSetCycleTimer(void);
#endif

char gCdiFecContext[32];
char gCdiFecServer[32];
char gCdiFecName[32];
char gCdiHome[128];
char gFecHome[128];
char gCdiSubsystem[16];
char gCdiFecDescription[64];
char gCdiFecLocation[64];
char gCdiFecHardware[64];
int gCdiSrvPort = 30;

extern int hdwIoDebugLevel;
void ServerInit(int argc, char *argv[]);

/* --------- Code for PC104 blinker lights ---------------- */
static int blink = 0;   
FILE * PPORT;
void blinkInit() {
  char dummy;
  PPORT=fopen("/dev/parlelport","w"); // uses driver pport.ko
  blink = (PPORT)? 1 : 0;
  if (!blink) { printf("PPORT open fails\n"); return; }
  setvbuf(PPORT,&dummy,_IONBF,1);
}
// -- use PP to drive blinking LED during operation ---------
void blinker() {
  static struct timeval ts; 
  static int tsec_old;
  static unsigned char blnk=0;
  gettimeofday(&ts,0);
  if (ts.tv_sec != tsec_old) {
  fwrite(&blnk,1,1,PPORT);
  blnk = (blnk)? 0 : 255;
  tsec_old = ts.tv_sec;
  }
}

void blinkOff() { // turn off on exit
   static unsigned char blnk=0;
   fwrite(&blnk,1,1,PPORT);
}
// -- -------------------------------------------------------- 
void PreSystemInit(void)
{
  char usrnam[32]; // expenv[128];
  char *usr = getenv("USERNAME");
  if (usr == NULL) usr = "UNKNOWN";
  strncpy(usrnam,usr,32);
  /* System parameters from older version (some default?)     */
  /* shown below.  13.10.10 Which of these have been changed? */
  MinPollingRate = 1000;
  MaxPollingRate = 30000;
  MaxNumContracts = 10000;
  MaxNumClients = 100;
  FeclogDepth = 100;
  MaxRPCTransportSize = 65535;
  putCommandsInFeclog = TRUE; 
  RegisterFecInformation(gCdiFecName,gCdiSubsystem,gCdiFecContext,gCdiFecDescription,gCdiFecLocation,gCdiFecHardware,usrnam,(UINT16)gCdiSrvPort);
  cdiSetExportName(gCdiFecServer); /* this should work on a linux machine */
  PdoClientInit();
  RegisterUserCommand("pdostats",PdoStatus,NULL,NULL,CA_READ);
}

void PostSystemInit(void)
{
  blinkInit();
  return;
}
int main(int argc, char *argv[]) {
  int cc;
  int t0=0,t1=0;
  setAppDate();             // for TINE stock properties
  setAppVersion();
  ServerInit(argc,argv);    // process cmd line args
  PreSystemInit();
  // --- set server to ignore client calls, initialize ---
  SetServerWaiting(TRUE);
  if ((cc=SystemInit(TRUE)) != 0) {
    printf(erlst[cc]);
    exit(1);
  }
  // -- other registrations in PostSystemInit() ----------
  PostSystemInit();
  while(IsServerRunning()==0) {
    SystemCycle(TRUE);  
  }
  printf("Server Init Finished\n");
  cdilog("Server Init Finished\n");
  // ----- Now init TPDO time intervals, seems to work ---
  while(PdoInitCycle()==0) {
    t1 = time(NULL);
    if ((t1-t0) < 2) continue;
    t0 = t1;
    SystemCycle(TRUE); // blocks so can't call very often?
  }
  SubscribeCoDevices();   // will it work here ?
  // ----- Now accept client calls ---
# ifdef MULTITHREADED_XX
  SystemSetCycleTimer();
  Sleep(INFINITE);
# else
  SystemCycle(TRUE);
  millisleep(SYSTEM_TICK);
  SetServerWaiting(FALSE);
  SetPdoStateflag(QDEFAULT); // start monitor
  for(;;)
  {
    SystemCycle(TRUE);
    if (blink) blinker();
    PdoMonitorCycle();
    
  }
# endif
  return 0;
}
/* ServerInit uses cmd line args to set file scope variables for FEC initialization           */
/* command line cdiHdwSrv /n=<fecname> /c=<context> /p=<port> /d=<directory> /h or /? => help */
void ServerInit(int argc, char *argv[]) {
  int i;
  char fecHomeStr[80];
  char *c, *ptr=NULL;
  int fecNameGiven=FALSE, srvNameGiven=FALSE;
  struct utsname unam;
  if (uname(&unam) == 0) ptr = unam.nodename;
  if (ptr == NULL) ptr = getenv("HOSTNAME");
  if (ptr != NULL) 
  {
    strncpy(gCdiFecName,ptr,16); 
    gCdiFecName[16] = 0; 
    strcpy(gCdiFecServer,gCdiFecName);
    strcat(gCdiFecName,".30");
    strcat(gCdiFecServer,".CDI");
  }
  gCdiHome[0] = 0;
  strcpy(gCdiFecContext,"HARDWARE");
  strcpy(gCdiFecDescription,"CDI Raw Hardware Server");
  strcpy(gCdiFecLocation,"unknown");
  strcpy(gCdiFecHardware,"unknown");
  strcpy(gCdiSubsystem,"DIAG");
  for (i=1; i<argc; i++)
  {
    c = argv[i];
    if (*c++ != '/') continue;
    switch (tolower(*c))
    {
      case 'n':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(gCdiFecName,c,32);
          fecNameGiven = TRUE;
        }
        continue;
      case 's':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(gCdiFecServer,c,32);
          srvNameGiven = TRUE;
        }
        continue;
      case 'c':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(gCdiFecContext,c,32);
        }
        continue;
      case 'u':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(gCdiSubsystem,c,16);
        }
        continue;
      case 'l':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(gCdiFecLocation,c,32);
        }
        continue;
      case 'p':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          gCdiSrvPort = atoi(c);
        }
        continue;
      case 'd':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          snprintf(gCdiHome,128,"CDI_HOME=%s",c);
          snprintf(gFecHome,128,"FEC_HOME=%s",c);
          strncpy(fecHomeStr,c,80);
        }
        continue;
      case 'h':
      case '?':
        printf("Usage:\n\n\tcdiHdwSrv [/n=<fecname> /s=<server name> /c=<context> /u=<subsystem> /l=<location> /p=<port offset> /d=<cdi home directory>]\n");
        printf("\nCalling cdiHdwSrv without arguments uses the default settings.\n");
        printf("These are: fecname = <COMPUTERNAME>.30\n");
        printf("           server = <COMPUTERNAME>.CDI\n");
        printf("           context = HARDWARE\n");
        printf("           subsystem = DIAG\n");
        printf("           location = unknown\n");
        printf("           port offset = 30\n");
        printf("           cdi home directory = <CDI_HOME> or <FEC_HOME>\n");
        printf("To override these use any or all of the swithes above.\n");
        printf("\nFor Example:\n\n\tcdiHdwSrv /n=ACCXPDOBPM /s=BPM /c=DORIS /u=DIAG /p=41 /d=L:\\cdi\\ACCXPDOBPM\n");
        exit(0);
      default:
        continue;
    }
  }
  if (fecNameGiven && !srvNameGiven)
  {
    strcpy(gCdiFecServer,gCdiFecName);
    strcat(gCdiFecServer,".CDI");
  }

  if (strlen(gCdiHome) > 0)
  {
    putenv(gCdiHome);
    putenv(gFecHome);
    SetFecHome(fecHomeStr);
  }
}
